
/*

	XDSP example - alignment
	Demonstrates requirement of buffer alignment


*/

#include "stdafx.h"
#include "../../include/xdsp.h"


int main(int argc, char* argv[])
{
	// initialize XDSP with for use with given samplerate
	xdsp.Init(44100, (1.0 / (1 << 20)));
	
	if (SSESupported)
		printf("SSE detected\n");
	
	dword const bufsize = 256;
	
	// allocate float buffer of 256 samples
	// buffer is guaranteed to be 16-byte aligned
	float *buffer = xdsp.NewBuffer(bufsize);
	
	// clear buffer
	xdsp.Clear_V(buffer, bufsize);
	
	// try to clear buffer starting at second sample
	// buffer+1 is not 16-byte aligned so SSE version crashes with privileged instruction exception
	xdsp.Clear_V(buffer+1, bufsize-1);

	// you can use xdsp_noalign instead in this case
	// however unaligned versions of the functions are not as fast
	xdsp_noalign.Clear_V(buffer+1, bufsize-1);		// no crash
	
	// free buffer
	xdsp.DeleteBuffer(buffer);
	
	
	
	return 0;
}
