

/*

	XDSP example - biquad filtering
	Creates a lowpass filter and runs it twice on white noise input

*/


#include "stdafx.h"
#include "../../include/xdsp.h"


int main(int argc, char* argv[])
{
	// initialize XDSP with for use with given samplerate
	xdsp.Init(44100, (1.0 / (1 << 20)));
	
	if (SSESupported)
		printf("SSE detected\n");
	
	dword const bufsize = 256;
	
	// allocate float buffer of 256 samples
	// buffer is guaranteed to be 16-byte aligned
	float *buffer = xdsp.NewBuffer(bufsize);
	
	// generate white noise to the buffer
	for (dword i = 0; i < bufsize; i++)
		buffer[i] = (float)(rand() - (RAND_MAX / 2));
	

	// create a one channel biquad filter
	CXBiquadFilter *bqf = xdsp.BiquadCreate(1);

	// set coefficients 
	// xdsp biquad does out[i] = a0 * out[i-1] + a1 * out[i-2] + b0 * i[i] + b1 * i[i-1] + b2 * i[i-2]
	// so setting a0 to 0.7 and b0 to 0.3 gives you a simple onepole lowpass filter
	xdsp.BiquadSetCoeffs(bqf, 0, 0.7f, 0, 0.3f, 0, 0);

	// run filter
	xdsp.BiquadRun(bqf, buffer, bufsize);

	// reset biquad states (past inputs and outputs)
	xdsp.BiquadResetState(bqf, 0);

	// run filter again
	xdsp.BiquadRun(bqf, buffer, bufsize);

	// delete biquad filter
	delete bqf;
	
	// free buffer
	xdsp.DeleteBuffer(buffer);
	
	return 0;
}
