
/*

	XDSP example - resampler
	generates white noise at 100000Hz and resamples it to 44100Hz for output

*/


#include "stdafx.h"
#include "../../include/xdsp.h"
#include "stdafx.h"
#include "../WaveOut.h"

#define INTERNAL_RATE			100000

#define OUTPUT_BUFFER_SIZE		256

__declspec(align(32)) static float OutputBuffer[2*OUTPUT_BUFFER_SIZE];	// stereo
__declspec(align(32)) static float MonoBuffer[OUTPUT_BUFFER_SIZE];	



static CXPreparedResamplerFilter *pFilter = NULL;
static CXResampler *pResampler = NULL;


void __fastcall ResamplerCB(float *pout, dword const n, void *context);

int main(int argc, char* argv[])
{
	// initialize XDSP with for use with given samplerate
	xdsp.Init(44100, 1.0 / (1 << 20));
	
	// design a filter
	CXResamplerFilter *prf = xdsp.CreateKaiserSincFilter(INTERNAL_RATE, 44100, 32, 0.7f, 6);

	// prepare a 1 channel resampling filter
	pFilter = xdsp.PrepareResamplerFilter(prf, 1);
	delete prf;	// original filter is no longer needed

	pResampler = xdsp.ResamplerCreate(pFilter, ResamplerCB, NULL, OUTPUT_BUFFER_SIZE, NULL);

	WaveOutStart();

	getch();

	WaveOutStop();

	delete pResampler;
	delete pFilter;

	return 0;
}



void __fastcall ResamplerCB(float *pout, dword const n, void *context)
{
	// generate at INTERNAL_RATE here

	for (dword i = 0; i < n; i++)
		pout[i] = (float)(rand() - 16384) * 0.5f;
		

}



float *WaveOutCallback(int &numsamples)
{
	numsamples = OUTPUT_BUFFER_SIZE;

	xdsp.ResamplerRun(pResampler, MonoBuffer, numsamples);

	// mono -> stereo
	for (dword i = 0; i < (dword)numsamples; i++)
	{
		OutputBuffer[2*i+0] = MonoBuffer[i];
		OutputBuffer[2*i+1] = MonoBuffer[i];
	}

	return OutputBuffer;
}
