#pragma once

class CIOAudioBufferRef
{
public:
	friend class CIOAudioBufferIterator;
	
	CIOAudioBufferRef(float * const ps, dword const nums) 
		: Samples(ps), numSamples(nums) 
	{
		assert(Samples != NULL);
		ASSERT_XDSP_ALIGNMENT(Samples);
		assert(numSamples > 0);
	}
	
	float * const GetSamples() const { return Samples; }
	dword const GetnumSamples() const { return numSamples; }
	
	// xdsp interface

	void Clear_V() const { xdsp.Clear_V(Samples, numSamples); }
	void Add_V_V(float const *pin) const { xdsp.Add_V_V(Samples, pin, numSamples); }
	void Sub_V_V(float const *pin) const { xdsp.Sub_V_V(Samples, pin, numSamples); }
	void Copy_V_V(float const *pin) const { xdsp.Copy_V_V(Samples, pin, numSamples); }
	void Copy_V_VU(float const *pin) const { xdsp.Copy_V_VU(Samples, pin, numSamples); }
	void Copy_V_S(float const a) const { xdsp.Copy_V_S(Samples, numSamples, a); }
	void Copy_V_V4(float const *pin) const { xdsp.Copy_V_V4(Samples, pin, numSamples); }
	void Mul_V_V(float const *pin) const { xdsp.Mul_V_V(Samples, pin, numSamples); }
	void Mul_V_S(float const a) const { xdsp.Mul_V_S(Samples, numSamples, a); }
	void Mul_V_L(float &a, float const da) const { xdsp.Mul_V_L(Samples, numSamples, a, da); }
	void Mul_V_E(float &a, float const da) const { xdsp.Mul_V_E(Samples, numSamples, a, da); }
	void MulCopy_V_V_S(float const *pin, float const a) const { xdsp.MulCopy_V_V_S(Samples, pin, numSamples, a); }
	void MulCopy_V_V_L(float const *pin, float a, float const da) const { xdsp.MulCopy_V_V_L(Samples, pin, numSamples, a, da); }
	void MulAdd_V_V_S(float const *pin, float const a) const { xdsp.MulAdd_V_V_S(Samples, pin, numSamples, a); }
	void MulAdd_V_V_L(float const *pin, float a, float const da) const { xdsp.MulAdd_V_V_L(Samples, pin, numSamples, a, da); }
	void Div_V_S(float const a) const { xdsp.Div_V_S(Samples, numSamples, a); }
	void Div_V_V(float const *pin) const { xdsp.Div_V_V(Samples, pin, numSamples); }
	void Sqrt_V() const { xdsp.Sqrt_V(Samples, numSamples); }
	float AbsMax_V() const { xdsp.AbsMax_V(Samples, numSamples); }
	void Pow2Approx2_V_V(float const *pin) const { xdsp.Pow2Approx2_V_V(Samples, pin, numSamples); }
	void Min_V_S(float const x) const { xdsp.Min_V_S(Samples, numSamples, x); }
	void Max_V_S(float const x) const { xdsp.Max_V_S(Samples, numSamples, x); }
	void MinMax_V_S(float const _min, float const _max) const { xdsp.MinMax_V_S(Samples, numSamples, _min, _max); }
	void And_VI_SI(int const x) const { xdsp.And_VI_SI((int *)Samples, numSamples, x); }
	void Or_VI_SI(int const x) const { xdsp.Or_VI_SI((int *)Samples, numSamples, x); }
	void Xor_VI_SI(int const x) const { xdsp.Xor_VI_SI((int *)Samples, numSamples, x); }
	void And_VI_VI(int const *pin) const { xdsp.And_VI_VI((int *)Samples, pin, numSamples); }
	void AndNot1_VI_VI(int const *pin) const { xdsp.AndNot1_VI_VI((int *)Samples, pin, numSamples); }
	void AndNot2_VI_VI(int const *pin) const { xdsp.AndNot2_VI_VI((int *)Samples, pin, numSamples); }
	void Or_VI_VI(int const *pin) const { xdsp.Or_VI_VI((int *)Samples, pin, numSamples); }
	void Xor_VI_VI(int const *pin) const { xdsp.Xor_VI_VI((int *)Samples, pin, numSamples); }
	bool AbsMaxGT_V_S(float const x) const { return xdsp.AbsMaxGT_V_S(Samples, numSamples, x); }
	void Copy_V_VS(short const *pin) const { xdsp.Copy_V_VS(Samples, pin, numSamples); }
	void Copy_V_VI(int const *pin) const { xdsp.Copy_V_VI(Samples, pin, numSamples); }
	void Copy_V_VIP() const { xdsp.Copy_V_VIP(Samples, numSamples); }
	void EvaluatePolynomial3_V(float *pc, float const *pdc) const { xdsp.EvaluatePolynomial3_V(Samples, numSamples, pc, pdc); }
	void EvaluatePolynomial4_V(float *pc, float const *pdc) const { xdsp.EvaluatePolynomial4_V(Samples, numSamples, pc, pdc); }

	void BiquadRun(CXBiquadFilter *pf) const { xdsp.BiquadRun(pf, Samples, numSamples); }
	void CombRun(CXCombFilter *pc) const { xdsp.CombRun(pc, Samples, numSamples); }
	void CombOnepoleRun(CXCombFilter *pc) const { xdsp.CombOnepoleRun(pc, Samples, numSamples); }
	void AllpassRun(CXAllpassFilter *pc) const { xdsp.AllpassRun(pc, Samples, numSamples); }
	void ResamplerRun(CXResampler *pr) const { xdsp.ResamplerRun(pr, Samples, numSamples); }
	void MWOscillatorRun(CXMWOscillator *po, float const rate) const { xdsp.MWOscillatorRun(po, Samples, numSamples, rate); }
	void OscArrayRun(CXOscillatorArray *poa) const { xdsp.OscArrayRun(poa, Samples, numSamples); }
	
	// c++ operators

	void operator += (float const *pin) { Add_V_V(pin);	}
	void operator -= (float const *pin) { Sub_V_V(pin);	}
	void operator *= (float const *pin) { Mul_V_V(pin);	}
	void operator /= (float const *pin) { Div_V_V(pin);	}
	void operator &= (int const *pin) { And_VI_VI(pin); }
	void operator |= (int const *pin) { Or_VI_VI(pin); }
	void operator ^= (int const *pin) { Xor_VI_VI(pin); }
	

private:
	float * const Samples;
	dword const numSamples;
};


class CIOAudioBufferIterator
{
public:
	CIOAudioBufferIterator(CIOAudioBufferRef const &ab) 
		: Samples(ab.Samples), numSamples(ab.numSamples), Pos(0) 
	{ 
	}
	
	bool const Next() { ++Pos; return Pos < numSamples; }
	
	void Set(float const &s) const { Samples[Pos] = s; }
	float const &Get() const { return Samples[Pos]; }
	
private:
	float * const Samples;
	dword const numSamples;
	dword Pos;
	
};



class CStereoIOAudioBufferRef
{
public:
	friend class CStereoIOAudioBufferIterator;

	CStereoIOAudioBufferRef(float * const ps, dword const nums) 
		: Samples(ps), numSamples(nums) 
	{
		assert(Samples != NULL);
		ASSERT_XDSP_ALIGNMENT(Samples);
		assert(numSamples > 0);
	}
	
	
	float * const GetSamples() const { return Samples; }
	dword const GetnumSamples() const { return numSamples; }


private:
	float * const Samples;
	dword const numSamples;
};


class CStereoIOAudioBufferIterator
{
public:
	CStereoIOAudioBufferIterator(CStereoIOAudioBufferRef const &ab) 
		: Samples(ab.Samples), numSamples(ab.numSamples), Pos(0) 
	{ 
	}

	bool const Next() { ++Pos; return Pos < numSamples; }
	
	
	void SetL(float const &s) const { Samples[2*Pos] = s; }
	void SetR(float const &s) const { Samples[2*Pos+1] = s; }
	
	float const &GetL() const { return Samples[2*Pos]; }
	float const &GetR() const { return Samples[2*Pos+1]; }
	
private:
	float * const Samples;
	dword const numSamples;
	dword Pos;
	
};


