

#ifndef __XDSP_AUDIO_BUFFER_POOL_H
#define __XDSP_AUDIO_BUFFER_POOL_H

#include "XDSPInterface.h"


class CXAudioBufferPoolImpl;

class CXAudioBufferPool : public CXObject
{
public:
	class CBuffer
	{
	public:
		operator float * () { return Data; }
		operator int * () { return (int *)Data; }
		
	public:
		float *Data;
		CBuffer *Next;
	};
	
	CXAudioBufferPool(dword const buffersize, dword const numbuffers);
	virtual ~CXAudioBufferPool();
	virtual dword QueryMemoryUsage() const;
	
	__forceinline CBuffer *AllocateBuffer()
	{
		if (FreeBuffers == NULL)
		{
			Grow();

			// if FreeBuffers == NULL after Grow() the internal limit has been reached
			// the most probable cause is a memory leak in user code

			assert(FreeBuffers != NULL);	

			if (FreeBuffers == NULL)
				return NULL;
		}
		
		CBuffer *t = FreeBuffers;
		FreeBuffers = FreeBuffers->Next;
		return t;
	}
	
	__forceinline void FreeBuffer(CBuffer *t)
	{
		assert(t != NULL);	
		
		t->Next = FreeBuffers;
		FreeBuffers = t;
	}

protected:
	void Grow();

protected:
	CBuffer *FreeBuffers;
	CXAudioBufferPoolImpl *pImpl;
		
};

class CXTempAudioBuffer
{
public:
	__forceinline CXTempAudioBuffer(CXAudioBufferPool &p, bool const alloc = true)
		: abp(p), pb(alloc ? abp.AllocateBuffer() : NULL)
	{
	}

	__forceinline ~CXTempAudioBuffer()
	{
		if (pb != NULL)
			abp.FreeBuffer(pb);
	}

	__forceinline operator float * () const { return pb->Data; }
	__forceinline operator int * () const { return (int *)pb->Data; }

	__forceinline void Acquire(CXTempAudioBuffer &a)
	{
		assert(pb == NULL);
		pb = a.pb;
		a.pb = NULL;
	}

private:
	CXAudioBufferPool::CBuffer *pb;
	CXAudioBufferPool &abp;
};



#endif
