
// Copyright 2002-2003 Oskari Tammelin (ot@iki.fi)

#ifndef __XDSP_INTERFACE_H
#define __XDSP_INTERFACE_H

#include "jxstream.h"

#define CC __fastcall

class CXObject
{
public:
	CXObject() {}
	virtual ~CXObject() {}

	virtual dword QueryMemoryUsage() const { return 0; }

private:	// make all CXObjects non-copyable
	CXObject (CXObject const &a);
	CXObject const & operator = (CXObject const &a);

public:
	enum type 
	{ 
		none, BiquadFilter, CombFilter, AllpassFilter, 
		ResamplerFilter, PreparedResamplerFilter, Resampler,
		MipWave, MWOscillator, OscillatorArray			
	};
	
};
 
class CXSerializable 
{ 
public:
	virtual bool Serialize(CStreamOut &so) = 0; 
	static CXObject *DeserializeObject(CStreamIn &si);
	
};

class CXBiquadFilter : public CXObject {};
class CXCombFilter : public CXObject {};
class CXAllpassFilter : public CXObject {};
class CXResamplerFilter : public CXObject, public CXSerializable {};
class CXPreparedResamplerFilter : public CXObject {};
class CXResampler : public CXObject {};
class CXMipWave : public CXObject {};
class CXMWOscillator : public CXObject {};
class CXOscillatorArray : public CXObject
{
public:
	float *Amplitudes;
	float *AmplitudeDeltas;
};

typedef void (CC *CXResamplerCallback)(float *pout, dword const n, void *context);

enum XLCAlgo { simple };

class CXDSPInterface
{
public:
	// initialization
	void (CC *Init)(dword const samplerate, float const silencethreshold = (1.0f / (1 << 30)));

	// buffer handling
	float *(CC *NewBuffer)(dword const numsamples);
	void (CC *DeleteBuffer)(float *pbuf);

	int *(CC *NewIntBuffer)(dword const numsamples);
	void (CC *DeleteIntBuffer)(int *pbuf);

	short *(CC *NewShortBuffer)(dword const numsamples);
	void (CC *DeleteShortBuffer)(short *pbuf);
	
	// basic operations
	void (CC *Clear_V)(float *p, dword const n);					

	void (CC *Add_V_V)(float *pout, float const *pin, dword const n);
	void (CC *Add_VI_VI)(int *pout, int const *pin, dword const n);
	void (CC *Add_VS_VS)(short *pout, short const *pin, dword const n);
	void (CC *Add_V2N_V2)(float *pout0, float *pout1, float const *pin, dword const n);
	void (CC *Add_V_S)(float *p, dword const n, float const a);
	void (CC *Add_V_L)(float *p, dword const n, float &a, float const da);
	void (CC *Add_VI_SI)(int *pout, dword const n, int const a);
	
	void (CC *Sub_V_V)(float *pout, float const *pin, dword const n);
	void (CC *Sub_V_V_V)(float *pout, float const *pina, float const *pinb, dword const n);
	void (CC *Sub_VI_VI)(int *pout, int const *pin, dword const n);
	void (CC *Sub_VS_VS)(short *pout, short const *pin, dword const n);
	
	void (CC *Copy_V_V)(float *pout, float const *pin, dword const n);
	void (CC *Copy_V_VU)(float *pout, float const *pin, dword const n);
	void (CC *Copy_V_S)(float *p, dword const n, float const a);
	void (CC *Copy_V_L)(float *p, dword const n, float &a, float const da);
	void (CC *Copy_V_E)(float *p, dword const n, float &a, float const da);
	void (CC *Copy_V2_V2N)(float *pout, float const *pin0, float const *pin1, dword const n);
	void (CC *Copy_V2N_V2)(float *pout0, float *pout1, float const *pin, dword const n);
	void (CC *Copy_V4_V4N)(float *pout, float const *pin0, float const *pin1, float const *pin2, float const *pin3, dword const n);
	void (CC *Copy_V4N_V4)(float *pout0, float *pout1, float *pout2, float *pout3, float const *pin, dword const n);
	void (CC *Copy_V_V4)(float *pout, float const *pin, dword const n);
	void (CC *Copy_V2_V4)(float *pout, float const *pin, dword const n);

	void (CC *Mul_V_V)(float *pout, float const *pin, dword const n);
	void (CC *Mul_V_S)(float *p, dword const n, float const a);
	void (CC *Mul_V2_S2)(float *p, dword const n, float const la, float const ra);
	void (CC *Mul_V4_S4)(float *p, dword const n, float const a0, float const a1, float const a2, float const a3);
	void (CC *Mul_V_L)(float *po, dword const n, float &a, float const da);
	void (CC *Mul_V_E)(float *po, dword const n, float &a, float const da);
	void (CC *MulSar_VI_SI)(int *p, dword const n, int const a, dword const shift);

	void (CC *MulCopy_V_V_S)(float *pout, float const *pin, dword const n, float const a);
	void (CC *MulCopy_V2_V2_S2)(float *pout, float const *pin, dword const n, float const la, float const ra);
	void (CC *MulCopy_V2_V_S2)(float *pout, float const *pin, dword const n, float const la, float const ra);
	void (CC *MulCopy_V_V_L)(float *pout, float const *pin, dword const n, float a, float const da);
	void (CC *MulCopy_V2_V2_L2)(float *pout, float const *pin, dword const n, float la, float ra, float const dla, float const dra);

	void (CC *MulAdd_V_V_S)(float *pout, float const *pin, dword const n, float const a);
	void (CC *MulAdd_V2_V2_S2)(float *pout, float const *pin, dword const n, float const la, float const ra);
	void (CC *MulAdd_V2_V_S2)(float *pout, float const *pin, dword const n, float const la, float const ra);
	void (CC *MulAdd_V2_V_L2)(float *pout, float const *pin, dword const n, float la, float ra, float const dla, float const dra);
	void (CC *MulAdd_V_V_L)(float *pout, float const *pin, dword const n, float a, float const da);
	void (CC *MulAdd_V2_V2_L2)(float *pout, float const *pin, dword const n, float la, float ra, float const dla, float const dra);
	
	void (CC *Div_V_S)(float *p, dword const n, float const a);
	void (CC *Div_V_V)(float *pout, float const *pin, dword const n);
	void (CC *ReciprocalApprox_V_V)(float *pout, float const *pin, dword const n);
	
	void (CC *Sqrt_V)(float *p, dword const n);
	void (CC *ReciprocalSqrtApprox_V_V)(float *pout, float const *pin, dword const n);
	
	void (CC *Abs_VI)(int *p, dword const n);
	void (CC *Abs_VS)(short *p, dword const n);
	
	float (CC *AbsMax_V)(float const *pin, dword const n);
	int (CC *AbsMax_VI)(int const *pin, dword const n);
	short (CC *AbsMax_VS)(short const *pin, dword const n);
	
	void (CC *DenormalToZero_V)(float *p, dword const n);

	// exponential/logaritmic operations
	void (CC *ExpApprox_V_V)(float *pout, float const *pin, dword const n);
	void (CC *Pow2Approx_V_V)(float *pout, float const *pin, dword const n);
	void (CC *Pow2Approx2_V_V)(float *pout, float const *pin, dword const n);
	void (CC *LogApprox_V_V)(float *pout, float const *pin, dword const n);
	void (CC *Log2Approx_V_V)(float *pout, float const *pin, dword const n);
	
	// trigonometric functions
	void (CC *SinCos_V2N_V)(float *psin, float *pcos, float const *pin, dword const n);
	void (CC *SinApprox_V_V)(float *pout, float const *pin, dword const n);
	void (CC *CosApprox_V_V)(float *pout, float const *pin, dword const n);
	void (CC *TanApprox_V_V)(float *pout, float const *pin, dword const n);
	void (CC *AtanApprox_V_V)(float *pout, float const *pin, dword const n);
	
	// hyperbolic functions
	void (CC *SinhApprox_V_V)(float *pout, float const *pin, dword const n);
	void (CC *CoshApprox_V_V)(float *pout, float const *pin, dword const n);
	void (CC *TanhApprox_V_V)(float *pout, float const *pin, dword const n);

	// complex number operations
	void (CC *Abs_V_VCN)(float *pout, float const *pr, float const *pi, dword const n);

	// minimum/maximum
	void (CC *Min_V_S)(float *p, dword const n, float const x);
	void (CC *Max_V_S)(float *p, dword const n, float const x);
	void (CC *MinMax_V_S)(float *p, dword const n, float const _min, float const _max);
	void (CC *Min_V_V)(float *pout, float const *pin, dword const n);
	void (CC *Max_V_V)(float *pout, float const *pin, dword const n);
	
	// shift operations
	void (CC *Shl_VI)(int *p, dword const n, dword const shift);
	void (CC *Shr_VI)(int *p, dword const n, dword const shift);
	void (CC *Sar_VI)(int *p, dword const n, dword const shift);
	
	// logical operations
	void (CC *And_VI_SI)(int *p, dword const n, int const x);
	void (CC *Or_VI_SI)(int *p, dword const n, int const x);
	void (CC *Xor_VI_SI)(int *p, dword const n, int const x);
	void (CC *And_VI_VI)(int *p, int const *pin, dword const n);
	void (CC *AndNot1_VI_VI)(int *p, int const *pin, dword const n);
	void (CC *AndNot2_VI_VI)(int *p, int const *pin, dword const n);
	void (CC *Or_VI_VI)(int *p, int const *pin, dword const n);
	void (CC *Xor_VI_VI)(int *p, int const *pin, dword const n);
	
	// comparison 
	void (CC *CompareLT_V_V_S)(dword *pout, float const *pin1, dword const n, float const a);
	void (CC *CompareLT_V_V_V)(dword *pout, float const *pin1, float const *pin2, dword const n);
	bool (CC *AbsMaxGT_V_S)(float const *pin, dword const n, float const x);

	// conditional operations
	void (CC *CompareLTAdd_V_S_S)(float *p, dword const n, float const a, float const b);
	void (CC *CompareLEAdd_V_S_S)(float *p, dword const n, float const a, float const b);
	void (CC *CompareGTAdd_V_S_S)(float *p, dword const n, float const a, float const b);
	void (CC *CompareGEAdd_V_S_S)(float *p, dword const n, float const a, float const b);
	void (CC *AbsCompareLTCopy_V_S_S)(float *p, dword const n, float const a, float const b);
	
	// conversion
	void (CC *Copy_V_VS)(float *pout, short const *pin, dword const n);
	void (CC *Copy_V_VI)(float *pout, int const *pin, dword const n);
	void (CC *Copy_V_VIP)(float *p, dword const n);
	void (CC *Copy_VS_V)(short *pout, float const *pin, dword const n);
	void (CC *Copy_VI_V)(int *pout, float const *pin, dword const n);
	void (CC *Copy_VI_VP)(int *p, dword const n);
	void (CC *CopyShift_VI_VP)(int *p, dword const n, dword const shift);

	// conversion with clipping to destination range
	void (CC *ClipCopy_VS_V)(short *pout, float const *pin, dword const n);
	void (CC *ClipCopy_VI_V)(int *pout, float const *pin, dword const n);
	void (CC *ClipCopy_VI_VP)(int *p, dword const n);

	// pseudorandom number generators
	void (CC *Rand_VI)(int *pout, dword const n);
	void (CC *RandExp_V)(float *pout, dword const n);
	void (CC *RandExp2_V)(float *pout, dword const n);
	
	// polynomial evaluation
	void (CC *EvaluatePolynomial3_V)(float *p, dword const n, float *pc, float const *pdc);
	void (CC *EvaluatePolynomial4_V)(float *p, dword const n, float *pc, float const *pdc);
	
	// biquad filter
	CXBiquadFilter *(CC *BiquadCreate)(dword const numchannels);
	void (CC *BiquadResetState)(CXBiquadFilter *pf, dword const channel);
	void (CC *BiquadSetCoeffs)(CXBiquadFilter *pf, dword const channel, float const a0, float const a1, float const b0, float const b1, float const b2);
	void (CC *BiquadRun)(CXBiquadFilter *pf, float *p, dword const n);
	void (CC *BiquadRun2x2)(CXBiquadFilter *pf, float *p, dword const n);
	
	// comb filter
	CXCombFilter *(CC *CombCreate)(dword const delay);
	void (CC *CombSetState)(CXCombFilter *pc, float const *state, float const onepolestate = 0);
	void (CC *CombSetFeedback)(CXCombFilter *pc, float const fb);
	void (CC *CombSetFilterCoeffs)(CXCombFilter *pc, float const pole, float const gain);
	void (CC *CombRun)(CXCombFilter *pc, float *p, dword const n);
	void (CC *CombOnepoleRun)(CXCombFilter *pc, float *p, dword const n);
	
	// allpass filter
	CXAllpassFilter *(CC *AllpassCreate)(dword const delay);
	void (CC *AllpassSetState)(CXAllpassFilter *pc, float const *state = NULL);
	void (CC *AllpassSetFeedback)(CXAllpassFilter *pc, float const fb);
	void (CC *AllpassRun)(CXAllpassFilter *pc, float *p, dword const n);
	
	// resampler
	CXResamplerFilter *(CC *CreateKaiserSincFilter)(dword inrate, dword outrate, dword fsize, float cutoff, float beta);
	CXPreparedResamplerFilter *(CC *PrepareResamplerFilter)(CXResamplerFilter *prf, dword numch);
	CXResampler *(CC *ResamplerCreate)(CXPreparedResamplerFilter *pf, CXResamplerCallback pcb, void *cbcontext, dword const maxbuffersize, float *tempbuffer);
	void (CC *ResamplerRun)(CXResampler *pr, float *pout, dword const n);

	// wavetable oscillator
	CXMipWave *(CC *CreateMipWave)(short const *samples, dword const numsamples, dword const size, float const freq_limit, float const alias_limit, float const alias_level_limit);
	CXMWOscillator *(CC *MWOscillatorCreate)();
	void (CC *MWOscillatorSetMW)(CXMWOscillator *po, CXMipWave const *pmw);
	void (CC *MWOscillatorRun)(CXMWOscillator *po, float *pout, dword const n, float const rate);
		
	// oscillator array
	CXOscillatorArray *(CC *OscArrayCreate)(dword const numoscs, float *tempbuffer);
	void (CC *OscArraySetFrequencies)(CXOscillatorArray *poa, float const *freqs);
	void (CC *OscArraySetPhases)(CXOscillatorArray *poa, float const *freqs);
	void (CC *OscArrayRun)(CXOscillatorArray *poa, float *pout, dword const n);
		
	// lossless compression
	dword (CC *Compress_VS)(byte *pout, short const *pin, dword const n, XLCAlgo const algo);
	dword (CC *Compress_VI)(byte *pout, int const *pin, dword const n, XLCAlgo const algo);
	bool (CC *Decompress_VS)(short *pout, byte const *pin, dword const n, dword const nbits, XLCAlgo const algo);
	bool (CC *Decompress_VI)(int *pout, byte const *pin, dword const n, dword const nbits, XLCAlgo const algo);
	
};


#undef CC

#endif