
#include "stdafx.h"
#include "AudioBufferPool.h"
#include <xmmintrin.h>
#include <list>
#include <memory>

class CBufferVector
{
public:
	CBufferVector(dword const buffersize, dword const numbuffers)
	{
		dword const actualbuffersize = (buffersize + 7) & ~7;
		dword const totalsize = actualbuffersize * numbuffers;
		
		Data = (float *)_aligned_malloc(totalsize * sizeof(float), 64);
		Buffers = new CXAudioBufferPool::CBuffer[numbuffers];
		
		for (dword i = 0; i < numbuffers; i++)
		{
			Buffers[i].Data = Data + i * actualbuffersize;
			Buffers[i].Next = (i < numbuffers-1) ? Buffers + i + 1 : NULL;
			
		}
		
	}

	~CBufferVector()
	{
		delete[] Buffers;
		_aligned_free(Data);
	}
	

public:
	CXAudioBufferPool::CBuffer *Buffers;
	float *Data;
	
};

typedef std::list<std::auto_ptr<CBufferVector> > CBufferVectorList;

class CXAudioBufferPoolImpl
{
public:
	CXAudioBufferPoolImpl(dword const buffersize, dword const numbuffers)
	{
		BufferSize = buffersize;
		numBuffers = numbuffers;
	}

	CXAudioBufferPool::CBuffer *NewBufferVector()
	{
		BufferVectors.push_back(std::auto_ptr<CBufferVector>(new CBufferVector(BufferSize, numBuffers)));
		return BufferVectors.back().get()->Buffers;
		
	}

public:
	dword numBuffers;
	dword BufferSize;
	CBufferVectorList BufferVectors;

};



CXAudioBufferPool::CXAudioBufferPool(dword const buffersize, dword const numbuffers)
{
	pImpl = new CXAudioBufferPoolImpl(buffersize, numbuffers);
	FreeBuffers = pImpl->NewBufferVector();
}


CXAudioBufferPool::~CXAudioBufferPool()
{
	delete pImpl;
}

dword CXAudioBufferPool::QueryMemoryUsage() const
{
//	return numBuffers * ((BufferSize + 7) & ~7);

	return 0;
}

void CXAudioBufferPool::Grow()
{
	assert(FreeBuffers == NULL);
	FreeBuffers = pImpl->NewBufferVector();
}