// AGInstructionBuffer.h: interface for the CAGInstructionBuffer class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_AGINSTRUCTIONBUFFER_H__B567A3CA_8287_46E7_BFD7_2D5C10860200__INCLUDED_)
#define AFX_AGINSTRUCTIONBUFFER_H__B567A3CA_8287_46E7_BFD7_2D5C10860200__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000




#include "AudioGraphPort.h"
#include "../../include/AudioGraph.h"

class CAGInstructionBuffer  
{
public:
	CAGInstructionBuffer();
	virtual ~CAGInstructionBuffer();

	
	void AddAllocBuffer(int const ibuf)
	{
		BeginInstruction(alloc_buffer);
		Data.push_back(ibuf);
	}

	void AddFreeBuffer(int const ibuf)
	{
		BeginInstruction(free_buffer);
		Data.push_back(ibuf);
	}

	void AddAllocCopyBuffer(int const ibuf, int const source)
	{
		BeginInstruction(alloc_copy_buffer);
		Data.push_back(ibuf);
		Data.push_back(source);
	}

	void AddAddBuffer(int const dest, int const source)
	{
		BeginInstruction(add_buffer);
		Data.push_back(dest);
		Data.push_back(source);
	}

	void AddCallback(CXAudioGraphCallback *pcb, AUDIO_GRAPH_CALLBACK_FUNC const pfunc, IntVector const &inputs, IntVector const &outputs)
	{
		BeginInstruction(callback);
		
		assert(sizeof(AUDIO_GRAPH_CALLBACK_FUNC) == 8);
		Data.push_back((int)pcb);
		Data.push_back(*(int *)&pfunc);
		Data.push_back(*((int *)&pfunc+1));
		
		Data.push_back(inputs.size());
		for (int i = 0; i < inputs.size(); i++)
			Data.push_back(inputs[i]);

		Data.push_back(outputs.size());
		for (i = 0; i < outputs.size(); i++)
			Data.push_back(outputs[i]);
		
	}

	void AddEnd()
	{
		BeginInstruction(end);
	}

	void Clear() { Data.clear(); }
	int GetSize() const { return Data.size(); }

	int const *GetPtr() const { return &Data[0]; }

	enum instruction { none=0, alloc_buffer, free_buffer, alloc_copy_buffer, add_buffer, callback, end };

private:

	void BeginInstruction(instruction i) { Data.push_back((int)i); }
	
private:
	IntVector Data;


};

#endif // !defined(AFX_AGINSTRUCTIONBUFFER_H__B567A3CA_8287_46E7_BFD7_2D5C10860200__INCLUDED_)
