

#ifndef __AUDIO_GRAPH_IMPL_H
#define __AUDIO_GRAPH_IMPL_H


#include "AudioGraphNodeImpl.h"
#include "AGInstructionBuffer.h"

#pragma pointers_to_members(full_generality, multiple_inheritance)


typedef std::list<std::auto_ptr<CXAudioGraphNode> > CAudioGraphNodeList;
typedef std::list<CXAudioGraphNode *> CAudioGraphNodePtrList;

class CAudioGraphBufferRef
{
public:
	CAudioGraphBufferRef()
	{
		Allocated = false;
		pBuffer = NULL;
	}

	CAudioGraphBufferRef(CXAudioBufferPool::CBuffer *p)
	{
		Allocated = true;
		pBuffer = p;
	}

public:
	CXAudioBufferPool::CBuffer *pBuffer;
	bool Allocated;
};

typedef std::vector<CAudioGraphBufferRef> CAudioBufferRefVector;

class CAGCompiler
{
public:
	CAGCompiler()
	{
		MaxInputs = MaxOutputs = 0;
	}

	int AllocBuffer()
	{
		int index = -1;

		for (int i = 0; i < Buffers.size(); i++)
		{
			if (!Buffers[i])
			{
				Buffers[i] = 1;
				index = i;
				break;
			}
		}
		
		Buffers.push_back(1);
		index = Buffers.size() - 1;

		InstrBuffer->AddAllocBuffer(index);

		return index;
	}

	void FreeBuffer(int const i)
	{
		assert(i >= 0);
		assert(i < Buffers.size());
		assert(Buffers[i]);
		
		Buffers[i] = 0;
		InstrBuffer->AddFreeBuffer(i);
	
	}

	int AllocCopyBuffer(int const source)
	{
		int index = -1;
		
		for (int i = 0; i < Buffers.size(); i++)
		{
			if (!Buffers[i])
			{
				Buffers[i] = 1;
				index = i;
				break;
			}
		}
		
		Buffers.push_back(1);
		index = Buffers.size() - 1;
		
		InstrBuffer->AddAllocCopyBuffer(index, source);
		
		return index;
	}
	
public:
	CAGInstructionBuffer *InstrBuffer;
	IntVector Buffers;
	int MaxInputs;
	int MaxOutputs;
	
};

class CXAudioGraphImpl  
{
public:
	CXAudioGraphImpl();
	virtual ~CXAudioGraphImpl();

	void SetAudioBufferPool(CXAudioBufferPool *pabp);

	bool AddNode(CXAudioGraphNode *pnode, bool const output_node = false);
	bool RemoveNode(CXAudioGraphNode *pnode);

	void Compile();

	bool IsCyclic();


	void Run(dword const nsamples);

private:
	CAudioGraphNodeList Nodes;
	CAudioGraphNodePtrList OutputNodes;
	CXAudioBufferPool *pABP;

	CAudioBufferRefVector Buffers;

	CJxCriticalSection CS;

	CAGInstructionBuffer *pInstructions;
	float **InputPtrs;
	float **OutputPtrs;

public:
	CAGCompiler *pCompiler;

};

#endif 
