// AudioGraphNode.cpp: implementation of the CAudioGraphNode class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "AudioGraphImpl.h"
#include "../../include/AudioGraphNode.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CXAudioGraphNodeImpl::CXAudioGraphNodeImpl()
{
	pGraph = NULL;
}

CXAudioGraphNodeImpl::~CXAudioGraphNodeImpl()
{
	for (CAudioGraphPortPtrVector::iterator i = InputPorts.begin(); i != InputPorts.end(); i++)
		delete *i;

	for (i = OutputPorts.begin(); i != OutputPorts.end(); i++)
		delete *i;
}


bool CXAudioGraphNodeImpl::AddPort(CAudioGraphPortPtrVector &v, dword const i)
{
	if (i > 255)
		return false;
	
	if (i < v.size() && v[i] != NULL)
		return false;
		
	if (i >= v.size())
		v.resize(i+1);

	v[i] = new CAudioGraphPort(this);
	
	return true;
	
}

bool CXAudioGraphNodeImpl::RemovePort(CAudioGraphPortPtrVector &v, dword const i)
{
	if (i > 255)
		return false;
	
	if (i >= v.size() || v[i] == NULL)
		return false;
	
	delete v[i];
	v[i] = NULL;

	return true;
}

bool CXAudioGraphNodeImpl::AddInputPort(dword const i)
{
	return AddPort(InputPorts, i);
}

bool CXAudioGraphNodeImpl::RemoveInputPort(dword const i)
{
	return RemovePort(InputPorts, i);
}

bool CXAudioGraphNodeImpl::AddOutputPort(dword const i)
{
	return AddPort(OutputPorts, i);
}

bool CXAudioGraphNodeImpl::RemoveOutputPort(dword const i)
{
	return RemovePort(OutputPorts, i);
}


bool Connect(CXAudioGraphNode *_psrc, dword const srcport, CXAudioGraphNode *_pdst, dword const dstport)
{
	CXAudioGraphNodeImpl *psrc = _psrc->pImpl;
	CXAudioGraphNodeImpl *pdst = _pdst->pImpl;
	
	if (psrc == NULL || pdst == NULL)
		return false;

	if (psrc->GetGraph() != pdst->GetGraph())
		return false;
	
	CAudioGraphPort *psrcport = psrc->GetOutputPort(srcport);
	CAudioGraphPort *pdstport = pdst->GetInputPort(dstport);
	
	if (psrcport == NULL || pdstport == NULL)
		return false;

	if (psrcport->HasConnection(pdstport) || pdstport->HasConnection(psrcport))
		return false;

	psrcport->Connections.push_back(pdstport);
	pdstport->Connections.push_back(psrcport);
	
	if (psrc->GetGraph()->IsCyclic())
	{
		Disconnect(_psrc, srcport, _pdst, dstport);
		return false;
	}

	return true;
}

bool Disconnect(CXAudioGraphNode *_psrc, dword const srcport, CXAudioGraphNode *_pdst, dword const dstport)
{
	CXAudioGraphNodeImpl *psrc = _psrc->pImpl;
	CXAudioGraphNodeImpl *pdst = _pdst->pImpl;

	if (psrc == NULL || pdst == NULL)
		return false;

	if (psrc->GetGraph() != pdst->GetGraph())
		return false;

	CAudioGraphPort *psrcport = psrc->GetOutputPort(srcport);
	CAudioGraphPort *pdstport = pdst->GetInputPort(dstport);
	
	if (psrcport == NULL || pdstport == NULL)
		return false;
	
	if (!psrcport->HasConnection(pdstport) || !pdstport->HasConnection(psrcport))
		return false;
	
	psrcport->RemoveConnection(pdstport);
	pdstport->RemoveConnection(psrcport);

	return true;
}

bool CXAudioGraphNodeImpl::CycleTest()
{
	if (CycleTestFlag)
		return true;

	CycleTestFlag = true;
	
	for (CAudioGraphPortPtrVector::iterator i = InputPorts.begin(); i != InputPorts.end(); i++)
	{
		if (*i != NULL)
		{
			for (CAudioGraphPortPtrList::iterator j = (*i)->Connections.begin(); j != (*i)->Connections.end(); j++)
			{
				if ((*j)->pNode->CycleTest())
					return true;
			}
				
		}
	}
	
	CycleTestFlag = false;

	return false;
}


void CXAudioGraphNodeImpl::Compile()
{
	IntVector ibufs;
	IntVector buffers_to_free;
	
	for (CAudioGraphPortPtrVector::iterator i = InputPorts.begin(); i != InputPorts.end(); i++)
	{
		IntVector portibufs;

		int mixbuffer = -1;

		if (*i != NULL)
		{
			for (CAudioGraphPortPtrList::iterator j = (*i)->Connections.begin(); j != (*i)->Connections.end(); j++)
			{
				CAudioGraphPort *pp = *j;

				if (!(*j)->pNode->Processed)
					(*j)->pNode->Compile();
			
				if (mixbuffer == -1 && (*j)->OutputCount == 1)
					mixbuffer = (*j)->BufferIndex;
				else
					portibufs.push_back((*j)->BufferIndex);
			}
			
		}

		if (portibufs.size() == 0 && mixbuffer == -1)
		{
			ibufs.push_back(-1);
			continue;
		}

		if (portibufs.size() == 0)
		{
			ibufs.push_back(mixbuffer);
			continue;
		}

		printf("mix buffers\n");
		
		for (int j = 0; j < portibufs.size(); j++)
		{
			if (portibufs[j] != mixbuffer)
			{
				if (mixbuffer == -1)
				{
					mixbuffer = pGraph->pCompiler->AllocCopyBuffer(portibufs[j]);
					buffers_to_free.push_back(mixbuffer);
					printf("  alloc/copy %d to %d\n", portibufs[j], mixbuffer);
					
				}
				else
				{
					printf("  add %d to %d\n", portibufs[j], mixbuffer);
					pGraph->pCompiler->InstrBuffer->AddAddBuffer(mixbuffer, portibufs[j]);
				}

			}
			
		}

		printf("\n");

		ibufs.push_back(mixbuffer);
			

	}

	printf("start process node %x\n", this);

	IntVector obufs;

	{
		int oindex = 0;

		for (CAudioGraphPortPtrVector::iterator i = OutputPorts.begin(); i != OutputPorts.end(); i++)
		{
			if (*i != NULL)
			{
				if ((*i)->Connections.size() > 0)
				{
					int const bufindex = pGraph->pCompiler->AllocBuffer();
					(*i)->BufferIndex = bufindex;
					(*i)->OutputCount = (*i)->Connections.size();
					obufs.push_back(bufindex);

					printf("  alloc buffer %d for output port %d (connection count %d)\n", bufindex, oindex, (*i)->OutputCount);
					
				}
				else
				{
					printf("  no connections for output port %d\n", oindex);
					obufs.push_back(-1);
				}
				
				oindex++;
			}
			
		}	
	}
	

	printf("  callback in [ ");
	
	{
		for (int i = 0; i < ibufs.size(); i++)
		{
			printf("%d ", ibufs[i]);
		}
	}

	printf("]  out [ ");

	{
		for (int i = 0; i < obufs.size(); i++)
		{
			printf("%d ", obufs[i]);
		}
	}
	
	printf("]\n");

	if (ibufs.size() > pGraph->pCompiler->MaxInputs)
		pGraph->pCompiler->MaxInputs = ibufs.size();

	if (obufs.size() > pGraph->pCompiler->MaxOutputs)
		pGraph->pCompiler->MaxOutputs = obufs.size();
	
	pGraph->pCompiler->InstrBuffer->AddCallback(pCB, pCBFunc, ibufs, obufs);
	

	
	{
		int iindex = 0;

		for (CAudioGraphPortPtrVector::iterator i = InputPorts.begin(); i != InputPorts.end(); i++)
		{
			if (*i != NULL)
			{
				for (CAudioGraphPortPtrList::iterator j = (*i)->Connections.begin(); j != (*i)->Connections.end(); j++)
				{
					(*j)->OutputCount--;
					if ((*j)->OutputCount == 0)
					{
						pGraph->pCompiler->FreeBuffer((*j)->BufferIndex);			
						printf("  free buffer %d\n", (*j)->BufferIndex);
						
					}

				}
				
			}

			iindex++;
		}
		
	}
	
	{
		for (int i = 0; i < buffers_to_free.size(); i++)
		{
			pGraph->pCompiler->FreeBuffer(buffers_to_free[i]);
			printf("  free buffer %d\n", buffers_to_free[i]);
			
		}
	}
	
	printf("end process\n\n");
	
	Processed = true;

}


// ---------------------------- interface -------------------------------

CXAudioGraphNode::CXAudioGraphNode()
{
	pImpl = new CXAudioGraphNodeImpl;
}

CXAudioGraphNode::~CXAudioGraphNode()
{
	delete pImpl;
}

bool CXAudioGraphNode::AddInputPort(dword const i) { return pImpl->AddInputPort(i); }
bool CXAudioGraphNode::RemoveInputPort(dword const i) { return pImpl->RemoveInputPort(i); }
bool CXAudioGraphNode::AddOutputPort(dword const i) { return pImpl->AddOutputPort(i); }
bool CXAudioGraphNode::RemoveOutputPort(dword const i) { return pImpl->RemoveOutputPort(i); }
void CXAudioGraphNode::SetCallback(CXAudioGraphCallback *pcb, AUDIO_GRAPH_CALLBACK_FUNC pfunc)
{
	pImpl->SetCallback(pcb, pfunc);
}


