
#include "stdafx.h"
#include "XDSPInterface.h"
#include "dspinternal.h"

void FPUClear_V(float *p, dword const n);
void FPUCopy_V_V(float *pout, float const *pin, dword const n);



// allpass

class CAllpassFPU : public CXAllpassFilter
{
public:
	CAllpassFPU(dword const delay)
	{
		BufferPos = Buffer = NULL;
		BufferLength = 0;
		Feedback = 0.5;
		SetDelay(delay);
	}
	
	virtual ~CAllpassFPU()
	{
		delete[] Buffer;
	}
	
	virtual dword QueryMemoryUsage() const { return sizeof(this) + BufferLength * sizeof(float); }
	
	
	void SetDelay(dword d)
	{
		BufferLength = d;
		delete[] Buffer;
		BufferPos = Buffer = new float [d];
		FPUClear_V(Buffer, d);
		
	}
	
	void SetState(float const *state)
	{
		if (state == NULL)
			FPUClear_V(Buffer, BufferLength);
		else
			FPUCopy_V_V(Buffer, state, BufferLength);
	}
	
public:
	float Feedback;
	dword BufferLength;
	float *BufferPos;
	float *Buffer;
	
};

//*XDSP*
static CXAllpassFilter *AllpassCreate(dword const delay)
{
	return new CAllpassFPU(delay);
}

//*XDSP*
static void AllpassSetFeedback(CXAllpassFilter *_pc, float const fb)
{
	CAllpassFPU *pc = (CAllpassFPU *)_pc;
	pc->Feedback = fb;		
}


//*XDSP*
static void AllpassSetState(CXAllpassFilter *_pc, float const *state)
{
	CAllpassFPU *pc = (CAllpassFPU *)_pc;
	pc->SetState(state);	
}


//*XDSP*
static void AllpassRun(CXAllpassFilter *_pc, float *p, dword const n)
{
	CAllpassFPU *pc = (CAllpassFPU *)_pc;
	
	dword num = n;
	
	do
	{
		dword const x = __min(num, pc->BufferLength - (pc->BufferPos - pc->Buffer));
		
		for (dword c = 0; c < x; c++, p++)
		{
			float b = *pc->BufferPos;
			undenormalise(b);
			
			float const in = *p;
			
			*p = -in + b;
			*pc->BufferPos = in + (b * pc->Feedback);
			pc->BufferPos++;
			
		}
		
		if (pc->BufferPos == pc->Buffer + pc->BufferLength)
			pc->BufferPos = pc->Buffer;
		
		num -= x;
		
	} while(num > 0);
	
}

/*
void SetAllpassFPU(CXDSPInterface *p)
{
	p->AllpassCreate = AllpassCreate;
	p->AllpassSetState = AllpassSetState;
	p->AllpassSetFeedback = AllpassSetFeedback;
	p->AllpassRun = AllpassRun;
}
*/

#include "set/AllpassFPU.set"