
#include "stdafx.h"
#include "XDSPInterface.h"
#include <xmmintrin.h>
#include "dspinternal.h"

void SSEClear_V(float *p, dword const n);
void FPUCopy_V_V(float *pout, float const *pin, dword const n);




class CAllpassSSE : public CXAllpassFilter
{
public:
	CAllpassSSE(dword const delay)
	{
		BufferPos = Buffer = NULL;
		BufferLength = 0;
		Feedback = 0.5;
		SetDelay(delay);
	}

	virtual ~CAllpassSSE()
	{
		_aligned_free(Buffer);
	}

	virtual dword QueryMemoryUsage() const { return sizeof(this) + BufferLength * sizeof(float); }
	
	void SetDelay(dword d)
	{
		BufferLength = d;
		_aligned_free(Buffer);
		BufferPos = Buffer = (float *)_aligned_malloc(d * sizeof(float), 32);
		SSEClear_V(Buffer, d);
		
	}

	void SetState(float const *state)
	{
		if (state == NULL)
			SSEClear_V(Buffer, BufferLength);
		else
			FPUCopy_V_V(Buffer, state, BufferLength);
	}

public:
	float Feedback;
	dword BufferLength;
	float *BufferPos;
	float *Buffer;
	
};

//*XDSP*S
static CXAllpassFilter *AllpassCreate(dword const delay)
{
	return new CAllpassSSE(delay);
}


//*XDSP*S
static void AllpassSetFeedback(CXAllpassFilter *_pc, float const fb)
{
	CAllpassSSE *pc = (CAllpassSSE *)_pc;
	pc->Feedback = fb;		
}

//*XDSP*S
static void AllpassSetState(CXAllpassFilter *_pc, float const *state)
{
	CAllpassSSE *pc = (CAllpassSSE *)_pc;
	pc->SetState(state);	
}


//*XDSP*S
static void AllpassRun(CXAllpassFilter *_pc, float *p, dword const n)
{
	CAllpassSSE *pc = (CAllpassSSE *)_pc;
	
	dword num = n;
	
	dword temp = 0x80000000;

	__m128 const signbitmask = _mm_load1_ps((float *)&temp);
	__m128 const fb = _mm_set1_ps(pc->Feedback);

	do
	{
		dword x = __min(num, pc->BufferLength - (pc->BufferPos - pc->Buffer));
		num -= x;
		
		if ((dword)p & 15)
		{
			float *dest = (float *)(((dword)p + 15) & (~0x0f));
			while(p != dest && x > 0)
			{
				__m128 const b = _mm_load_ss(pc->BufferPos);
				__m128 const in = _mm_load_ss(p);
				_mm_store_ss(p, _mm_add_ss(_mm_xor_ps(in, signbitmask), b));
				_mm_store_ss(pc->BufferPos, _mm_add_ss(in, _mm_mul_ss(b, fb)));
				pc->BufferPos++;
				p++;
				x--;
			}

		}

		for (dword c = 0; c < x/4; c++, p += 4)
		{
			__m128 const b = _mm_loadu_ps(pc->BufferPos);
			__m128 const in = _mm_load_ps(p);
			_mm_store_ps(p, _mm_add_ps(_mm_xor_ps(in, signbitmask), b));
			_mm_storeu_ps(pc->BufferPos, _mm_add_ps(in, _mm_mul_ps(b, fb)));
			pc->BufferPos += 4;
		}
		
		dword const last = x & 3;
		if (last > 0)
		{
			for (dword c = 0; c < last; c++, p++)
			{
				__m128 const b = _mm_load_ss(pc->BufferPos);
				__m128 const in = _mm_load_ss(p);
				_mm_store_ss(p, _mm_add_ss(_mm_xor_ps(in, signbitmask), b));
				_mm_store_ss(pc->BufferPos, _mm_add_ss(in, _mm_mul_ss(b, fb)));
				pc->BufferPos++;
			}
			
		}

		if (pc->BufferPos == pc->Buffer + pc->BufferLength)
			pc->BufferPos = pc->Buffer;
		
		
	} while(num > 0);
	
} 

/*
void SetAllpassSSE(CXDSPInterface *p)
{
	p->AllpassCreate = AllpassCreate;
	p->AllpassSetFeedback = AllpassSetFeedback;
	p->AllpassSetState = AllpassSetState;
	p->AllpassRun = AllpassRun;
}
*/

#include "set/AllpassSSE.set"