// DSPBase.cpp: implementation of the CDSPBase class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include <malloc.h>
#include "XDSPInterface.h"
#include "dspinternal.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

float g_SilenceThreshold = (1.0f / (1 << 20));

//*XDSP*
static void Init(dword const samplerate, float const silencethreshold)
{
	g_SilenceThreshold = silencethreshold;
}


void FPUClear_V(float *p, dword const n)
{
	memset(p, 0, n * sizeof(float));

}



void FPUCopy_V_V(float *pout, float const *pin, dword const n)
{
	memcpy(pout, pin, n * sizeof(float));
}




//*XDSP*
static void Clear_V(float *p, dword const n)
{
	memset(p, 0, n * sizeof(float));
}

//*XDSP*
static void Copy_V_V(float *pout, float const *pin, dword const n)
{
	memcpy(pout, pin, n * sizeof(float));
}

//*XDSP*
static void Copy_V_VU(float *pout, float const *pin, dword const n)
{
	memcpy(pout, pin, n * sizeof(float));
}

//*XDSP*
static void Copy_V_S(float *p, dword const n, float const a)
{
	for (dword c = 0; c < n/4; c++, p += 4)
	{
		p[0] = a;
		p[1] = a;
		p[2] = a;
		p[3] = a;
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++)
			p[c] = a;
	}
	
}

//*XDSP*
static void Add_V_S(float *p, dword const n, float const a)
{
	for (dword c = 0; c < n/4; c++, p += 4)
	{
		p[0] += a;
		p[1] += a;
		p[2] += a;
		p[3] += a;
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++)
			p[c] += a;
	}
	
}


//*XDSP*
static void Add_V_V(float *pout, float const *pin, dword const n)
{
	int i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		pout[i] += pin[i];
		pout[i+1] += pin[i+1];
		pout[i+2] += pin[i+2];
		pout[i+3] += pin[i+3];
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			pout[i] += pin[i];
	}
}

//*XDSP*
static void Add_VI_VI(int *pout, int const *pin, dword const n)
{
	int i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		pout[i] += pin[i];
		pout[i+1] += pin[i+1];
		pout[i+2] += pin[i+2];
		pout[i+3] += pin[i+3];
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			pout[i] += pin[i];
	}
}

//*XDSP*
static void Add_VS_VS(short *pout, short const *pin, dword const n)
{
	int i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		pout[i] += pin[i];
		pout[i+1] += pin[i+1];
		pout[i+2] += pin[i+2];
		pout[i+3] += pin[i+3];
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			pout[i] += pin[i];
	}
}

//*XDSP*
static void Add_VI_SI(int *pout, dword const n, int const a)
{
	int i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		pout[i] += a;
		pout[i+1] += a;
		pout[i+2] += a;
		pout[i+3] += a;
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			pout[i] += a;
	}
}


//*XDSP*
static void Sub_V_V(float *pout, float const *pin, dword const n)
{
	int i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		pout[i] -= pin[i];
		pout[i+1] -= pin[i+1];
		pout[i+2] -= pin[i+2];
		pout[i+3] -= pin[i+3];
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			pout[i] -= pin[i];
	}
}

//*XDSP*
static void Sub_V_V_V(float *pout, float const *pina, float const *pinb, dword const n)
{
	int i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		pout[i] = pina[i] - pinb[i];
		pout[i+1] = pina[i+1] - pinb[i+1];
		pout[i+2] = pina[i+2] - pinb[i+2];
		pout[i+3] = pina[i+3] - pinb[i+3];
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			pout[i] = pina[i] - pinb[i];
	}
}

//*XDSP*
static void Sub_VI_VI(int *pout, int const *pin, dword const n)
{
	int i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		pout[i] -= pin[i];
		pout[i+1] -= pin[i+1];
		pout[i+2] -= pin[i+2];
		pout[i+3] -= pin[i+3];
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			pout[i] -= pin[i];
	}
}

//*XDSP*
static void Sub_VS_VS(short *pout, short const *pin, dword const n)
{
	int i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		pout[i] -= pin[i];
		pout[i+1] -= pin[i+1];
		pout[i+2] -= pin[i+2];
		pout[i+3] -= pin[i+3];
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			pout[i] -= pin[i];
	}
}

//*XDSP*
static void Mul_V_V(float *pout, float const *pin, dword const n)
{
	int i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		pout[i] *= pin[i];
		pout[i+1] *= pin[i+1];
		pout[i+2] *= pin[i+2];
		pout[i+3] *= pin[i+3];
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			pout[i] *= pin[i];
	}
}


//*XDSP*
static void Mul_V_S(float *p, dword const n, float const a)
{
	for (dword c = 0; c < n/4; c++, p += 4)
	{
		p[0] *= a;
		p[1] *= a;
		p[2] *= a;
		p[3] *= a;
	}

	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++)
			p[c] *= a;
	}
}

//*XDSP*
static void Mul_V2_S2(float *p, dword const n, float const la, float const ra)
{
	for (dword c = 0; c < n/2; c++, p += 4)
	{
		p[0] *= la;
		p[1] *= ra;
		p[2] *= la;
		p[3] *= ra;
	}
	
	if (n & 1)
	{
		p[0] *= la;
		p[1] *= ra;
	}
}

//*XDSP*
static void Mul_V4_S4(float *p, dword const n, float const a0, float const a1, float const a2, float const a3)
{
	for (dword c = 0; c < n; c++, p += 4)
	{
		p[0] *= a0;
		p[1] *= a1;
		p[2] *= a2;
		p[3] *= a3;
	}
	
}



//*XDSP*
static void MulCopy_V_V_S(float *pout, float const *pin, dword const n, float const a)
{
	dword i = 0;

	for (dword c = 0; c < n/4; c++, i += 4)
	{
		pout[i] = pin[i] * a;
		pout[i+1] = pin[i+1] * a;
		pout[i+2] = pin[i+2] * a;
		pout[i+3] = pin[i+3] * a;
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			pout[i] = pin[i] * a;
	}
}


//*XDSP*
static void MulCopy_V2_V2_S2(float *pout, float const *pin, dword const n, float const la, float const ra)
{
	dword i = 0;
	
	for (dword c = 0; c < n/2; c++, i += 4)
	{
		pout[i] = pin[i] * la;
		pout[i+1] = pin[i+1] * ra;
		pout[i+2] = pin[i+2] * la;
		pout[i+3] = pin[i+3] * ra;
	}
	
	if (n & 1)
	{
		pout[i] = pin[i] * la;
		pout[i+1] = pin[i+1] * ra;
	}
}

//*XDSP*
static void MulCopy_V2_V_S2(float *pout, float const *pin, dword const n, float const la, float const ra)
{
	dword i = 0;
	
	if ((*(int *)&la & 0x7fffffff) == 0)
	{
		for (dword c = 0; c < n/2; c++, i += 2)
		{
			pout[i*2+1] = pin[i] * ra;
			pout[i*2+3] = pin[i+1] * ra;
		}
		
		if (n & 1)
			pout[i*2+1] = pin[i] * ra;
		
	}
	else if ((*(int *)&ra & 0x7fffffff) == 0)
	{
		for (dword c = 0; c < n/2; c++, i += 2)
		{
			pout[i*2+0] = pin[i] * la;
			pout[i*2+2] = pin[i+1] * la;
		}
		
		if (n & 1)
			pout[i*2+0] = pin[i] * la;
	}
	else
	{
		for (dword c = 0; c < n/2; c++, i += 2)
		{
			pout[i*2+0] = pin[i] * la;
			pout[i*2+1] = pin[i] * ra;
			pout[i*2+2] = pin[i+1] * la;
			pout[i*2+3] = pin[i+1] * ra;
		}
		
		if (n & 1)
		{
			pout[i*2+0] = pin[i] * la;
			pout[i*2+1] = pin[i] * ra;
		}
	}
}


//*XDSP*
static void Mul_V_L(float *ps, dword const n, float &refa, float const da)
{
	float a = refa;

	dword i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		ps[i] *= a; a += da;
		ps[i+1] *= a; a += da;
		ps[i+2] *= a; a += da;
		ps[i+3] *= a; a += da;
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
		{
			ps[i] *= a; a += da;
		}
	}

	refa = a;
}

//*XDSP*
static void Copy_V_L(float *ps, dword const n, float &refa, float const da)
{
	float a = refa;

	dword i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		ps[i] = a; a += da;
		ps[i+1] = a; a += da;
		ps[i+2] = a; a += da;
		ps[i+3] = a; a += da;
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
		{
			ps[i] = a; a += da;
		}
	}

	refa = a;
}

//*XDSP*
static void Add_V_L(float *ps, dword const n, float &refa, float const da)
{
	float a = refa;

	dword i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		ps[i] += a; a += da;
		ps[i+1] += a; a += da;
		ps[i+2] += a; a += da;
		ps[i+3] += a; a += da;
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
		{
			ps[i] += a; a += da;
		}
	}

	refa = a;
}


//*XDSP*
static void Mul_V_E(float *ps, dword const n, float &refa, float const da)
{
	float a = refa;

	dword i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		ps[i] *= a; a *= da;
		ps[i+1] *= a; a *= da;
		ps[i+2] *= a; a *= da;
		ps[i+3] *= a; a *= da;
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
		{
			ps[i] *= a; a *= da;
		}
	}

	refa = a;
}

//*XDSP*
static void Copy_V_E(float *ps, dword const n, float &refa, float const da)
{
	float a = refa;

	dword i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		ps[i] = a; a *= da;
		ps[i+1] = a; a *= da;
		ps[i+2] = a; a *= da;
		ps[i+3] = a; a *= da;
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
		{
			ps[i] = a; a *= da;
		}
	}

	refa = a;
}

//*XDSP*
static void MulCopy_V_V_L(float *pout, float const *pin, dword const n, float a, float const da)
{
	dword i = 0;

	for (dword c = 0; c < n/4; c++, i += 4)
	{
		pout[i] = pin[i] * a; a += da;
		pout[i+1] = pin[i+1] * a; a += da;
		pout[i+2] = pin[i+2] * a; a += da;
		pout[i+3] = pin[i+3] * a; a += da;
	}

	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
		{
			pout[i] = pin[i] * a; a += da;
		}
	}

}

//*XDSP*
static void MulCopy_V2_V2_L2(float *pout, float const *pin, dword const n, float la, float ra, float const dla, float const dra)
{
	dword i = 0;
	
	for (dword c = 0; c < n/2; c++, i += 4)
	{
		pout[i] = pin[i] * la; la += dla;
		pout[i+1] = pin[i+1] * ra; ra += dra;
		pout[i+2] = pin[i+2] * la; la += dla;
		pout[i+3] = pin[i+3] * ra; ra += dra;
	}
	
	if (n & 1)
	{
		pout[i] = pin[i] * la;
		pout[i+1] = pin[i+1] * ra; 
	}
	
}


//*XDSP*
static void MulAdd_V_V_S(float *pout, float const *pin, dword const n, float const a)
{
	dword i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		pout[i] += pin[i] * a;
		pout[i+1] += pin[i+1] * a;
		pout[i+2] += pin[i+2] * a;
		pout[i+3] += pin[i+3] * a;
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			pout[i] += pin[i] * a;
	}
}

//*XDSP*
static void MulAdd_V2_V2_S2(float *pout, float const *pin, dword const n, float const la, float const ra)
{
	dword i = 0;
	
	for (dword c = 0; c < n/2; c++, i += 4)
	{
		pout[i] += pin[i] * la;
		pout[i+1] += pin[i+1] * ra;
		pout[i+2] += pin[i+2] * la;
		pout[i+3] += pin[i+3] * ra;
	}
	
	if (n & 1)
	{
		pout[i] += pin[i] * la;
		pout[i+1] += pin[i+1] * ra;
	}
}

//*XDSP*
static void MulAdd_V2_V_S2(float *pout, float const *pin, dword const n, float const la, float const ra)
{
	dword i = 0;

	if ((*(int *)&la & 0x7fffffff) == 0)
	{
		for (dword c = 0; c < n/2; c++, i += 2)
		{
			pout[i*2+1] += pin[i] * ra;
			pout[i*2+3] += pin[i+1] * ra;
		}
		
		if (n & 1)
			pout[i*2+1] += pin[i] * ra;
		
	}
	else if ((*(int *)&ra & 0x7fffffff) == 0)
	{
		for (dword c = 0; c < n/2; c++, i += 2)
		{
			pout[i*2+0] += pin[i] * la;
			pout[i*2+2] += pin[i+1] * la;
		}
		
		if (n & 1)
			pout[i*2+0] += pin[i] * la;
	}
	else
	{
		for (dword c = 0; c < n/2; c++, i += 2)
		{
			pout[i*2+0] += pin[i] * la;
			pout[i*2+1] += pin[i] * ra;
			pout[i*2+2] += pin[i+1] * la;
			pout[i*2+3] += pin[i+1] * ra;
		}
		
		if (n & 1)
		{
			pout[i*2+0] += pin[i] * la;
			pout[i*2+1] += pin[i] * ra;
		}
	}
}

//*XDSP*
static void MulAdd_V2_V_L2(float *pout, float const *pin, dword const n, float la, float ra, float const lstep, float const rstep)
{
	dword i = 0;
	
	for (dword c = 0; c < n/2; c++, i += 2)
	{
		pout[i*2+0] += pin[i] * la; la += lstep;
		pout[i*2+1] += pin[i] * ra; ra += rstep;
		pout[i*2+2] += pin[i+1] * la; la += lstep;
		pout[i*2+3] += pin[i+1] * ra; ra += rstep;
	}
	
	if (n & 1)
	{
		pout[i*2+0] += pin[i] * la; la += lstep;
		pout[i*2+1] += pin[i] * ra; ra += rstep;
	}
}


//*XDSP*
static void MulAdd_V_V_L(float *pout, float const *pin, dword const n, float a, float const da)
{
	
	dword i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		pout[i] += pin[i] * a; a += da;
		pout[i+1] += pin[i+1] * a; a += da;
		pout[i+2] += pin[i+2] * a; a += da;
		pout[i+3] += pin[i+3] * a; a += da;
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
		{
			pout[i] += pin[i] * a; a += da;
		}
	}
	
}

//*XDSP*
static void MulAdd_V2_V2_L2(float *pout, float const *pin, dword const n, float la, float ra, float const dla, float const dra)
{
	dword i = 0;
	
	for (dword c = 0; c < n/2; c++, i += 4)
	{
		pout[i] += pin[i] * la; la += dla;
		pout[i+1] += pin[i+1] * ra; ra += dra;
		pout[i+2] += pin[i+2] * la; la += dla;
		pout[i+3] += pin[i+3] * ra; ra += dra;
	}
	
	if (n & 1)
	{
		pout[i] += pin[i] * la;
		pout[i+1] += pin[i+1] * ra; 
	}
	
}


//*XDSP*
static void Copy_V2_V2N(float *pout, float const *pin0, float const *pin1, dword const n)
{
	for (dword c = 0; c < n-1; c += 2, pout += 4)
	{
		pout[0] = pin0[c];
		pout[1] = pin1[c];

		pout[2] = pin0[c+1];
		pout[3] = pin1[c+1];
	}

	if (n & 1)
	{
		pout[0] = pin0[c];
		pout[1] = pin1[c];
	}
}

//*XDSP*
static void Copy_V2N_V2(float *pout0, float *pout1, float const *pin, dword const n)
{
	for (dword c = 0; c < n-1; c += 2, pin += 4)
	{
		pout0[c] = pin[0];
		pout1[c] = pin[1];

		pout0[c+1] = pin[2];
		pout1[c+1] = pin[3];
		
	}

	if (n & 1)
	{
		pout0[c] = pin[0];
		pout1[c] = pin[1];
	}

}


//*XDSP*
static void Add_V2N_V2(float *pout0, float *pout1, float const *pin, dword const n)
{
	for (dword c = 0; c < n-1; c += 2, pin += 4)
	{
		pout0[c] += pin[0];
		pout1[c] += pin[1];
		
		pout0[c+1] += pin[2];
		pout1[c+1] += pin[3];
		
	}
	
	if (n & 1)
	{
		pout0[c] += pin[0];
		pout1[c] += pin[1];
	}
	
}



//*XDSP*
static void Copy_V4_V4N(float *pout, float const *pin0, float const *pin1, float const *pin2, float const *pin3, dword const n)
{

	for (dword c = 0; c < n; c++, pout += 4)
	{
		pout[0] = pin0[c];
		pout[1] = pin1[c];
		pout[2] = pin2[c];
		pout[3] = pin3[c];
	}
		
}

//*XDSP*
static void Copy_V4N_V4(float *pout0, float *pout1, float *pout2, float *pout3, float const *pin, dword const n)
{
	for (dword c = 0; c < n; c++, pin += 4)
	{
		pout0[c] = pin[0];
		pout1[c] = pin[1];
		pout2[c] = pin[2];
		pout3[c] = pin[3];
	}
}



//*XDSP*
static void Copy_V_VS(float *pout, short const *pin, dword const n)
{
	dword i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		pout[i] = (float)pin[i];
		pout[i+1] = (float)pin[i+1];
		pout[i+2] = (float)pin[i+2];
		pout[i+3] = (float)pin[i+3];
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			pout[i] = (float)pin[i];
	}
	
}


//*XDSP*
static void Copy_V_VI(float *pout, int const *pin, dword const n)
{
	dword i = 0;

	for (dword c = 0; c < n/4; c++, i += 4)
	{
		pout[i] = (float)pin[i];
		pout[i+1] = (float)pin[i+1];
		pout[i+2] = (float)pin[i+2];
		pout[i+3] = (float)pin[i+3];
	}

	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			pout[i] = (float)pin[i];
	}
	
}

//*XDSP*
static void Copy_V_VIP(float *p, dword const n)
{
	dword i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		p[i] = (float)((int *)p)[i];
		p[i+1] = (float)((int *)p)[i+1];
		p[i+2] = (float)((int *)p)[i+2];
		p[i+3] = (float)((int *)p)[i+3];
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			p[i] = (float)((int *)p)[i];
	}
	
}

//*XDSP*
static void ClipCopy_VS_V(short *pout, float const *pin, dword const n)
{
	double const d2i = (1.5 * (1 << 26) * (1 << 26));
 
	dword * const pouti = (dword *)pout;

	for (dword c = 0; c < n/2; c++)
	{
		float x = pin[c*2+0];
		if (x < _SHORT_MIN)
			x = _SHORT_MIN;
		else if (x > _SHORT_MAX)
			x = _SHORT_MAX;

		float y = pin[c*2+1];
		if (y < _SHORT_MIN)
			y = _SHORT_MIN;
		else if (y > _SHORT_MAX)
			y = _SHORT_MAX;
		
		double res = ((double)x) + d2i;
		int r = *(int *)&res;
		
		double res2 = ((double)y) + d2i;
		int r2 = *(int *)&res2;

		pouti[c] = (r2 << 16) | (r & 0xffff);
	}

	if (n & 1)
	{
		float x = pin[n-1];
		if (x < _SHORT_MIN)
			x = _SHORT_MIN;
		else if (x > _SHORT_MAX)
			x = _SHORT_MAX;
		
		double res = ((double)x) + d2i;
		int r = *(int *)&res;
		pout[n-1] = (short)r;
	}
	
}

//*XDSP*
static void Copy_VS_V(short *pout, float const *pin, dword const n)
{
	double const d2i = (1.5 * (1 << 26) * (1 << 26));
	
	dword * const pouti = (dword *)pout;
	
	for (dword c = 0; c < n/2; c++)
	{
		float x = pin[c*2+0];
		float y = pin[c*2+1];
		
		double res = ((double)x) + d2i;
		int r = *(int *)&res;
		
		double res2 = ((double)y) + d2i;
		int r2 = *(int *)&res2;
		
		pouti[c] = (r2 << 16) | (r & 0xffff);
	}
	
	if (n & 1)
	{
		float x = pin[n-1];
		
		double res = ((double)x) + d2i;
		int r = *(int *)&res;
		pout[n-1] = (short)r;
	}
	
}


//*XDSP*
static void ClipCopy_VI_V(int *pout, float const *pin, dword const n)
{
	double const d2i = (1.5 * (1 << 26) * (1 << 26));
	
	for (dword c = 0; c < n; c++)
	{
		double x = pin[c];
		if (x < _INT_MIN) 
			x = _INT_MIN;
		else if (x > _INT_MAX)
			x = _INT_MAX;
		
		double res = ((double)x) + d2i;
		int r = *(int *)&res;
		pout[c] = r;
		
	}
	
}

//*XDSP*
static void Copy_VI_V(int *pout, float const *pin, dword const n)
{
	double const d2i = (1.5 * (1 << 26) * (1 << 26));
	
	for (dword c = 0; c < n; c++)
	{
		double x = pin[c];
		
		double res = ((double)x) + d2i;
		int r = *(int *)&res;
		pout[c] = r;
		
	}
	
}


//*XDSP*
static void ClipCopy_VI_VP(int *p, dword const n)
{
	double const d2i = (1.5 * (1 << 26) * (1 << 26));
	
	for (dword c = 0; c < n; c++)
	{
		double x = ((float *)p)[c];
		if (x < _INT_MIN) 
			x = _INT_MIN;
		else if (x > _INT_MAX)
			x = _INT_MAX;
		
		double res = ((double)x) + d2i;
		int r = *(int *)&res;
		p[c] = r;
		
	}
	
}

//*XDSP*
static void Copy_VI_VP(int *p, dword const n)
{
	double const d2i = (1.5 * (1 << 26) * (1 << 26));
	
	for (dword c = 0; c < n; c++)
	{
		double x = ((float *)p)[c];
		
		double res = ((double)x) + d2i;
		int r = *(int *)&res;
		p[c] = r;
		
	}
	
}

//*XDSP*
static void CopyShift_VI_VP(int *p, dword const n, dword const shift)
{
	double const d2i = (1.5 * (1 << 26) * (1 << 26));
	
	for (dword c = 0; c < n; c++)
	{
		double x = ((float *)p)[c];
		
		double res = ((double)x) + d2i;
		int r = *(int *)&res;
		p[c] = r << shift;
		
	}
	
}


//*XDSP*
static void Copy_V_V4(float *pout, float const *pin, dword const n)
{
	for (dword c = 0; c < n; c++, pin += 4)
	{
		pout[c] = pin[0] + pin[1] + pin[2] + pin[3];
	}
}

//*XDSP*
static void Copy_V2_V4(float *pout, float const *pin, dword const n)
{
	for (dword c = 0; c < n; c++, pout += 2, pin += 4)
	{
		pout[0] = pin[0] + pin[2];
		pout[1] = pin[1] + pin[3];
	}
}


static void Resample(float *pout, float const *pin, dword const n)
{
}


//*XDSP*
static bool AbsMaxGT_V_S(float const *pin, dword const n, float const x)
{
	dword i = 0;

	for (dword c = 0; c < n/4; c++, i += 4)
	{
		if (fabs(pin[i+0]) > x)
			return true;
		if (fabs(pin[i+1]) > x)
			return true;
		if (fabs(pin[i+2]) > x)
			return true;
		if (fabs(pin[i+3]) > x)
			return true;
	}

	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			if (pin[i] > x)
				return true;
	}
	
	return false;
}


//*XDSP*
static float AbsMax_V(float const *pin, dword const n)
{
	float m = 0;

	dword i = 0;

	for (dword c = 0; c < n/4; c++, i += 4)
	{
		m = (float)__max(m, fabs(pin[i+0]));
		m = (float)__max(m, fabs(pin[i+1]));
		m = (float)__max(m, fabs(pin[i+2]));
		m = (float)__max(m, fabs(pin[i+3]));
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			m = (float)__max(m, fabs(pin[i]));
	}
	
	return m;
}

//*XDSP*
static int AbsMax_VI(int const *pin, dword const n)
{
	int m = 0;

	for (dword i = 0; i < n; i++)
	{
		int x = labs(pin[i]);
		if (x > m)
			m = x;
	}
		
	return m;
}

//*XDSP*
static short AbsMax_VS(short const *pin, dword const n)
{
	short m = 0;
	
	for (dword i = 0; i < n; i++)
	{
		short x = (short)labs(pin[i]);
		if (x > m)
			m = x;
	}
	
	return m;
}

//*XDSP*
static void Abs_VI(int *p, dword const n)
{
	for (dword c = 0; c < n/4; c++, p += 4)
	{
		p[0] = labs(p[0]);
		p[1] = labs(p[1]);
		p[2] = labs(p[2]);
		p[3] = labs(p[3]);
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++)
			p[c] = labs(p[c]);
	}
}

//*XDSP*
static void Abs_VS(short *p, dword const n)
{
	for (dword c = 0; c < n/4; c++, p += 4)
	{
		p[0] = (short)labs(p[0]);
		p[1] = (short)labs(p[1]);
		p[2] = (short)labs(p[2]);
		p[3] = (short)labs(p[3]);
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++)
			p[c] = (short)labs(p[c]);
	}
}

//*XDSP*
static void And_VI_SI(int *p, dword const n, int const x)
{
	for (dword c = 0; c < n/4; c++, p += 4)
	{
		p[0] &= x;
		p[1] &= x;
		p[2] &= x;
		p[3] &= x;
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++)
			p[c] &= x;
	}
}

//*XDSP*
static void Or_VI_SI(int *p, dword const n, int const x)
{
	for (dword c = 0; c < n/4; c++, p += 4)
	{
		p[0] |= x;
		p[1] |= x;
		p[2] |= x;
		p[3] |= x;
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++)
			p[c] |= x;
	}
}

//*XDSP*
static void Xor_VI_SI(int *p, dword const n, int const x)
{
	for (dword c = 0; c < n/4; c++, p += 4)
	{
		p[0] ^= x;
		p[1] ^= x;
		p[2] ^= x;
		p[3] ^= x;
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++)
			p[c] ^= x;
	}
}

//*XDSP*
void And_VI_VI(int *pout, int const *pin, dword const n)
{
	int i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		pout[i] &= pin[i];
		pout[i+1] &= pin[i+1];
		pout[i+2] &= pin[i+2];
		pout[i+3] &= pin[i+3];
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			pout[i] &= pin[i];
	}
}

//*XDSP*
void AndNot1_VI_VI(int *pout, int const *pin, dword const n)
{
	int i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		pout[i] = ~pout[i] & pin[i];
		pout[i+1] = ~pout[i+1] & pin[i+1];
		pout[i+2] = ~pout[i+2] & pin[i+2];
		pout[i+3] = ~pout[i+3] & pin[i+3];
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			pout[i] = ~pout[i] & pin[i];
	}
}

//*XDSP*
void AndNot2_VI_VI(int *pout, int const *pin, dword const n)
{
	int i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		pout[i] &= ~pin[i];
		pout[i+1] &= ~pin[i+1];
		pout[i+2] &= ~pin[i+2];
		pout[i+3] &= ~pin[i+3];
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			pout[i] &= ~pin[i];
	}
}

//*XDSP*
static void Or_VI_VI(int *pout, int const *pin, dword const n)
{
	int i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		pout[i] |= pin[i];
		pout[i+1] |= pin[i+1];
		pout[i+2] |= pin[i+2];
		pout[i+3] |= pin[i+3];
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			pout[i] |= pin[i];
	}
}

//*XDSP*
static void Xor_VI_VI(int *pout, int const *pin, dword const n)
{
	int i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		pout[i] ^= pin[i];
		pout[i+1] ^= pin[i+1];
		pout[i+2] ^= pin[i+2];
		pout[i+3] ^= pin[i+3];
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			pout[i] ^= pin[i];
	}
}

/*
static void EvaluatePolynomial_(float *p, dword const n, float *pc, dword const npc)
{
	for (dword i = 0; i < n; i++)
	{
		float const x = p[i];
		float product = pc[0];

		for (dword k = 1; k < npc; k++)
		{
			product *= x;
			product += pc[k];
		}

		p[i] = product;		
	}

}
*/

//*XDSP*
static void EvaluatePolynomial3_V(float *p, dword const n, float *pc, float const *pdc)
{
	if (pdc != NULL)
	{
		for (dword i = 0; i < n; i++)
		{
			float const x = p[i];
			p[i] = (pc[0] * x + pc[1]) * x + pc[2];
			pc[0] += pdc[0];
			pc[1] += pdc[1];
			pc[2] += pdc[2];
		}
	}
	else
	{
		for (dword i = 0; i < n; i++)
		{
			float const x = p[i];
			p[i] = (pc[0] * x + pc[1]) * x + pc[2];
		}
	
	}
	
}

//*XDSP*
static void EvaluatePolynomial4_V(float *p, dword const n, float *pc, float const *pdc)
{
	if (pdc != NULL)
	{
		for (dword i = 0; i < n; i++)
		{
			float const x = p[i];
			p[i] = ((pc[0] * x + pc[1]) * x + pc[2]) * x + pc[3];
			pc[0] += pdc[0];
			pc[1] += pdc[1];
			pc[2] += pdc[2];
			pc[3] += pdc[3];
		}
	}
	else
	{
		for (dword i = 0; i < n; i++)
		{
			float const x = p[i];
			p[i] = ((pc[0] * x + pc[1]) * x + pc[2]) * x + pc[3];
		}
		
	}
	
}

static int g_stat = 0x16BA2118;

// idum = 1664525L*idum + 1013904223L;

#define RAND_MUL	1664525
#define RAND_ADD	1013904223

//*XDSP*
static void Rand_VI(int *pout, dword const n)
{
	int stat = g_stat;

	dword i = 0;

	for (dword c = 0; c < n/4; c++, i += 4)
	{
		stat = (stat * RAND_MUL + RAND_ADD);
		pout[i] = stat;
		
		stat = (stat * RAND_MUL + RAND_ADD);
		pout[i+1] = stat;
		
		stat = (stat * RAND_MUL + RAND_ADD);
		pout[i+2] = stat;
		
		stat = (stat * RAND_MUL + RAND_ADD);
		pout[i+3] = stat;
	}

	dword const r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
		{
			stat = (stat * RAND_MUL + RAND_ADD);
			pout[i] = stat;
		}
	}

	g_stat = stat;
}

//*XDSP*
static void RandExp_V(float *pout, dword const n)
{
	Rand_VI((int *)pout, n);
	And_VI_SI((int *)pout, n, 0x87ffffff);
	Add_VI_SI((int *)pout, n, (127-16) << 23);
	
}

//*XDSP*
static void RandExp2_V(float *pout, dword const n)
{
	Rand_VI((int *)pout, n);
	And_VI_SI((int *)pout, n, 0x83ffffff);
	Add_VI_SI((int *)pout, n, (127-8) << 23);
	
}


//*XDSP*
static void MulSar_VI_SI(int *p, dword const n, int const a, dword const shift)
{
	if (shift == 0)
	{
		for (dword c = 0; c < n/4; c++, p += 4)
		{
			p[0] *= a;
			p[1] *= a;
			p[2] *= a;
			p[3] *= a;
		}
		
		dword r = n & 3;
		if (r > 0)
		{
			for (dword c = 0; c < r; c++)
				p[c] *= a;
		}
		
	}
	else
	{
		for (dword c = 0; c < n/4; c++, p += 4)
		{
			p[0] = (p[0] * a) >> shift;
			p[1] = (p[1] * a) >> shift;
			p[2] = (p[2] * a) >> shift;
			p[3] = (p[3] * a) >> shift;
		}
		
		dword r = n & 3;
		if (r > 0)
		{
			for (dword c = 0; c < r; c++)
				p[c] = (p[c] * a) >> shift;
		}
		
	}

}



//*XDSP*
static void Shl_VI(int *p, dword const n, dword const shift)
{
	int i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		p[i] <<= shift;
		p[i+1] <<= shift;
		p[i+2] <<= shift;
		p[i+3] <<= shift;
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			p[i] <<= shift;
	}
	
}

//*XDSP*
static void Shr_VI(int *_p, dword const n, dword const shift)
{
	dword *p = (dword *)_p;
	int i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		p[i] >>= shift;
		p[i+1] >>= shift;
		p[i+2] >>= shift;
		p[i+3] >>= shift;
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			p[i] >>= shift;
	}
	
}

//*XDSP*
static void Sar_VI(int *p, dword const n, dword const shift)
{
	int i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		p[i] >>= shift;
		p[i+1] >>= shift;
		p[i+2] >>= shift;
		p[i+3] >>= shift;
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			p[i] >>= shift;
	}
	
}

#pragma optimize ("p", on)	// improve float consistency for the next function to avoid /x -> * (1/x)
 
//*XDSP*
static void Div_V_S(float *p, dword const n, float const a)
{
	for (dword c = 0; c < n/4; c++, p += 4)
	{
		p[0] /= a;
		p[1] /= a;
		p[2] /= a;
		p[3] /= a;
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++)
			p[c] /= a;
	}
	
}

#pragma optimize("", on)

//*XDSP*
static void Div_V_V(float *pout, float const *pin, dword const n)
{
	int i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		pout[i] /= pin[i];
		pout[i+1] /= pin[i+1];
		pout[i+2] /= pin[i+2];
		pout[i+3] /= pin[i+3];
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			pout[i] /= pin[i];
	}
}

//*XDSP*
static void ReciprocalApprox_V_V(float *pout, float const *pin, dword const n)
{
	int i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		pout[i] = 1.0f / pin[i];
		pout[i+1] = 1.0f / pin[i+1];
		pout[i+2] = 1.0f/ pin[i+2];
		pout[i+3] = 1.0f / pin[i+3];
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			pout[i] = 1.0f / pin[i];
	}
}

	
//*XDSP*
static void Sqrt_V(float *p, dword const n)
{
	for (dword c = 0; c < n/4; c++, p += 4)
	{
		p[0] = sqrt(p[0]);
		p[1] = sqrt(p[1]);
		p[2] = sqrt(p[2]);
		p[3] = sqrt(p[3]);
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++)
			p[c] = sqrt(p[c]);
	}
}

//*XDSP*
static void ReciprocalSqrtApprox_V_V(float *pout, float const *pin, dword const n)
{
	int i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		pout[i] = 1.0f / sqrt(pin[i]);
		pout[i+1] = 1.0f / sqrt(pin[i+1]);
		pout[i+2] = 1.0f/ sqrt(pin[i+2]);
		pout[i+3] = 1.0f / sqrt(pin[i+3]);
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			pout[i] = 1.0f / sqrt(pin[i]);
	}
}

//*XDSP*
static void SinCos_V2N_V(float *psin, float *pcos, float const *pin, dword const n)
{
	__asm
	{
		mov ecx, n
		mov esi, pin
		mov edx, psin
		mov edi, pcos
loopi:
		fld [esi]
		add esi, 4
		fsincos
		fstp [edi]
		add edi, 4
		fstp [edx]
		add edx, 4
			
		dec ecx
		jnz loopi
			
	}
	
}

//*XDSP*
void CompareLT_V_V_S(dword *pout, float const *pin, dword const n, float const a)
{
	int i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		pout[i+0] = pin[i+0] < a ? 0xffffffff : 0;
		pout[i+1] = pin[i+1] < a ? 0xffffffff : 0;
		pout[i+2] = pin[i+2] < a ? 0xffffffff : 0;
		pout[i+3] = pin[i+3] < a ? 0xffffffff : 0;
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			pout[i] = pin[i] < a ? 0xffffffff : 0;
	}
	
}

//*XDSP*
void CompareLT_V_V_V(dword *pout, float const *pin1, float const *pin2, dword const n)
{
	int i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		pout[i+0] = pin1[i+0] < pin2[i+0] ? 0xffffffff : 0;
		pout[i+1] = pin1[i+1] < pin2[i+1] ? 0xffffffff : 0;
		pout[i+2] = pin1[i+2] < pin2[i+2] ? 0xffffffff : 0;
		pout[i+3] = pin1[i+3] < pin2[i+3] ? 0xffffffff : 0;
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			pout[i] = pin1[i] < pin2[i] ? 0xffffffff : 0;
	}
	
}

//*XDSP*
static void Min_V_S(float *p, dword const n, float const a)
{
	for (dword c = 0; c < n/4; c++, p += 4)
	{
		p[0] = __min(p[0], a);
		p[1] = __min(p[1], a);
		p[2] = __min(p[2], a);
		p[3] = __min(p[3], a);
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++)
			p[c] = __min(p[c], a);
	}
}

//*XDSP*
static void Max_V_S(float *p, dword const n, float const a)
{
	for (dword c = 0; c < n/4; c++, p += 4)
	{
		p[0] = __max(p[0], a);
		p[1] = __max(p[1], a);
		p[2] = __max(p[2], a);
		p[3] = __max(p[3], a);
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++)
			p[c] = __max(p[c], a);
	}
}

//*XDSP*
static void MinMax_V_S(float *p, dword const n, float const _min, float const _max)
{
	for (dword c = 0; c < n/4; c++, p += 4)
	{
		p[0] = __max(__min(p[0], _min), _max);
		p[1] = __max(__min(p[1], _min), _max);
		p[2] = __max(__min(p[2], _min), _max);
		p[3] = __max(__min(p[3], _min), _max);
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++)
			p[c] = __max(__min(p[c], _min), _max);
	}
}

//*XDSP*
static void Min_V_V(float *pout, float const *pin, dword const n)
{
	dword i = 0;

	for (dword c = 0; c < n/4; c++, i += 4)
	{
		pout[i+0] = __min(pout[i+0], pin[i+0]);
		pout[i+1] = __min(pout[i+1], pin[i+1]);
		pout[i+2] = __min(pout[i+2], pin[i+2]);
		pout[i+3] = __min(pout[i+3], pin[i+3]);
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; i++)
			pout[i] = __min(pout[i], pin[i]);
	}
}

//*XDSP*
static void Max_V_V(float *pout, float const *pin, dword const n)
{
	dword i = 0;

	for (dword c = 0; c < n/4; c++, i += 4)
	{
		pout[i+0] = __max(pout[i+0], pin[i+0]);
		pout[i+1] = __max(pout[i+1], pin[i+1]);
		pout[i+2] = __max(pout[i+2], pin[i+2]);
		pout[i+3] = __max(pout[i+3], pin[i+3]);
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; i++)
			pout[i] = __max(pout[i], pin[i]);
	}
}

//*XDSP*
void CompareLTAdd_V_S_S(float *p, dword const n, float const a, float const b)
{
	dword i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		if (p[i+0] < a) p[i+0] += b;
		if (p[i+1] < a) p[i+1] += b;
		if (p[i+2] < a) p[i+2] += b;
		if (p[i+3] < a) p[i+3] += b;
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; i++)
			if (p[i] < a) p[i] += b;
	}
	
}

//*XDSP*
void CompareGTAdd_V_S_S(float *p, dword const n, float const a, float const b)
{
	dword i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		if (p[i+0] > a) p[i+0] += b;
		if (p[i+1] > a) p[i+1] += b;
		if (p[i+2] > a) p[i+2] += b;
		if (p[i+3] > a) p[i+3] += b;
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; i++)
			if (p[i] > a) p[i] += b;
	}
	
}

//*XDSP*
void CompareLEAdd_V_S_S(float *p, dword const n, float const a, float const b)
{
	dword i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		if (p[i+0] <= a) p[i+0] += b;
		if (p[i+1] <= a) p[i+1] += b;
		if (p[i+2] <= a) p[i+2] += b;
		if (p[i+3] <= a) p[i+3] += b;
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; i++)
			if (p[i] <= a) p[i] += b;
	}
	
}

//*XDSP*
void CompareGEAdd_V_S_S(float *p, dword const n, float const a, float const b)
{
	dword i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		if (p[i+0] >= a) p[i+0] += b;
		if (p[i+1] >= a) p[i+1] += b;
		if (p[i+2] >= a) p[i+2] += b;
		if (p[i+3] >= a) p[i+3] += b;
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; i++)
			if (p[i] >= a) p[i] += b;
	}
	
}

//*XDSP*
void AbsCompareLTCopy_V_S_S(float *_p, dword const n, float const _a, float const _b)
{
	int *p = (int *)_p;
	int const a = *(int const *)&_a;
	int const b = *(int const *)&_a;
	
	dword i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		if ((p[i+0] & 0x7fffffff) < a) p[i+0] = b;
		if ((p[i+1] & 0x7fffffff) < a) p[i+1] = b;
		if ((p[i+2] & 0x7fffffff) < a) p[i+2] = b;
		if ((p[i+3] & 0x7fffffff) < a) p[i+3] = b;
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; i++)
			if ((p[i] & 0x7fffffff) < a) p[i] = b;
	}
	
}


//*XDSP*
static void Abs_V_VCN(float *pout, float const *pr, float const *pi, dword const n)
{
	int i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		pout[i+0] = sqrt(pr[i+0] * pr[i+0] + pi[i+0] * pi[i+0]);
		pout[i+1] = sqrt(pr[i+1] * pr[i+1] + pi[i+1] * pi[i+1]);
		pout[i+2] = sqrt(pr[i+2] * pr[i+2] + pi[i+2] * pi[i+2]);
		pout[i+3] = sqrt(pr[i+3] * pr[i+3] + pi[i+3] * pi[i+3]);
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			pout[i] = sqrt(pr[i] * pr[i] + pi[i] * pi[i]);
	}
}

//*XDSP*
static void DenormalToZero_V(float *_p, dword const n)
{
	int *p = (int *)_p;
	
	for (dword c = 0; c < n/4; c++, p += 4)
	{
		if ((p[0] & 0x7f800000) == 0) p[0] = 0;
		if ((p[1] & 0x7f800000) == 0) p[1] = 0;
		if ((p[2] & 0x7f800000) == 0) p[2] = 0;
		if ((p[3] & 0x7f800000) == 0) p[3] = 0;
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, p++)
			if ((*p & 0x7f800000) == 0) *p = 0;

	}
	
}

/////////////////////////////////////////////////////////////////////////

/*
extern void SetCombFPU(CXDSPInterface *p);
extern void SetAllpassFPU(CXDSPInterface *p);
extern void SetBiquadFPU(CXDSPInterface *p);
extern void SetResamplerFPU(CXDSPInterface *p);
extern void SetFilterDesign(CXDSPInterface *p);
extern void SetMipWave(CXDSPInterface *p);
extern void SetMWOscillator(CXDSPInterface *p);
extern void SetOscArray(CXDSPInterface *p);
extern void SetCompress(CXDSPInterface *p);
extern void SetPow2(CXDSPInterface *p);
*/

//*XDSP*
static float *NewBuffer(dword const numsamples)
{
	return (float *)_aligned_malloc(numsamples * sizeof(float), 32);
}

//*XDSP*
static void DeleteBuffer(float *pbuf)
{
	_aligned_free(pbuf);
}

//*XDSP*
static int *NewIntBuffer(dword const numsamples)
{
	return (int *)_aligned_malloc(numsamples * sizeof(int), 32);
}

//*XDSP*
static void DeleteIntBuffer(int *pbuf)
{
	_aligned_free(pbuf);
} 

//*XDSP*
static short *NewShortBuffer(dword const numsamples)
{
	return (short *)_aligned_malloc(numsamples * sizeof(short), 32);
}

//*XDSP*
static void DeleteShortBuffer(short *pbuf)
{
	_aligned_free(pbuf);
}



/*
void SetFPU(CXDSPInterface *p)
{
	
	p->Init = Init;
	p->NewBuffer = NewBuffer;
	p->DeleteBuffer = DeleteBuffer;
	p->NewIntBuffer = NewIntBuffer;
	p->DeleteIntBuffer = DeleteIntBuffer;
	p->NewShortBuffer = NewShortBuffer;
	p->DeleteShortBuffer = DeleteShortBuffer;
	
	p->Clear_V = FPUClear_V;
	p->Copy_V_V = FPUCopy_V_V;
	p->Copy_V_VU = FPUCopy_V_V;
	p->Add_V_V = Add_V_V;
	p->Add_VI_VI = Add_VI_VI;
	p->Add_VS_VS = Add_VS_VS;
	p->Sub_V_V = Sub_V_V;
	p->Sub_VI_VI = Sub_VI_VI;
	p->Sub_VS_VS = Sub_VS_VS;
	p->Mul_V_V = Mul_V_V;
	p->Copy_V_S = Copy_V_S;
	p->Mul_V_S = Scale;
	p->Mul_V2_S2 = Scale2;
	p->Mul_V4_S4 = Scale4;
	p->Mul_V_L = ScaleRamp;
	p->Mul_V_E = ScaleRampExp;
	p->MulCopy_V_V_S = ScaleCopy;
	p->MulCopy_V2_V2_S2 = ScaleCopy2;
	p->MulCopy_V2_V_S2 = ScaleCopy1to2;
	p->MulCopy_V_V_L = ScaleRampCopy;
	p->MulCopy_V2_V2_L2 = ScaleRampCopy2;
	p->MulAdd_V_V_S = ScaleAdd;
	p->MulAdd_V2_V2_S2 = ScaleAdd2;
	p->MulAdd_V2_V_S2 = ScaleAdd1to2;
	p->MulAdd_V2_V_L2 = ScaleRampAdd1to2;
	p->MulAdd_V_V_L = ScaleRampAdd;
	p->MulAdd_V2_V2_L2 = ScaleRampAdd2;
	p->MulSar_VI_SI = MulSar_VI_SI;

	p->Shl_VI = Shl_VI;
	p->Shr_VI = Shr_VI;
	p->Sar_VI = Sar_VI;
	
	p->And_VI_SI = And;
	p->Or_VI_SI = Or;
	p->Xor_VI_SI = Xor;
	p->And_VI_VI = AndBySignal;
	p->AndNot1_VI_VI = AndNot1_VI_VI;
	p->AndNot2_VI_VI = AndNot2_VI_VI;
	p->Or_VI_VI = OrBySignal;
	p->Xor_VI_VI = XorBySignal;
	
	p->Copy_V2_V2N = Interleave2;
	p->Copy_V2N_V2 = Deinterleave2;
	p->Add_V2N_V2 = DeinterleaveAdd2;
	p->Copy_V4_V4N = Interleave4;
	p->Copy_V4N_V4 = Deinterleave4;

	p->Copy_V_V4 = AddInterleavedChannels4to1;
	p->Copy_V2_V4 = AddInterleavedChannels4to2;

	p->Abs_VI = Abs_VI;
	p->Abs_VS = Abs_VS;
	p->AbsMax_V = AbsMax;
	p->AbsMax_VI = AbsMax_VI;
	p->AbsMax_VS = AbsMax_VS;

	p->CompareLT_V_V_S = CompareLT_V_V_S;
	p->CompareLT_V_V_V = CompareLT_V_V_V;
	p->AbsMaxGT_V_S = AbsoluteMaximumGreaterThan;

	p->Div_V_S = Div_V_S;
	p->Div_V_V = Div_V_V;
	p->Sqrt_V = Sqrt_V;

	p->SinCos_V2N_V = SinCos_V2N_V;

	p->Min_V_S = Min_V_S;
	p->Max_V_S = Max_V_S;
	p->MinMax_V_S = MinMax_V_S;

	p->Copy_V_VS = ShortToFloat;
	p->Copy_V_VI = IntToFloat;
	p->Copy_V_VIP = IntToFloatInPlace;
	p->ClipCopy_VS_V = FloatToShort;
	p->Copy_VS_V = FloatToShortNoClip;
	p->ClipCopy_VI_V = FloatToInt;
	p->Copy_VI_V = FloatToIntNoClip;
	p->ClipCopy_VI_VP = FloatToIntInPlace;
	p->Copy_VI_VP = FloatToIntNoClipInPlace;
	p->CopyShift_VI_VP = FloatToIntNoClipShiftInPlace;


	p->Rand_VI = Rand_VI;
	p->EvaluatePolynomial3_V = EvaluatePolynomial3_V;
	p->EvaluatePolynomial4_V = EvaluatePolynomial4_V;
	
	//	p->Resample= Resample;

	SetBiquadFPU(p);
	SetCombFPU(p);
	SetAllpassFPU(p);
	SetResamplerFPU(p);
	SetFilterDesign(p);
	SetMipWave(p);	
	SetMWOscillator(p);
	SetOscArray(p);
	SetCompress(p);
	SetPow2(p);
		
}
*/

#include "set/BasicFPU.set"
