
#include "stdafx.h"
#include "XDSPInterface.h"
#include <xmmintrin.h>
#include "dspinternal.h"

extern float g_SilenceThreshold;

//*XDSP*S
static void Init(dword const samplerate, float const silencethreshold)
{
	g_SilenceThreshold = silencethreshold;
	_MM_SET_FLUSH_ZERO_MODE(_MM_FLUSH_ZERO_ON);
	
}


void SSEClear_V(float *p, dword const n)
{
	int z = 0;

	dword c = n / 16;
	if (c != 0)
	{
		__asm
		{
			movss xmm0, z
			shufps xmm0, xmm0, 0

			mov eax, c
			mov edi, p

	lop:
			movaps [edi], xmm0
			movaps [edi+16], xmm0
			movaps [edi+32], xmm0
			movaps [edi+48], xmm0
			add edi, 64
			dec eax
			jnz lop
				
		}
	}

	
	dword r = n & 15;
	
	if (r > 0)
		memset(p + c * 16, 0, r * sizeof(float));
}

//*XDSP*S
static void Clear_V(float *p, dword const n)
{
	int z = 0;
	
	dword c = n / 16;
	if (c != 0)
	{
		__asm
		{
			movss xmm0, z
				shufps xmm0, xmm0, 0
				
				mov eax, c
				mov edi, p
				
lop:
			movaps [edi], xmm0
				movaps [edi+16], xmm0
				movaps [edi+32], xmm0
				movaps [edi+48], xmm0
				add edi, 64
				dec eax
				jnz lop
				
		}
	}
	
	
	dword r = n & 15;
	
	if (r > 0)
		memset(p + c * 16, 0, r * sizeof(float));
}

//*XDSP*S
static void Copy_V_S(float *p, dword const n, float const a)
{
	dword const c = n / 16;
	if (c != 0)
	{
		__asm
		{
			movss xmm0, a
			shufps xmm0, xmm0, 0
			
			mov eax, c
			mov edi, p
				
lop:
			movaps [edi], xmm0
			movaps [edi+16], xmm0
			movaps [edi+32], xmm0
			movaps [edi+48], xmm0
			add edi, 64
			dec eax
			jnz lop
				
		}
	}
	
	
	dword r = n & 15;
	if (r > 0)
	{
		for (dword i = 0; i < r; i++)
			p[i + c * 16] = a;
	}
}

//*XDSP*S
static void Copy_V_V(float *pout, float const *pin, dword const n)
{

	dword c = n / 16;
	if (c != 0)
	{
		__asm
		{
				
			mov eax, c
			mov esi, pin
			mov edi, pout
				
lop:
			movaps xmm0, [esi]
			movaps [edi], xmm0

			movaps xmm0, [esi+16]
			movaps [edi+16], xmm0
			
			movaps xmm0, [esi+32]
			movaps [edi+32], xmm0

			movaps xmm0, [esi+48]
			movaps [edi+48], xmm0
			
			add esi, 64
			add edi, 64

			dec eax
			jnz lop
				
		}
	}
	
	
	dword r = n & 15;
	if (r > 0)
		memcpy(pout + c * 16, pin + c * 16, r * sizeof(float));
}

void FPUCopy_V_V(float *pout, float const *pin, dword const n);

//*XDSP*S
static void Copy_V_VU(float *pout, float const *pin, dword const n)
{
	/*
	if (!((dword)pin & 8))
	{
		// fpu copy is faster for non-8byte aligned addressess on P3
		FPUCopy_V_V(pout, pin,  n);
		return;
	}
	*/
	dword c = n / 16;
	if (c != 0)
	{
		__asm
		{
			
			mov eax, c
			mov esi, pin
			mov edi, pout
				
lop:
			movups xmm0, [esi]
			movaps [edi], xmm0
				
			movups xmm0, [esi+16]
			movaps [edi+16], xmm0
				
			movups xmm0, [esi+32]
			movaps [edi+32], xmm0
				
			movups xmm0, [esi+48]
			movaps [edi+48], xmm0
				
			add esi, 64
			add edi, 64
				
			dec eax
			jnz lop
				
		}
	}
	
	
	dword r = n & 15;
	if (r > 0)
		memcpy(pout + c * 16, pin + c * 16, r * sizeof(float));
}

//*XDSP*S
static void Add_V_V(float *pout, float const *pin, dword const n)
{
	dword i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
		_mm_store_ps(pout+i, _mm_add_ps(_mm_load_ps(pout+i), _mm_load_ps(pin+i)));
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			_mm_store_ss(pout+i, _mm_add_ss(_mm_load_ss(pout+i), _mm_load_ss(pin+i)));
	}
	
}

//*XDSP*M
static void Add_VI_VI(int *_pout, int const *_pin, dword const n)
{
	dword i = 0;
	__m64 *pout = (__m64 *)_pout;
	__m64 *pin = (__m64 *)_pin;
	
	for (dword c = 0; c < n/4; c++, i += 2)
	{
		pout[i] = _mm_add_pi32(pout[i], pin[i]);
		pout[i+1] = _mm_add_pi32(pout[i+1], pin[i+1]);
	}
	
	_mm_empty();

	i *= 2;

	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			_pout[i] += _pin[i];
	}
	
}

//*XDSP*M
static void Add_VS_VS(short *_pout, short const *_pin, dword const n)
{
	dword i = 0;
	__m64 *pout = (__m64 *)_pout;
	__m64 *pin = (__m64 *)_pin;
	
	for (dword c = 0; c < n/8; c++, i += 2)
	{
		pout[i] = _mm_add_pi16(pout[i], pin[i]);
		pout[i+1] = _mm_add_pi16(pout[i+1], pin[i+1]);
	}
	
	_mm_empty();
	
	i *= 4;
	
	dword r = n & 7;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			_pout[i] += _pin[i];
	}
	
}

//*XDSP*M
static void Add_VI_SI(int *_pout, dword const n, int const a)
{
	dword i = 0;
	__m64 *pout = (__m64 *)_pout;
	__m64 const a2 = _mm_set_pi32(a, a);

	for (dword c = 0; c < n/4; c++, i += 2)
	{
		pout[i] = _mm_add_pi32(pout[i], a2);
		pout[i+1] = _mm_add_pi32(pout[i+1], a2);
	}
	
	_mm_empty();

	i *= 2;

	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			_pout[i] += a;
	}
	
}


//*XDSP*S
static void Sub_V_V(float *pout, float const *pin, dword const n)
{
	dword i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
		_mm_store_ps(pout+i, _mm_sub_ps(_mm_load_ps(pout+i), _mm_load_ps(pin+i)));
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			_mm_store_ss(pout+i, _mm_sub_ss(_mm_load_ss(pout+i), _mm_load_ss(pin+i)));
	}
	
}

//*XDSP*S
static void Sub_V_V_V(float *pout, float const *pina, float const *pinb, dword const n)
{
	dword i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
		_mm_store_ps(pout+i, _mm_sub_ps(_mm_load_ps(pina+i), _mm_load_ps(pinb+i)));
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			_mm_store_ss(pout+i, _mm_sub_ss(_mm_load_ss(pina+i), _mm_load_ss(pinb+i)));
	}
	
}


//*XDSP*M
static void Sub_VI_VI(int *_pout, int const *_pin, dword const n)
{
	dword i = 0;
	__m64 *pout = (__m64 *)_pout;
	__m64 *pin = (__m64 *)_pin;
	
	for (dword c = 0; c < n/4; c++, i += 2)
	{
		pout[i] = _mm_sub_pi32(pout[i], pin[i]);
		pout[i+1] = _mm_sub_pi32(pout[i+1], pin[i+1]);
	}
	
	_mm_empty();
	
	i *= 2;
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			_pout[i] -= _pin[i];
	}
	
}

//*XDSP*M
static void Sub_VS_VS(short *_pout, short const *_pin, dword const n)
{
	dword i = 0;
	__m64 *pout = (__m64 *)_pout;
	__m64 *pin = (__m64 *)_pin;
	
	for (dword c = 0; c < n/8; c++, i += 2)
	{
		pout[i] = _mm_sub_pi16(pout[i], pin[i]);
		pout[i+1] = _mm_sub_pi16(pout[i+1], pin[i+1]);
	}
	
	_mm_empty();
	
	i *= 4;
	
	dword r = n & 7;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			_pout[i] -= _pin[i];
	}
	
}

//*XDSP*S
static void Mul_V_V(float *pout, float const *pin, dword const n)
{
	dword i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
		_mm_store_ps(pout+i, _mm_mul_ps(_mm_load_ps(pout+i), _mm_load_ps(pin+i)));
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			_mm_store_ss(pout+i, _mm_mul_ss(_mm_load_ss(pout+i), _mm_load_ss(pin+i)));
	}
	
}


//*XDSP*S
static void Mul_V_S(float *p, dword const n, float const a)
{
	__m128 const a4 = _mm_set_ps1(a);
	
	for (dword c = 0; c < n/4; c++, p += 4)
		_mm_store_ps(p, _mm_mul_ps(_mm_load_ps(p), a4));

	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, p++)
			_mm_store_ss(p, _mm_mul_ss(_mm_load_ss(p), a4));
	}
}

//*XDSP*S
static void Add_V_S(float *p, dword const n, float const a)
{
	__m128 const a4 = _mm_set_ps1(a);
	
	for (dword c = 0; c < n/4; c++, p += 4)
		_mm_store_ps(p, _mm_add_ps(_mm_load_ps(p), a4));

	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, p++)
			_mm_store_ss(p, _mm_add_ss(_mm_load_ss(p), a4));
	}
}

//*XDSP*S
static void Mul_V2_S2(float *p, dword const n, float const la, float const ra)
{
	__m128 const a4 = _mm_set_ps(ra, la, ra, la);
	
	for (dword c = 0; c < n/2; c++, p += 4)
		_mm_store_ps(p, _mm_mul_ps(_mm_load_ps(p), a4));
	
	if (n & 1)
		_mm_store_ss(p, _mm_mul_ss(_mm_load_ss(p), a4));
}

//*XDSP*S
static void Mul_V4_S4(float *p, dword const n, float const a0, float const a1, float const a2, float const a3)
{
	__m128 const a4 = _mm_set_ps(a3, a2, a1, a0);
	
	for (dword c = 0; c < n; c++, p += 4)
		_mm_store_ps(p, _mm_mul_ps(_mm_load_ps(p), a4));
	
}



//*XDSP*S
static void Mul_V_L(float *ps, dword const n, float &refa, float const da)
{
	__m128 a4 = _mm_set_ss(refa);
	__m128 da4 = _mm_set_ps1(da);
	
	{
		__m128 t = _mm_add_ss(a4, da4);
		a4 = _mm_unpacklo_ps(a4, t);
		a4 = _mm_movelh_ps(a4, _mm_add_ps(a4, _mm_mul_ps(da4, _mm_set_ps1(2))));
	}
	
	da4 = _mm_mul_ps(da4, _mm_set_ps1(4));
	
	dword i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		_mm_store_ps(ps+i, _mm_mul_ps(_mm_load_ps(ps+i), a4));
		a4 = _mm_add_ps(a4, da4);
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		da4 = _mm_set_ps1(da);
		
		for (dword c = 0; c < r; c++, i++)
		{
			_mm_store_ss(ps+i, _mm_mul_ss(_mm_load_ss(ps+i), a4));
			a4 = _mm_add_ss(a4, da4);
		}
		
	}

	_mm_store_ss(&refa, a4);
}

//*XDSP*S
static void Copy_V_L(float *ps, dword const n, float &refa, float const da)
{
	__m128 a4 = _mm_set_ss(refa);
	__m128 da4 = _mm_set_ps1(da);
	
	{
		__m128 t = _mm_add_ss(a4, da4);
		a4 = _mm_unpacklo_ps(a4, t);
		a4 = _mm_movelh_ps(a4, _mm_add_ps(a4, _mm_mul_ps(da4, _mm_set_ps1(2))));
	}
	
	da4 = _mm_mul_ps(da4, _mm_set_ps1(4));
	
	dword i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		_mm_store_ps(ps+i, a4);
		a4 = _mm_add_ps(a4, da4);
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		da4 = _mm_set_ps1(da);
		
		for (dword c = 0; c < r; c++, i++)
		{
			_mm_store_ss(ps+i, a4);
			a4 = _mm_add_ss(a4, da4);
		}
		
	}

	_mm_store_ss(&refa, a4);
}

//*XDSP*S
static void Add_V_L(float *ps, dword const n, float &refa, float const da)
{
	__m128 a4 = _mm_set_ss(refa);
	__m128 da4 = _mm_set_ps1(da);
	
	{
		__m128 t = _mm_add_ss(a4, da4);
		a4 = _mm_unpacklo_ps(a4, t);
		a4 = _mm_movelh_ps(a4, _mm_add_ps(a4, _mm_mul_ps(da4, _mm_set_ps1(2))));
	}
	
	da4 = _mm_mul_ps(da4, _mm_set_ps1(4));
	
	dword i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		_mm_store_ps(ps+i, _mm_add_ps(_mm_load_ps(ps+i), a4));
		a4 = _mm_add_ps(a4, da4);
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		da4 = _mm_set_ps1(da);
		
		for (dword c = 0; c < r; c++, i++)
		{
			_mm_store_ss(ps+i, _mm_add_ss(_mm_load_ss(ps+i), a4));
			a4 = _mm_add_ss(a4, da4);
		}
		
	}

	_mm_store_ss(&refa, a4);
}


//*XDSP*S
static void Mul_V_E(float *ps, dword const n, float &refa, float const da)
{
	__m128 a4 = _mm_set_ss(refa);
	__m128 da4 = _mm_set_ps1(da);
	
	__m128 da42 = _mm_mul_ps(da4, da4);
	
	{
		__m128 t = _mm_mul_ss(a4, da4);
		a4 = _mm_unpacklo_ps(a4, t);
		a4 = _mm_movelh_ps(a4, _mm_mul_ps(a4, da42));
	}
	
	da4 = _mm_mul_ps(da42, da42);
	
	dword i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		_mm_store_ps(ps+i, _mm_mul_ps(_mm_load_ps(ps+i), a4));
		a4 = _mm_mul_ps(a4, da4);
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		da4 = _mm_set_ps1(da);
		
		for (dword c = 0; c < r; c++, i++)
		{
			_mm_store_ss(ps+i, _mm_mul_ss(_mm_load_ss(ps+i), a4));
			a4 = _mm_mul_ss(a4, da4);
		}
		
	}
	
	_mm_store_ss(&refa, a4);
	
}

//*XDSP*S
static void Copy_V_E(float *ps, dword const n, float &refa, float const da)
{
	__m128 a4 = _mm_set_ss(refa);
	__m128 da4 = _mm_set_ps1(da);
	
	__m128 da42 = _mm_mul_ps(da4, da4);
	
	{
		__m128 t = _mm_mul_ss(a4, da4);
		a4 = _mm_unpacklo_ps(a4, t);
		a4 = _mm_movelh_ps(a4, _mm_mul_ps(a4, da42));
	}
	
	da4 = _mm_mul_ps(da42, da42);
	
	dword i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		_mm_store_ps(ps+i, a4);
		a4 = _mm_mul_ps(a4, da4);
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		da4 = _mm_set_ps1(da);
		
		for (dword c = 0; c < r; c++, i++)
		{
			_mm_store_ss(ps+i, a4);
			a4 = _mm_mul_ss(a4, da4);
		}
		
	}
	
	_mm_store_ss(&refa, a4);
	
}

//*XDSP*S
static void MulCopy_V_V_S(float *pout, float const *pin, dword const n, float const a)
{
	__m128 const a4 = _mm_set_ps1(a);
	
	dword i = 0;

	for (dword c = 0; c < n/4; c++, i += 4)
		_mm_store_ps(pout+i, _mm_mul_ps(_mm_load_ps(pin+i), a4));
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			_mm_store_ss(pout+i, _mm_mul_ss(_mm_load_ss(pin+i), a4));
	}
	
}

//*XDSP*S
static void MulCopy_V2_V2_S2(float *pout, float const *pin, dword const n, float const la, float const ra)
{
	__m128 const a4 = _mm_set_ps(ra, la, ra, la);
	
	dword i = 0;
	
	for (dword c = 0; c < n/2; c++, i += 4)
		_mm_store_ps(pout+i, _mm_mul_ps(_mm_load_ps(pin+i), a4));
	
	if (n & 1)
		_mm_store_ss(pout+i, _mm_mul_ss(_mm_load_ss(pin+i), a4));
	
}

//*XDSP*S
static void MulCopy_V2_V_S2(float *pout, float const *pin, dword const n, float const la, float const ra)
{
	__m128 const a4 = _mm_set_ps(ra, la, ra, la);
	
	dword i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		__m128 const in = _mm_load_ps(pin+i);
		__m128 low = _mm_shuffle_ps(in, in, _MM_SHUFFLE(1, 1, 0, 0));
		__m128 high = _mm_shuffle_ps(in, in, _MM_SHUFFLE(3, 3, 2, 2));
		
		_mm_store_ps(pout+2*i+0, _mm_mul_ps(low, a4));
		_mm_store_ps(pout+2*i+4, _mm_mul_ps(high, a4));
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		__m128 const _la = _mm_set_ss(la);
		__m128 const _ra = _mm_set_ss(ra);
		
		for (dword c = 0; c < r; c++, i++)
		{
			__m128 const in = _mm_load_ss(pin+i);
			_mm_store_ss(pout+2*i+0, _mm_mul_ss(in, _la));
			_mm_store_ss(pout+2*i+1, _mm_mul_ss(in, _ra));
			
		}
	}
	
}



//*XDSP*S
static void MulCopy_V_V_L(float *pout, float const *pin, dword const n, float a, float const da)
{
	__m128 a4 = _mm_set_ss(a);
	__m128 da4 = _mm_set_ps1(da);
	
	{
		__m128 t = _mm_add_ss(a4, da4);
		a4 = _mm_unpacklo_ps(a4, t);

		a4 = _mm_movelh_ps(a4, _mm_add_ps(a4, _mm_mul_ps(da4, _mm_set_ps1(2))));

	}
	
	da4 = _mm_mul_ps(da4, _mm_set_ps1(4));

	dword i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		_mm_store_ps(pout+i, _mm_mul_ps(_mm_load_ps(pin+i), a4));
		a4 = _mm_add_ps(a4, da4);
	}
		
	dword r = n & 3;
	if (r > 0)
	{
		da4 = _mm_set_ps1(da);

		for (dword c = 0; c < r; c++, i++)
		{
			_mm_store_ss(pout+i, _mm_mul_ss(_mm_load_ss(pin+i), a4));
			a4 = _mm_add_ss(a4, da4);
		}

	}
	
}

//*XDSP*S
static void MulCopy_V2_V2_L2(float *pout, float const *pin, dword const n, float la, float ra, float const dla, float const dra)
{
	__m128 a4 = _mm_set_ps(ra+dra, la+dla, ra, la);
	__m128 da4 = _mm_set_ps(dra, dla, dra, dla);
	
	da4 = _mm_mul_ps(da4, _mm_set_ps1(2));
	
	dword i = 0;
	
	for (dword c = 0; c < n/2; c++, i += 4)
	{
		_mm_store_ps(pout+i, _mm_mul_ps(_mm_load_ps(pin+i), a4));
		a4 = _mm_add_ps(a4, da4);
	}
	
	if (n & 1)
		_mm_store_ss(pout+i, _mm_mul_ss(_mm_load_ss(pin+i), a4));
	
}


//*XDSP*S
static void MulAdd_V_V_S(float *pout, float const *pin, dword const n, float const a)
{
	__m128 const a4 = _mm_set_ps1(a);
	
	dword i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
		_mm_store_ps(pout+i, _mm_add_ps(_mm_load_ps(pout+i), _mm_mul_ps(_mm_load_ps(pin+i), a4)));
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			_mm_store_ss(pout+i, _mm_add_ss(_mm_load_ss(pout+i), _mm_mul_ss(_mm_load_ss(pin+i), a4)));
	}
	
}

//*XDSP*S
static void MulAdd_V2_V2_S2(float *pout, float const *pin, dword const n, float const la, float const ra)
{
	__m128 const a4 = _mm_set_ps(ra, la, ra, la);
	
	dword i = 0;
	
	for (dword c = 0; c < n/2; c++, i += 4)
		_mm_store_ps(pout+i, _mm_add_ps(_mm_load_ps(pout+i), _mm_mul_ps(_mm_load_ps(pin+i), a4)));
	
	if (n & 1)
		_mm_store_ss(pout+i, _mm_add_ss(_mm_load_ss(pout+i), _mm_mul_ss(_mm_load_ss(pin+i), a4)));
	
}


//*XDSP*S
static void MulAdd_V2_V_S2(float *pout, float const *pin, dword const n, float const la, float const ra)
{
	__m128 const a4 = _mm_set_ps(ra, la, ra, la);
	
	dword i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		__m128 const in = _mm_load_ps(pin+i);
		__m128 low = _mm_shuffle_ps(in, in, _MM_SHUFFLE(1, 1, 0, 0));
		__m128 high = _mm_shuffle_ps(in, in, _MM_SHUFFLE(3, 3, 2, 2));
		
		_mm_store_ps(pout+2*i+0, _mm_add_ps(_mm_load_ps(pout+2*i+0), _mm_mul_ps(low, a4)));
		_mm_store_ps(pout+2*i+4, _mm_add_ps(_mm_load_ps(pout+2*i+4), _mm_mul_ps(high, a4)));
	}

	dword r = n & 3;
	if (r > 0)
	{
		__m128 const _la = _mm_set_ss(la);
		__m128 const _ra = _mm_set_ss(ra);
		
		for (dword c = 0; c < r; c++, i++)
		{
			__m128 const in = _mm_load_ss(pin+i);
			_mm_store_ss(pout+2*i+0, _mm_add_ss(_mm_load_ss(pout+2*i+0), _mm_mul_ss(in, _la)));
			_mm_store_ss(pout+2*i+1, _mm_add_ss(_mm_load_ss(pout+2*i+1), _mm_mul_ss(in, _ra)));

		}
	}

}



//*XDSP*S
static void MulAdd_V_V_L(float *pout, float const *pin, dword const n, float a, float const da)
{
	__m128 a4 = _mm_set_ss(a);
	__m128 da4 = _mm_set_ps1(da);
	
	{
		__m128 t = _mm_add_ss(a4, da4);
		a4 = _mm_unpacklo_ps(a4, t);
		a4 = _mm_movelh_ps(a4, _mm_add_ps(a4, _mm_mul_ps(da4, _mm_set_ps1(2))));
	}
	
	da4 = _mm_mul_ps(da4, _mm_set_ps1(4));
	
	dword i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		_mm_store_ps(pout+i, _mm_add_ps(_mm_load_ps(pout+i), _mm_mul_ps(_mm_load_ps(pin+i), a4)));
		a4 = _mm_add_ps(a4, da4);
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		da4 = _mm_set_ps1(da);
		
		for (dword c = 0; c < r; c++, i++)
		{
			_mm_store_ss(pout+i, _mm_add_ss(_mm_load_ss(pout+i), _mm_mul_ss(_mm_load_ss(pin+i), a4)));
			a4 = _mm_add_ss(a4, da4);
		}
		
	}
	
}

//*XDSP*S
static void MulAdd_V2_V2_L2(float *pout, float const *pin, dword const n, float la, float ra, float const dla, float const dra)
{
	__m128 a4 = _mm_set_ps(ra+dra, la+dla, ra, la);
	__m128 da4 = _mm_set_ps(dra, dla, dra, dla);
	
	da4 = _mm_mul_ps(da4, _mm_set_ps1(2));
	
	dword i = 0;
	
	for (dword c = 0; c < n/2; c++, i += 4)
	{
		_mm_store_ps(pout+i, _mm_add_ps(_mm_load_ps(pout+i), _mm_mul_ps(_mm_load_ps(pin+i), a4)));
		a4 = _mm_add_ps(a4, da4);
	}
	
	if (n & 1)
		_mm_store_ss(pout+i, _mm_add_ps(_mm_load_ps(pout+i), _mm_mul_ss(_mm_load_ss(pin+i), a4)));
	
}



//*XDSP*SM
static void Copy_V_VS(float *pout, short const *pin, dword const n)
{

	
	dword x = n / 4;

	if (x > 0)
	{
		__asm
		{
			mov edx, pin
			mov edi, pout
			mov ecx, x

			pxor mm2, mm2
			xorps xmm1, xmm1
	lop:
			movq mm0, [edx]
			movq mm1, mm0
			movq mm3, mm2
			pcmpgtw mm3, mm1
			movq mm1, mm3
			movq mm4, mm0
			punpckhwd mm4, mm3
			cvtpi2ps xmm0, mm4
			punpcklwd mm0, mm1
			movaps xmm2, xmm1
			movlhps xmm2, xmm0
			cvtpi2ps xmm2, mm0
			add edx, 8
			movaps [edi], xmm2
			add edi, 16

			dec ecx
			jnz lop
			
			emms


		}
	}
	
	dword y = n & 3;

	if (y > 0)
	{
		for (dword c = 0; c < y; c++)
			pout[x*4+c] = pin[x*4+c];
	}

}


//*XDSP*SM
static void Copy_V_VI(float *pout, int const *pin, dword const n)
{
	dword x = n / 4;


	if (x > 0)
	{
		__asm
		{
			mov edx, pin
			mov edi, pout
			mov ecx, x
				
	lop:
			movq mm0, [edx]
			cvtpi2ps xmm0, mm0
			movq mm1, [edx+8]
			cvtpi2ps xmm1, mm1
			movlhps xmm0, xmm1
			add edx, 16
			movaps [edi], xmm0
			add edi, 16
			dec ecx
			jnz lop
				
			emms
				
				
		}
	}
	
	dword y = n & 3;
	
	if (y > 0)
	{
		for (dword c = 0; c < y; c++)
			pout[x*4+c] = (float)pin[x*4+c];
	}
	
}

//*XDSP*SM
static void Copy_V_VIP(float *p, dword const n)
{
	dword x = n / 4;
	
	if (x > 0)
	{
		__asm
		{
				pushad
				mov edx, p
				mov ecx, x
				
	lop:
				movq mm0, [edx]
				cvtpi2ps xmm0, mm0  
				movq mm1, [edx+8]
				cvtpi2ps xmm1, mm1
				movlhps xmm0, xmm1
				movaps [edx], xmm0
				add edx, 16
				dec ecx
				jnz lop
				 
				emms
				
				popad
				
		}
	}
	
	dword y = n & 3;
	
	if (y > 0)
	{
		for (dword c = 0; c < y; c++)
			p[x*4+c] = (float)((int *)p)[x*4+c];
	}
	
}



//*XDSP*SM
static void ClipCopy_VS_V(short *pout, float const *pin, dword const n)
{
	__m128 const mini = _mm_set_ps1(_SHORT_MIN);
	__m128 const maxi = _mm_set_ps1(_SHORT_MAX);
	
	dword i = 0;
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		__m128 x = _mm_load_ps(pin+i);
		
		x = _mm_min_ps(x, maxi);
		x = _mm_max_ps(x, mini);

		__m64 const low = _mm_cvttps_pi32(x);
		__m64 const high = _mm_cvttps_pi32(_mm_movehl_ps(x, x));
		
		__m64 const comb = _mm_packs_pi32(low, high);

		*(__m64 *)(pout+i) = comb;
	}

	_mm_empty();

	dword y = n & 3;
	
	if (y > 0)
	{
		for (dword c = 0; c < y; c++)
		{
			float const a = pin[i+c];
			if (a < _SHORT_MIN)
				pout[i+c] = _SHORT_MINI;
			else if (a > _SHORT_MAX)
				pout[i+c] = _SHORT_MAXI;
			else
				pout[i+c] = (short)(int)a;
		}
	}
	
}

//*XDSP*SM
static void Copy_VS_V(short *pout, float const *pin, dword const n)
{
	dword i = 0;
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		__m128 x = _mm_load_ps(pin+i);
		
		__m64 const low = _mm_cvttps_pi32(x);
		__m64 const high = _mm_cvttps_pi32(_mm_movehl_ps(x, x));
		
		__m64 const comb = _mm_packs_pi32(low, high);
		
		*(__m64 *)(pout+i) = comb;
	}
	
	_mm_empty();
	
	dword y = n & 3;
	
	if (y > 0)
	{
		for (dword c = 0; c < y; c++)
		{
			float const a = pin[i+c];
			pout[i+c] = (short)(int)a;
		}
	}
	
}


//*XDSP*SM
static void ClipCopy_VI_V(int *pout, float const *pin, dword const n)
{
	__m128 const mini = _mm_set_ps1((float)_INT_MIN);
	__m128 const maxi = _mm_set_ps1((float)_INT_MAX);
	
	dword i = 0;
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		__m128 x = _mm_load_ps(pin+i);
		
		x = _mm_min_ps(x, maxi);
		x = _mm_max_ps(x, mini);

		__m64 const low = _mm_cvttps_pi32(x);
		__m64 const high = _mm_cvttps_pi32(_mm_movehl_ps(x, x));
		
		*(__m64 *)(pout+i) = low;
		*(__m64 *)(pout+i+2) = high;
	}
	
	_mm_empty();
	
	dword y = n & 3;
	
	if (y > 0)
	{
		for (dword c = 0; c < y; c++)
		{
			float const a = pin[i+c];
			if (a < _INT_MIN)
				pout[i+c] = _INT_MINI;
			else if (a > _INT_MAX)
				pout[i+c] = _INT_MAXI;
			else
				pout[i+c] = (int)a;
		}
	}
	
}

//*XDSP*SM
static void Copy_VI_V(int *pout, float const *pin, dword const n)
{
	dword i = 0;
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		__m128 x = _mm_load_ps(pin+i);
		
		__m64 const low = _mm_cvttps_pi32(x);
		__m64 const high = _mm_cvttps_pi32(_mm_movehl_ps(x, x));
		
		*(__m64 *)(pout+i) = low;
		*(__m64 *)(pout+i+2) = high;
	}
	
	_mm_empty();
	
	dword y = n & 3;
	
	if (y > 0)
	{
		for (dword c = 0; c < y; c++)
		{
			float const a = pin[i+c];
			pout[i+c] = (int)a;
		}
	}
	
}

//*XDSP*SM
static void ClipCopy_VI_VP(int *p, dword const n)
{
	__m128 const mini = _mm_set_ps1((float)_INT_MIN);
	__m128 const maxi = _mm_set_ps1((float)_INT_MAX);
	
	dword i = 0;
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		__m128 x = _mm_load_ps((float *)p+i);
		
		x = _mm_min_ps(x, maxi);
		x = _mm_max_ps(x, mini);
		
		__m64 const low = _mm_cvttps_pi32(x);
		__m64 const high = _mm_cvttps_pi32(_mm_movehl_ps(x, x));
		
		*(__m64 *)(p+i) = low;
		*(__m64 *)(p+i+2) = high;
	}
	
	_mm_empty();
	
	dword y = n & 3;
	
	if (y > 0)
	{
		for (dword c = 0; c < y; c++)
		{
			float const a = ((float *)p)[i+c];
			if (a < _INT_MIN)
				p[i+c] = _INT_MINI;
			else if (a > _INT_MAX)
				p[i+c] = _INT_MAXI;
			else
				p[i+c] = (int)a;
		}
	}
	
}

//*XDSP*SM
static void Copy_VI_VP(int *p, dword const n)
{

	dword i = 0;
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		__m128 x = _mm_load_ps((float *)p+i);
		
		__m64 const low = _mm_cvttps_pi32(x);
		__m64 const high = _mm_cvttps_pi32(_mm_movehl_ps(x, x));
		
		*(__m64 *)(p+i) = low;
		*(__m64 *)(p+i+2) = high;
	}
	
	_mm_empty();
	
	dword y = n & 3;
	
	if (y > 0)
	{
		for (dword c = 0; c < y; c++)
		{
			float const a = ((float *)p)[i+c];
			p[i+c] = (int)a;
		}
	}
	
}

//*XDSP*SM
static void CopyShift_VI_VP(int *p, dword const n, dword const shift)
{
	__m64 const pshift = _mm_set_pi32(shift, shift);

	dword i = 0;
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		__m128 x = _mm_load_ps((float *)p+i);
		
		__m64 low = _mm_cvttps_pi32(x);
		__m64 high = _mm_cvttps_pi32(_mm_movehl_ps(x, x));
		
		low = _m_pslld(low, pshift);
		high = _m_pslld(high, pshift);
		
		*(__m64 *)(p+i) = low;
		*(__m64 *)(p+i+2) = high;
	}
	
	_mm_empty();
	
	dword y = n & 3;
	
	if (y > 0)
	{
		for (dword c = 0; c < y; c++)
		{
			float const a = ((float *)p)[i+c];
			p[i+c] = (int)a << shift;
		}
	}
	
}


//*XDSP*S
static bool AbsMaxGT_V_S(float const *pin, dword const n, float const x)
{
	dword i = 0;
	
	__m128 const x4 = _mm_set_ps1(x);

	static dword const am = 0x7fffffff;
	static __m128 const absmask = _mm_load_ps1((float const *)&am);

	for (dword c = 0; c < n/8; c++, i += 8)
	{
		__m128 a4 = _mm_max_ps(_mm_and_ps(_mm_load_ps(pin+i), absmask), _mm_and_ps(_mm_load_ps(pin+i+4), absmask));
		int z = _mm_movemask_ps(_mm_cmpgt_ps(a4, x4));
		if (z)
			return true;
	}

	dword r = n & 7;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			if (pin[i] > x)
				return true;
	}


	
	return false;
}

//*XDSP*S
static float AbsMax_V(float const *pin, dword const n)
{

	dword i = 0;
	
	__m128 mini = _mm_setzero_ps();
	__m128 maxi = _mm_setzero_ps();
	
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		__m128 x = _mm_load_ps(pin+i);
		mini = _mm_min_ps(mini, x);
		maxi = _mm_max_ps(maxi, x);
	}

	maxi = _mm_max_ps(maxi, _mm_mul_ps(mini, _mm_set_ps1(-1)));

	float m = maxi.m128_f32[0];
	m = __max(m, maxi.m128_f32[1]);
	m = __max(m, maxi.m128_f32[2]);
	m = __max(m, maxi.m128_f32[3]);
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			m = (float)__max(m, fabs(pin[i]));
		
	}
	
	
	return false;
}

//*XDSP*S
static void And_VI_SI(int *p, dword const n, int const x)
{
	__m128 const a4 = _mm_load_ps1((float *)&x);
	
	for (dword c = 0; c < n/4; c++, p += 4)
		_mm_store_ps((float *)p, _mm_and_ps(_mm_load_ps((float *)p), a4));
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, p++)
			_mm_store_ss((float *)p, _mm_and_ps(_mm_load_ss((float *)p), a4));
	}
}

	  

//*XDSP*S
static void Or_VI_SI(int *p, dword const n, int const x)
{
	__m128 const a4 = _mm_load_ps1((float *)&x);
	
	for (dword c = 0; c < n/4; c++, p += 4)
		_mm_store_ps((float *)p, _mm_or_ps(_mm_load_ps((float *)p), a4));
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, p++)
			_mm_store_ss((float *)p, _mm_or_ps(_mm_load_ss((float *)p), a4));
	}
}

//*XDSP*S
static void Xor_VI_SI(int *p, dword const n, int const x)
{
	__m128 const a4 = _mm_load_ps1((float *)&x);
	
	for (dword c = 0; c < n/4; c++, p += 4)
		_mm_store_ps((float *)p, _mm_xor_ps(_mm_load_ps((float *)p), a4));
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, p++)
			_mm_store_ss((float *)p, _mm_xor_ps(_mm_load_ss((float *)p), a4));
	}
}

//*XDSP*S
static void And_VI_VI(int *pout, int const *pin, dword const n)
{
	dword i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
		_mm_store_ps((float *)pout+i, _mm_and_ps(_mm_load_ps((float *)pout+i), _mm_load_ps((float const *)pin+i)));
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			_mm_store_ss((float *)pout+i, _mm_and_ps(_mm_load_ss((float *)pout+i), _mm_load_ss((float const *)pin+i)));
	}
	
}

//*XDSP*S
static void AndNot1_VI_VI(int *pout, int const *pin, dword const n)
{
	dword i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
		_mm_store_ps((float *)pout+i, _mm_andnot_ps(_mm_load_ps((float *)pout+i), _mm_load_ps((float const *)pin+i)));
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			_mm_store_ss((float *)pout+i, _mm_andnot_ps(_mm_load_ss((float *)pout+i), _mm_load_ss((float const *)pin+i)));
	}
	
}

//*XDSP*S
static void AndNot2_VI_VI(int *pout, int const *pin, dword const n)
{
	dword i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
		_mm_store_ps((float *)pout+i, _mm_andnot_ps(_mm_load_ps((float const *)pin+i), _mm_load_ps((float *)pout+i)));
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			_mm_store_ss((float *)pout+i, _mm_andnot_ps(_mm_load_ss((float const *)pin+i), _mm_load_ss((float *)pout+i)));
	}
	
}

//*XDSP*S
static void Or_VI_VI(int *pout, int const *pin, dword const n)
{
	dword i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
		_mm_store_ps((float *)pout+i, _mm_or_ps(_mm_load_ps((float *)pout+i), _mm_load_ps((float const *)pin+i)));
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			_mm_store_ss((float *)pout+i, _mm_or_ps(_mm_load_ss((float *)pout+i), _mm_load_ss((float const *)pin+i)));
	}
	
}

//*XDSP*S
static void Xor_VI_VI(int *pout, int const *pin, dword const n)
{
	dword i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
		_mm_store_ps((float *)pout+i, _mm_xor_ps(_mm_load_ps((float *)pout+i), _mm_load_ps((float const *)pin+i)));
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			_mm_store_ss((float *)pout+i, _mm_xor_ps(_mm_load_ss((float *)pout+i), _mm_load_ss((float const *)pin+i)));
	}
	
}

//*XDSP*M
static void Abs_VI(int *_p, dword const n)
{
	__m64 *p = (__m64 *)_p;

	for (dword c = 0; c < n/4; c++, p += 2)
	{
		__m64 in1 = p[0];
		__m64 in2 = p[1];

		__m64 const negmask1 = _mm_srai_pi32(in1, 31);
		__m64 const negmask2 = _mm_srai_pi32(in2, 31);
		
		in1 = _mm_xor_si64(in1, negmask1);
		in2 = _mm_xor_si64(in2, negmask2);
		
		in1 = _mm_sub_pi32(in1, negmask1);
		in2 = _mm_sub_pi32(in2, negmask2);
		
		p[0] = in1;
		p[1] = in2;
		


	}

	_p = (int *)p;

	_mm_empty();
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++)
			_p[c] = labs(_p[c]);
	}

}


//*XDSP*M
static void Abs_VS(short *_p, dword const n)
{
	__m64 *p = (__m64 *)_p;
	
	for (dword c = 0; c < n/8; c++, p += 2)
	{
		__m64 in1 = p[0];
		__m64 in2 = p[1];
		
		__m64 const negmask1 = _mm_srai_pi16(in1, 15);
		__m64 const negmask2 = _mm_srai_pi16(in2, 15);
		
		in1 = _mm_xor_si64(in1, negmask1);
		in2 = _mm_xor_si64(in2, negmask2);
		
		in1 = _mm_sub_pi16(in1, negmask1);
		in2 = _mm_sub_pi16(in2, negmask2);
		
		p[0] = in1;
		p[1] = in2;
		
	}
	
	_p = (short *)p;
	
	_mm_empty();
	
	dword r = n & 7;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++)
			_p[c] = (short)labs(_p[c]);
	}
	
}

//*XDSP*M
static int AbsMax_VI(int const *_p, dword const n)
{
	__m64 const *p = (__m64 *)_p;
	
	__m64 max1 = _mm_setzero_si64();
	__m64 max2 = _mm_setzero_si64();
	
	for (dword c = 0; c < n/4; c++, p += 2)
	{
		__m64 in1 = p[0];
		__m64 in2 = p[1];
		
		__m64 const negmask1 = _mm_srai_pi32(in1, 31);
		__m64 const negmask2 = _mm_srai_pi32(in2, 31);
		
		in1 = _mm_xor_si64(in1, negmask1);
		in2 = _mm_xor_si64(in2, negmask2);
		
		in1 = _mm_sub_pi32(in1, negmask1);
		in2 = _mm_sub_pi32(in2, negmask2);
		
		__m64 mask1 = _mm_cmpgt_pi32(in1, max1);
		__m64 mask2 = _mm_cmpgt_pi32(in2, max2);
		max1 = _mm_or_si64(_mm_and_si64(mask1, in1), _mm_andnot_si64(mask1, max1));
		max2 = _mm_or_si64(_mm_and_si64(mask2, in2), _mm_andnot_si64(mask2, max2));
	
	}
	
	__m64 mask = _mm_cmpgt_pi32(max1, max2);
	max1 = _mm_or_si64(_mm_and_si64(mask, max1), _mm_andnot_si64(mask, max2));
	
	int maks = __max(((int *)&max1)[0], ((int *)&max1)[1]);

	_p = (int const *)p;
	
	_mm_empty();
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++)
			maks = __max(maks, _p[c]);
	}

	return maks;
}

//*XDSP*M
static short AbsMax_VS(short const *_p, dword const n)
{
	__m64 const *p = (__m64 *)_p;
	
	__m64 max1 = _mm_setzero_si64();
	__m64 max2 = _mm_setzero_si64();
	
	for (dword c = 0; c < n/8; c++, p += 2)
	{
		__m64 in1 = p[0];
		__m64 in2 = p[1];
		
		__m64 const negmask1 = _mm_srai_pi16(in1, 15);
		__m64 const negmask2 = _mm_srai_pi16(in2, 15);
		
		in1 = _mm_xor_si64(in1, negmask1);
		in2 = _mm_xor_si64(in2, negmask2);
		
		in1 = _mm_sub_pi16(in1, negmask1);
		in2 = _mm_sub_pi16(in2, negmask2);
		
		__m64 mask1 = _mm_cmpgt_pi16(in1, max1);
		__m64 mask2 = _mm_cmpgt_pi16(in2, max2);
		max1 = _mm_or_si64(_mm_and_si64(mask1, in1), _mm_andnot_si64(mask1, max1));
		max2 = _mm_or_si64(_mm_and_si64(mask2, in2), _mm_andnot_si64(mask2, max2));
		
	}
	
	__m64 mask = _mm_cmpgt_pi16(max1, max2);
	max1 = _mm_or_si64(_mm_and_si64(mask, max1), _mm_andnot_si64(mask, max2));
	
	short maks = __max(((short *)&max1)[0], ((short *)&max1)[1]);
	maks = __max(maks, ((short *)&max1)[2]);
	maks = __max(maks, ((short *)&max1)[3]);
	
	_p = (short const *)p;
	
	_mm_empty();
	
	dword r = n & 7;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++)
			maks = __max(maks, _p[c]);
	}
	
	return maks;
}


/*
static void AddInterleavedChannels4to1(float *pout, float *pin, dword const n)
{	

	SLOW
  
	dword i = 0;
	for (dword c = 0; c < n/4; c++, i += 4, pin += 16)
	{
		__m128 row0 = _mm_load_ps(pin);
		__m128 row1 = _mm_load_ps(pin+4); 
		__m128 row2 = _mm_load_ps(pin+8);
		__m128 row3 = _mm_load_ps(pin+12);


		__m128 tmp0, tmp1, tmp2, tmp3; 
		tmp0 = _mm_shuffle_ps(row0, row1, 0x44);
		tmp2 = _mm_shuffle_ps(row0, row1, 0xEE);
		tmp1 = _mm_shuffle_ps(row2, row3, 0x44);
		tmp3 = _mm_shuffle_ps(row2, row3, 0xEE);
		
		__m128 sum = _mm_shuffle_ps(tmp0, tmp1, 0x88);
		sum = _mm_add_ps(sum, _mm_shuffle_ps(tmp0, tmp1, 0xDD));
		sum = _mm_add_ps(sum, _mm_shuffle_ps(tmp2, tmp3, 0x88));
		sum = _mm_add_ps(sum, _mm_shuffle_ps(tmp2, tmp3, 0xDD));
 
		_mm_store_ps(pout + i, sum);
	}

	dword y = n & 3;
	
	if (y > 0)
	{
		for (dword c = 0; c < y; c++, pin += 4)
		{
			pout[c+i] = pin[0] + pin[1] + pin[2] + pin[3];
		}
	}
	
}
*/




__declspec(align(32)) float fc0[64];
__declspec(align(32)) float fc1[64];
int str = 4;

static void Resample(float *pout, float const *pin, dword const n)
{
	

	for (dword c = 0; c < n; c++, pin += str)
	{
/*
		__m128 a = _mm_mul_ps(_mm_load_ps(fc), _mm_load_ps(pin));
		a = _mm_add_ps(a, _mm_mul_ps(_mm_load_ps(fc+4), _mm_load_ps(pin+4)));
		a = _mm_add_ps(a, _mm_mul_ps(_mm_load_ps(fc+8), _mm_load_ps(pin+8)));
		a = _mm_add_ps(a, _mm_mul_ps(_mm_load_ps(fc+12), _mm_load_ps(pin+12)));
		a = _mm_add_ps(a, _mm_mul_ps(_mm_load_ps(fc+16), _mm_load_ps(pin+16)));
		a = _mm_add_ps(a, _mm_mul_ps(_mm_load_ps(fc+20), _mm_load_ps(pin+20)));
		a = _mm_add_ps(a, _mm_mul_ps(_mm_load_ps(fc+24), _mm_load_ps(pin+24)));
		a = _mm_add_ps(a, _mm_mul_ps(_mm_load_ps(fc+28), _mm_load_ps(pin+28)));
*/
		
		__m128 t = _mm_load_ps(pin);
		__m128 a = _mm_mul_ps(_mm_load_ps(fc0), t);
		__m128 b = _mm_mul_ps(_mm_load_ps(fc1), t);

		t = _mm_load_ps(pin+4);
		a = _mm_add_ps(a, _mm_mul_ps(_mm_load_ps(fc0+4), t));
		b = _mm_add_ps(b, _mm_mul_ps(_mm_load_ps(fc1+4), t));

		t = _mm_load_ps(pin+8);
		a = _mm_add_ps(a, _mm_mul_ps(_mm_load_ps(fc0+8), t));
		b = _mm_add_ps(b, _mm_mul_ps(_mm_load_ps(fc1+8), t));

		t = _mm_load_ps(pin+12);
		a = _mm_add_ps(a, _mm_mul_ps(_mm_load_ps(fc0+12), t));
		b = _mm_add_ps(b, _mm_mul_ps(_mm_load_ps(fc1+12), t));
		
		t = _mm_load_ps(pin+16);
		a = _mm_add_ps(a, _mm_mul_ps(_mm_load_ps(fc0+16), t));
		b = _mm_add_ps(b, _mm_mul_ps(_mm_load_ps(fc1+16), t));

		t = _mm_load_ps(pin+20);
		a = _mm_add_ps(a, _mm_mul_ps(_mm_load_ps(fc0+20), t));
		b = _mm_add_ps(b, _mm_mul_ps(_mm_load_ps(fc1+20), t));

		t = _mm_load_ps(pin+24);
		a = _mm_add_ps(a, _mm_mul_ps(_mm_load_ps(fc0+24), t));
		b = _mm_add_ps(b, _mm_mul_ps(_mm_load_ps(fc1+24), t));

		t = _mm_load_ps(pin+28);
		a = _mm_add_ps(a, _mm_mul_ps(_mm_load_ps(fc0+28), t));
		b = _mm_add_ps(b, _mm_mul_ps(_mm_load_ps(fc1+28), t));
		
		
		__m128 x = _mm_add_ps(a, _mm_movehl_ps(a, a));
		x = _mm_add_ss(x, _mm_shuffle_ps(x, x, _MM_SHUFFLE(3, 2, 1, 1)));

		__m128 y = _mm_add_ps(b, _mm_movehl_ps(b, b));
		y = _mm_add_ss(y, _mm_shuffle_ps(y, y, _MM_SHUFFLE(3, 2, 1, 1)));
		
		_mm_store_ss(pout+c, _mm_add_ps(x, y));
	}
	
}

inline void PreparePackedRamp(float const &c, float const &dc, __m128 &pc, __m128 &pdc)
{
	pc = _mm_load_ps1(&c);
	pdc = _mm_load_ps1(&dc);
	
	__m128 const t = _mm_add_ss(pc, pdc);
	pc = _mm_unpacklo_ps(pc, t);
	pc = _mm_movelh_ps(pc, _mm_add_ps(pc, _mm_mul_ps(pdc, _mm_set_ps1(2))));

	pdc = _mm_mul_ps(pdc, _mm_set_ps1(4));
}


//*XDSP*S
static void EvaluatePolynomial3_V(float *p, dword const n, float *pc, float const *pdc)
{
	if (pdc != NULL)
	{
		__m128 pc0, pc1, pc2;
		__m128 dpc0, dpc1, dpc2;

		PreparePackedRamp(pc[0], pdc[0], pc0, dpc0);
		PreparePackedRamp(pc[1], pdc[1], pc1, dpc1);
		PreparePackedRamp(pc[2], pdc[2], pc2, dpc2);
		
		for (dword c = 0; c < n/4; c++, p += 4)
		{
			__m128 const x = _mm_load_ps(p);
			_mm_store_ps(p, _mm_add_ps(_mm_mul_ps(_mm_add_ps(_mm_mul_ps(pc0, x), pc1), x), pc2));
			pc0 = _mm_add_ps(pc0, dpc0);
			pc1 = _mm_add_ps(pc1, dpc1);
			pc2 = _mm_add_ps(pc2, dpc2);
		}
		
		dword r = n & 3;
		if (r > 0)
		{
			for (dword c = 0; c < r; c++, p++)
			{
				__m128 const x = _mm_load_ss(p);
				_mm_store_ss(p, _mm_add_ss(_mm_mul_ss(_mm_add_ss(_mm_mul_ss(pc0, x), pc1), x), pc2));
				pc0 = _mm_add_ss(pc0, dpc0);
				pc1 = _mm_add_ss(pc1, dpc1);
				pc2 = _mm_add_ss(pc2, dpc2);
			}
		}		
		
		_mm_store_ss(pc+0, pc0);
		_mm_store_ss(pc+1, pc1);
		_mm_store_ss(pc+2, pc2);
	}
	else
	{
		__m128 pc0 = _mm_set_ps1(pc[0]);
		__m128 pc1 = _mm_set_ps1(pc[1]);
		__m128 pc2 = _mm_set_ps1(pc[2]);
		
		for (dword c = 0; c < n/4; c++, p += 4)
		{
			__m128 const x = _mm_load_ps(p);
			_mm_store_ps(p, _mm_add_ps(_mm_mul_ps(_mm_add_ps(_mm_mul_ps(pc0, x), pc1), x), pc2));
		}
		
		dword r = n & 3;
		if (r > 0)
		{
			for (dword c = 0; c < r; c++, p++)
			{
				__m128 const x = _mm_load_ss(p);
				_mm_store_ss(p, _mm_add_ss(_mm_mul_ss(_mm_add_ss(_mm_mul_ss(pc0, x), pc1), x), pc2));
			}
		}		

		_mm_store_ss(pc+0, pc0);
		_mm_store_ss(pc+1, pc1);
		_mm_store_ss(pc+2, pc2);
	}
	
}

//*XDSP*S
static void EvaluatePolynomial4_V(float *p, dword const n, float *pc, float const *pdc)
{
	if (pdc != NULL)
	{
		__m128 pc0, pc1, pc2, pc3;
		__m128 dpc0, dpc1, dpc2, dpc3;
		
		PreparePackedRamp(pc[0], pdc[0], pc0, dpc0);
		PreparePackedRamp(pc[1], pdc[1], pc1, dpc1);
		PreparePackedRamp(pc[2], pdc[2], pc2, dpc2);
		PreparePackedRamp(pc[3], pdc[3], pc3, dpc3);
		
		for (dword c = 0; c < n/4; c++, p += 4)
		{
			__m128 const x = _mm_load_ps(p);
			_mm_store_ps(p, _mm_add_ps(_mm_mul_ps(_mm_add_ps(_mm_mul_ps(_mm_add_ps(_mm_mul_ps(pc0, x), pc1), x), pc2), x), pc3));
			pc0 = _mm_add_ps(pc0, dpc0);
			pc1 = _mm_add_ps(pc1, dpc1);
			pc2 = _mm_add_ps(pc2, dpc2);
			pc3 = _mm_add_ps(pc3, dpc3);
		}
		
		dword r = n & 3;
		if (r > 0)
		{
			for (dword c = 0; c < r; c++, p++)
			{
				__m128 const x = _mm_load_ss(p);
				_mm_store_ss(p, _mm_add_ss(_mm_mul_ss(_mm_add_ss(_mm_mul_ss(_mm_add_ss(_mm_mul_ss(pc0, x), pc1), x), pc2), x), pc3));
				pc0 = _mm_add_ss(pc0, dpc0);
				pc1 = _mm_add_ss(pc1, dpc1);
				pc2 = _mm_add_ss(pc2, dpc2);
				pc3 = _mm_add_ss(pc3, dpc3);
			}
		}		
		
		_mm_store_ss(pc+0, pc0);
		_mm_store_ss(pc+1, pc1);
		_mm_store_ss(pc+2, pc2);
		_mm_store_ss(pc+3, pc3);
	}
	else
	{
		__m128 pc0 = _mm_set_ps1(pc[0]);
		__m128 pc1 = _mm_set_ps1(pc[1]);
		__m128 pc2 = _mm_set_ps1(pc[2]);
		__m128 pc3 = _mm_set_ps1(pc[3]);
		
		for (dword c = 0; c < n/4; c++, p += 4)
		{
			__m128 const x = _mm_load_ps(p);
			_mm_store_ps(p, _mm_add_ps(_mm_mul_ps(_mm_add_ps(_mm_mul_ps(_mm_add_ps(_mm_mul_ps(pc0, x), pc1), x), pc2), x), pc3));
		}
		
		dword r = n & 3;
		if (r > 0)
		{
			for (dword c = 0; c < r; c++, p++)
			{
				__m128 const x = _mm_load_ss(p);
				_mm_store_ss(p, _mm_add_ss(_mm_mul_ss(_mm_add_ss(_mm_mul_ss(_mm_add_ss(_mm_mul_ss(pc0, x), pc1), x), pc2), x), pc3));
			}
		}		
		
		_mm_store_ss(pc+0, pc0);
		_mm_store_ss(pc+1, pc1);
		_mm_store_ss(pc+2, pc2);
		_mm_store_ss(pc+3, pc3);
	}
	
}

//*XDSP*M
static void Shl_VI(int *_p, dword const n, dword const shift)
{
	dword i = 0;
	__m64 *p = (__m64 *)_p;
	
	__m64 const sh = _mm_set_pi32(shift, shift);

	for (dword c = 0; c < n/4; c++, i += 2)
	{
		p[i] = _mm_sll_pi32(p[i], sh);
		p[i+1] = _mm_sll_pi32(p[i+1], sh);
	}
	
	_mm_empty();
	
	i *= 2;
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			_p[i] <<= shift;
	}
	
}

//*XDSP*M
static void Shr_VI(int *_p, dword const n, dword const shift)
{
	dword i = 0;
	__m64 *p = (__m64 *)_p;
	
	__m64 const sh = _mm_set_pi32(shift, shift);
	
	for (dword c = 0; c < n/4; c++, i += 2)
	{
		p[i] = _mm_srl_pi32(p[i], sh);
		p[i+1] = _mm_srl_pi32(p[i+1], sh);
	}
	
	_mm_empty();
	
	i *= 2;
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			((dword *)_p)[i] >>= shift;
	}
	
}

//*XDSP*M
static void Sar_VI(int *_p, dword const n, dword const shift)
{
	dword i = 0;
	__m64 *p = (__m64 *)_p;
	
	__m64 const sh = _mm_set_pi32(shift, shift);
	
	for (dword c = 0; c < n/4; c++, i += 2)
	{
		p[i] = _mm_sra_pi32(p[i], sh);
		p[i+1] = _mm_sra_pi32(p[i+1], sh);
	}
	
	_mm_empty();
	
	i *= 2;
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			_p[i] >>= shift;
	}
	
}

//*XDSP*S
static void Div_V_S(float *p, dword const n, float const a)
{
	__m128 const a4 = _mm_set_ps1(a);

	for (dword c = 0; c < n/4; c++, p += 4)
		_mm_store_ps(p, _mm_div_ps(_mm_load_ps(p), a4));
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, p++)
			_mm_store_ss(p, _mm_div_ss(_mm_load_ss(p), a4));
	}
}

//*XDSP*S
static void Div_V_V(float *pout, float const *pin, dword const n)
{
	dword i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
		_mm_store_ps(pout+i, _mm_div_ps(_mm_load_ps(pout+i), _mm_load_ps(pin+i)));
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			_mm_store_ss(pout+i, _mm_div_ss(_mm_load_ss(pout+i), _mm_load_ss(pin+i)));
	}
	
}

//*XDSP*S
static void ReciprocalApprox_V_V(float *pout, float const *pin, dword const n)
{
	dword i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
		_mm_store_ps(pout+i, _mm_rcp_ps(_mm_load_ps(pin+i)));
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			_mm_store_ss(pout+i, _mm_rcp_ss(_mm_load_ss(pin+i)));
	}
	
}

//*XDSP*S
static void Sqrt_V(float *p, dword const n)
{
	for (dword c = 0; c < n/4; c++, p += 4)
		_mm_store_ps(p, _mm_sqrt_ps(_mm_load_ps(p)));
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, p++)
			_mm_store_ss(p, _mm_sqrt_ss(_mm_load_ss(p)));
	}
}

//*XDSP*S
static void ReciprocalSqrtApprox_V_V(float *pout, float const *pin, dword const n)
{
	dword i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
		_mm_store_ps(pout+i, _mm_rsqrt_ps(_mm_load_ps(pin+i)));
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			_mm_store_ss(pout+i, _mm_rsqrt_ss(_mm_load_ss(pin+i)));
	}
	
}

//*XDSP*S
static void CompareLT_V_V_S(dword *pout, float const *pin, dword const n, float const a)
{
	__m128 const a4 = _mm_set_ps1(a);

	dword i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
		_mm_store_ps((float *)pout+i, _mm_cmplt_ps(_mm_load_ps(pin+i), a4));
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			_mm_store_ss((float *)pout+i, _mm_cmplt_ss(_mm_load_ss(pin+i), a4));
	}
	
}

//*XDSP*S
static void CompareLT_V_V_V(dword *pout, float const *pin1, float const *pin2, dword const n)
{
	dword i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
		_mm_store_ps((float *)pout+i, _mm_cmplt_ps(_mm_load_ps(pin1+i), _mm_load_ps(pin2+i)));
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			_mm_store_ss((float *)pout+i, _mm_cmplt_ss(_mm_load_ss(pin1+i), _mm_load_ss(pin2+i)));
	}
	
}

//*XDSP*S
static void Min_V_S(float *p, dword const n, float const a)
{
	__m128 const a4 = _mm_set_ps1(a);
	
	for (dword c = 0; c < n/4; c++, p += 4)
		_mm_store_ps(p, _mm_min_ps(_mm_load_ps(p), a4));
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, p++)
			_mm_store_ss(p, _mm_min_ss(_mm_load_ss(p), a4));
	}
}

//*XDSP*S
static void Max_V_S(float *p, dword const n, float const a)
{
	__m128 const a4 = _mm_set_ps1(a);
	
	for (dword c = 0; c < n/4; c++, p += 4)
		_mm_store_ps(p, _mm_max_ps(_mm_load_ps(p), a4));
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, p++)
			_mm_store_ss(p, _mm_max_ss(_mm_load_ss(p), a4));
	}
}

//*XDSP*S
static void MinMax_V_S(float *p, dword const n, float const _min, float const _max)
{
	__m128 const min4 = _mm_set_ps1(_min);
	__m128 const max4 = _mm_set_ps1(_max);
	
	for (dword c = 0; c < n/4; c++, p += 4)
		_mm_store_ps(p, _mm_max_ps(_mm_min_ps(_mm_load_ps(p), min4), max4));
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, p++)
			_mm_store_ss(p, _mm_max_ss(_mm_min_ss(_mm_load_ss(p), min4), max4));
	}
}

//*XDSP*S
static void Min_V_V(float *pout, float const *pin, dword const n)
{
	dword i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
		_mm_store_ps(pout+i, _mm_min_ps(_mm_load_ps(pout+i), _mm_load_ps(pin+i)));
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			_mm_store_ss(pout+i, _mm_min_ss(_mm_load_ss(pout+i), _mm_load_ss(pin+i)));
	}
	
}

//*XDSP*S
static void Max_V_V(float *pout, float const *pin, dword const n)
{
	dword i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
		_mm_store_ps(pout+i, _mm_max_ps(_mm_load_ps(pout+i), _mm_load_ps(pin+i)));
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			_mm_store_ss(pout+i, _mm_max_ss(_mm_load_ss(pout+i), _mm_load_ss(pin+i)));
	}
	
}

//*XDSP*S
static void CompareLTAdd_V_S_S(float *p, dword const n, float const a, float const b)
{
	__m128 const a4 = _mm_set_ps1(a);
	__m128 const b4 = _mm_set_ps1(b);
	
	for (dword c = 0; c < n/4; c++, p += 4)
	{
		__m128 const in = _mm_load_ps(p);
		_mm_store_ps(p, _mm_add_ps(in, _mm_and_ps(_mm_cmplt_ps(in, a4), b4)));
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, p++)
		{
			__m128 const in = _mm_load_ss(p);
			_mm_store_ss(p, _mm_add_ss(in, _mm_and_ps(_mm_cmplt_ss(in, a4), b4)));
		}
	}
}

//*XDSP*S
static void CompareGTAdd_V_S_S(float *p, dword const n, float const a, float const b)
{
	__m128 const a4 = _mm_set_ps1(a);
	__m128 const b4 = _mm_set_ps1(b);
	
	for (dword c = 0; c < n/4; c++, p += 4)
	{
		__m128 const in = _mm_load_ps(p);
		_mm_store_ps(p, _mm_add_ps(in, _mm_and_ps(_mm_cmpgt_ps(in, a4), b4)));
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, p++)
		{
			__m128 const in = _mm_load_ss(p);
			_mm_store_ss(p, _mm_add_ss(in, _mm_and_ps(_mm_cmpgt_ss(in, a4), b4)));
		}
	}
}

//*XDSP*S
static void CompareLEAdd_V_S_S(float *p, dword const n, float const a, float const b)
{
	__m128 const a4 = _mm_set_ps1(a);
	__m128 const b4 = _mm_set_ps1(b);
	
	for (dword c = 0; c < n/4; c++, p += 4)
	{
		__m128 const in = _mm_load_ps(p);
		_mm_store_ps(p, _mm_add_ps(in, _mm_and_ps(_mm_cmple_ps(in, a4), b4)));
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, p++)
		{
			__m128 const in = _mm_load_ss(p);
			_mm_store_ss(p, _mm_add_ss(in, _mm_and_ps(_mm_cmple_ss(in, a4), b4)));
		}
	}
}

//*XDSP*S
static void CompareGEAdd_V_S_S(float *p, dword const n, float const a, float const b)
{
	__m128 const a4 = _mm_set_ps1(a);
	__m128 const b4 = _mm_set_ps1(b);
	
	for (dword c = 0; c < n/4; c++, p += 4)
	{
		__m128 const in = _mm_load_ps(p);
		_mm_store_ps(p, _mm_add_ps(in, _mm_and_ps(_mm_cmpge_ps(in, a4), b4)));
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, p++)
		{
			__m128 const in = _mm_load_ss(p);
			_mm_store_ss(p, _mm_add_ss(in, _mm_and_ps(_mm_cmpge_ss(in, a4), b4)));
		}
	}
}


//*XDSP*S
static void AbsCompareLTCopy_V_S_S(float *p, dword const n, float const a, float const b)
{
	__m128 const a4 = _mm_set_ps1(a);
	__m128 const b4 = _mm_set_ps1(b);
	int const imask = 0x7fffffff;
	__m128 const mask = _mm_load_ps1((float const *)&imask);

	for (dword c = 0; c < n/4; c++, p += 4)
	{
		__m128 const in = _mm_load_ps(p);
		__m128 const abs = _mm_and_ps(in, mask);
		__m128 const cm = _mm_cmplt_ps(abs, a4);
		_mm_store_ps(p, MASK_COMBINE(b4, in, cm));
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, p++)
		{
			__m128 const in = _mm_load_ss(p);
			__m128 const abs = _mm_and_ps(in, mask);
			__m128 const cm = _mm_cmplt_ss(abs, a4);
			_mm_store_ss(p, MASK_COMBINE(b4, in, cm));
		}
	}
}


//*XDSP*S
static void Abs_V_VCN(float *pout, float const *pr, float const *pi, dword const n)
{
	dword i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		__m128 const re = _mm_load_ps(pr+i);
		__m128 const im = _mm_load_ps(pi+i);
		_mm_store_ps(pout+i, _mm_sqrt_ps(_mm_add_ps(_mm_mul_ps(re, re), _mm_mul_ps(im, im))));
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
		{
			__m128 const re = _mm_load_ss(pr+i);
			__m128 const im = _mm_load_ss(pi+i);
			_mm_store_ss(pout+i, _mm_sqrt_ss(_mm_add_ss(_mm_mul_ss(re, re), _mm_mul_ss(im, im))));
			
		}
	}
	
}

static int g_stat = 0x16BA2118;

// idum = 1664525L*idum + 1013904223L;

#define RAND_MUL	1664525
#define RAND_ADD	1013904223

//*XDSP*S
static void Rand_VI(int *pout, dword const n)
{
	int stat = g_stat;

	dword i = 0;

	for (dword c = 0; c < n/4; c++, i += 4)
	{
		stat = (stat * RAND_MUL + RAND_ADD);
		pout[i] = stat;
		
		stat = (stat * RAND_MUL + RAND_ADD);
		pout[i+1] = stat;
		
		stat = (stat * RAND_MUL + RAND_ADD);
		pout[i+2] = stat;
		
		stat = (stat * RAND_MUL + RAND_ADD);
		pout[i+3] = stat;
	}

	dword const r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
		{
			stat = (stat * RAND_MUL + RAND_ADD);
			pout[i] = stat;
		}
	}

	g_stat = stat;
}

//*XDSP*S
static void RandExp_V(float *pout, dword const n)
{
	Rand_VI((int *)pout, n);
	And_VI_SI((int *)pout, n, 0x87ffffff);
	Add_VI_SI((int *)pout, n, (127-16) << 23);
	
}

//*XDSP*S
static void RandExp2_V(float *pout, dword const n)
{
	Rand_VI((int *)pout, n);
	And_VI_SI((int *)pout, n, 0x83ffffff);
	Add_VI_SI((int *)pout, n, (127-8) << 23);
	
}


//*XDSP*S
static void DenormalToZero_V(float *p, dword const n)
{
	dword const x = 0x7f800000;
	__m128 const mask = _mm_set_ps1(*(float *)&x);
	__m128 const zero = _mm_setzero_ps();



	for (dword c = 0; c < n/4; c++, p += 4)
	{
		__m128 const in = _mm_load_ps(p);
		__m128 const x = _mm_and_ps(in, _mm_cmpeq_ps(_mm_and_ps(in, mask), zero));
		_mm_store_ps(p, _mm_xor_ps(in, x));
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, p++)
		{
			__m128 const in = _mm_load_ss(p);
			__m128 const x = _mm_and_ps(in, _mm_cmpeq_ss(_mm_and_ps(in, mask), zero));
			_mm_store_ss(p, _mm_xor_ps(in, x));
			
		}
	}
}


/////////////////////////////////////////////////////////////////////////



#include "set/BasicSSE.set"