
#include "stdafx.h"
#include "XDSPInterface.h"
#include "dspinternal.h"

void FPUClear(float *p, dword const n);




// biquad filters

__declspec(align(32)) class CBiquadFPU : public CXBiquadFilter
{
public:
	class CBiquadData
	{
	public:
		CBiquadData()
		{
			a[0] = a[1] = 0;
			b[0] = b[1] = b[2] = 0;
			
			ResetState();
		}

		void ResetState()
		{
			i[0] = i[1] = 0;
			o[0] = o[1] = 0;
		}
		
	public:
		float a[2];
		float b[3];
		
		float i[2];
		float o[2];
	};


	CBiquadFPU(int const numc)
		: numChannels(numc)
	{
	}


	virtual ~CBiquadFPU()
	{
	}

	virtual dword QueryMemoryUsage() const { return sizeof(this); }
	

public:
	CBiquadData Data[4];
	int numChannels;

};

//*XDSP*
static CXBiquadFilter *BiquadCreate(dword const numchannels)
{
	return new CBiquadFPU(numchannels);
}


//*XDSP*
static void BiquadResetState(CXBiquadFilter *pf, dword const channel)
{
	CBiquadFPU *f = (CBiquadFPU *)pf;
	f->Data[channel].ResetState();
}


//*XDSP*
static void BiquadSetCoeffs(CXBiquadFilter *pf, dword const channel, float const a0, float const a1, float const b0, float const b1, float const b2)
{
	CBiquadFPU *f = (CBiquadFPU *)pf;

	CBiquadFPU::CBiquadData *p = f->Data + channel;

	p->a[0] = a0;
	p->a[1] = a1;
	p->b[0] = b0;
	p->b[1] = b1;
	p->b[2] = b2;
}



static void BiquadRun1(CBiquadFPU::CBiquadData *f, float *p, dword const n)
{
	for (dword c = 0; c < n; c++, p++)
	{
		double x = f->b[0] * *p + f->b[1] * f->i[0] + f->b[2] * f->i[1] 
			+ f->a[0] * f->o[0] + f->a[1] * f->o[1];
	
		f->i[1] = f->i[0];
		f->i[0] = *p;
		
		f->o[1] = f->o[0];
		
		float fx = (float)x;
		
		undenormalise(fx);
		
		f->o[0] = fx;
		*p = fx;
		
	}
}

static void BiquadRun2(CBiquadFPU::CBiquadData *f, float *p, dword const n)
{
	for (dword c = 0; c < n; c++, p += 2)
	{
		double i = p[0];

		double x = f->b[0] * i + f->b[1] * f->i[0] + f->b[2] * f->i[1] 
			+ f->a[0] * f->o[0] + f->a[1] * f->o[1];

		
		f->i[1] = f->i[0];
		f->i[0] = (float)i;
		
		f->o[1] = f->o[0];
		
		float fx = (float)x;
		
		undenormalise(fx);
		
		f->o[0] = fx;
		p[0] = fx;
		


		i = p[1];
		
		x = f[1].b[0] * i + f[1].b[1] * f[1].i[0] + f[1].b[2] * f[1].i[1] 
			+ f[1].a[0] * f[1].o[0] + f[1].a[1] * f[1].o[1];
		
		
		f[1].i[1] = f[1].i[0];
		f[1].i[0] = (float)i;
		
		f[1].o[1] = f[1].o[0];
		
		fx = (float)x;
		
		undenormalise(fx);
		
		f[1].o[0] = fx;
		p[1] = fx;
		
	}
}


static void BiquadRun3(CBiquadFPU::CBiquadData *f, float *p, dword const n)
{
	for (dword c = 0; c < n; c++, p += 3)
	{
		double i = p[0];
		double x = f->b[0] * i + f->b[1] * f->i[0] + f->b[2] * f->i[1] 
			+ f->a[0] * f->o[0] + f->a[1] * f->o[1];
		f->i[1] = f->i[0];
		f->i[0] = (float)i;
		f->o[1] = f->o[0];
		float fx = (float)x;
		undenormalise(fx);
		f->o[0] = fx;
		p[0] = fx;
		
		
		i = p[1];
		x = f[1].b[0] * i + f[1].b[1] * f[1].i[0] + f[1].b[2] * f[1].i[1] 
			+ f[1].a[0] * f[1].o[0] + f[1].a[1] * f[1].o[1];
		f[1].i[1] = f[1].i[0];
		f[1].i[0] = (float)i;
		f[1].o[1] = f[1].o[0];
		fx = (float)x;
		undenormalise(fx);
		f[1].o[0] = fx;
		p[1] = fx;

		i = p[2];
		x = f[2].b[0] * i + f[2].b[1] * f[2].i[0] + f[2].b[2] * f[2].i[1] 
			+ f[2].a[0] * f[2].o[0] + f[2].a[1] * f[2].o[1];
		f[2].i[1] = f[2].i[0];
		f[2].i[0] = (float)i;
		f[2].o[1] = f[2].o[0];
		fx = (float)x;
		undenormalise(fx);
		f[2].o[0] = fx;
		p[2] = fx;
		
	}
}


static void BiquadRun4(CBiquadFPU::CBiquadData *f, float *p, dword const n)
{
	for (dword c = 0; c < n; c++, p += 4)
	{
		double i = p[0];
		double x = f->b[0] * i + f->b[1] * f->i[0] + f->b[2] * f->i[1] 
			+ f->a[0] * f->o[0] + f->a[1] * f->o[1];
		f->i[1] = f->i[0];
		f->i[0] = (float)i;
		f->o[1] = f->o[0];
		float fx = (float)x;
		undenormalise(fx);
		f->o[0] = fx;
		p[0] = fx;
		
		
		i = p[1];
		x = f[1].b[0] * i + f[1].b[1] * f[1].i[0] + f[1].b[2] * f[1].i[1] 
			+ f[1].a[0] * f[1].o[0] + f[1].a[1] * f[1].o[1];
		f[1].i[1] = f[1].i[0];
		f[1].i[0] = (float)i;
		f[1].o[1] = f[1].o[0];
		fx = (float)x;
		undenormalise(fx);
		f[1].o[0] = fx;
		p[1] = fx;
		
		
		i = p[2];
		x = f[2].b[0] * i + f[2].b[1] * f[2].i[0] + f[2].b[2] * f[2].i[1] 
			+ f[2].a[0] * f[2].o[0] + f[2].a[1] * f[2].o[1];
		f[2].i[1] = f[2].i[0];
		f[2].i[0] = (float)i;
		f[2].o[1] = f[2].o[0];
		fx = (float)x;
		undenormalise(fx);
		f[2].o[0] = fx;
		p[2] = fx;


		i = p[3];
		x = f[3].b[0] * i + f[3].b[1] * f[3].i[0] + f[3].b[2] * f[3].i[1] 
			+ f[3].a[0] * f[3].o[0] + f[3].a[1] * f[3].o[1];
		f[3].i[1] = f[3].i[0];
		f[3].i[0] = (float)i;
		f[3].o[1] = f[3].o[0];
		fx = (float)x;
		undenormalise(fx);
		f[3].o[0] = fx;
		p[3] = fx;
		
	}
}


//*XDSP*
static void BiquadRun2x2(CXBiquadFilter *pf, float *p, dword const n)
{
	CBiquadFPU::CBiquadData *f = ((CBiquadFPU *)pf)->Data;
	
	for (dword c = 0; c < n; c++, p += 2)
	{
		// left
		double i = p[0];
		double x = f->b[0] * (i + f->i[1]) + f->b[1] * f->i[0]  
			+ f->a[0] * f->o[0] + f->a[1] * f->o[1];
			
		f->i[1] = f->i[0];
		f->i[0] = (float)i;
		f->o[1] = f->o[0];
		float fx = (float)x;
		undenormalise(fx);
		f->o[0] = fx;
	
		// second left
		i = f->o[1];
		x = f[2].b[0] * (i + f[2].i[1]) + f[2].b[1] * f[2].i[0] 
			+ f[2].a[0] * f[2].o[0] + f[2].a[1] * f[2].o[1];
		
		f[2].i[1] = f[2].i[0];
		f[2].i[0] = (float)i;
		f[2].o[1] = f[2].o[0];
		fx = (float)x;
		undenormalise(fx);
		f[2].o[0] = fx;
		p[0] = fx;
		
		
		// right

		i = p[1];
		x = f[1].b[0] * (i + f[1].i[1]) + f[1].b[1] * f[1].i[0] 
			+ f[1].a[0] * f[1].o[0] + f[1].a[1] * f[1].o[1];
			
		f[1].i[1] = f[1].i[0];
		f[1].i[0] = (float)i;
		f[1].o[1] = f[1].o[0];
		fx = (float)x;
		undenormalise(fx);
		f[1].o[0] = fx;

		// second right
		
		i = f[1].o[1];
		x = f[3].b[0] * (i + f[3].i[1]) + f[3].b[1] * f[3].i[0] 
			+ f[3].a[0] * f[3].o[0] + f[3].a[1] * f[3].o[1];
		
		f[3].i[1] = f[3].i[0];
		f[3].i[0] = (float)i;
		f[3].o[1] = f[3].o[0];
		fx = (float)x;
		undenormalise(fx);
		f[3].o[0] = fx;
		p[1] = fx;
		
	}
	
}


//*XDSP*
static void BiquadRun(CXBiquadFilter *pf, float *p, dword const n)
{
	CBiquadFPU *f = (CBiquadFPU *)pf;

	switch(f->numChannels)
	{
	case 1: BiquadRun1(f->Data, p, n); break;
	case 2: BiquadRun2(f->Data, p, n); break;
	case 3: BiquadRun3(f->Data, p, n); break;
	case 4: BiquadRun4(f->Data, p, n); break;
	default: break;
		
	}
}

/*
void SetBiquadFPU(CXDSPInterface *p)
{
	p->BiquadCreate = BiquadCreate;
	p->BiquadResetState = BiquadResetState;
	p->BiquadSetCoeffs = BiquadSetCoeffs;
	p->BiquadRun = BiquadRun;
	p->BiquadRun2x2 = BiquadRun2x2;
}
*/

#include "set/BiquadFPU.set"