
#include "stdafx.h"
#include "XDSPInterface.h"
#include <xmmintrin.h>
#include "dspinternal.h"




// biquad 

class CBiquadSSE : public CXBiquadFilter
{
public:
	__declspec(align(32)) class CData
	{
	public:
		CData()
		{
			a[0] = a[1] = b[0] = b[1] = b[2] = _mm_setzero_ps();
			ResetState();
		}
			
		void ResetState()
		{
			i[0] = i[1] = o[0] = o[1] = _mm_setzero_ps();
			
		}

		
		__m128 i[2];
		__m128 o[2];
		
		__m128 a[2];
		__m128 b[3];
	};
		

	CBiquadSSE(int const numch)
	{
		numChannels = numch;
	}

	virtual ~CBiquadSSE()
	{

	}
	
	virtual dword QueryMemoryUsage() const { return sizeof(this); }
	
	void *operator new(size_t size)
	{
		return _aligned_malloc(size, 32);
	}

	void operator delete(void *p)
	{
		_aligned_free(p);
	}
	
public:
	CData Data;

	int numChannels;
	
};

//*XDSP*S
static CXBiquadFilter *BiquadCreate(dword const numchannels)
{
	return new CBiquadSSE(numchannels);
}

//*XDSP*S
static void BiquadResetState(CXBiquadFilter *pf, dword const channel)
{
	CBiquadSSE::CData *f = (CBiquadSSE::CData *)&((CBiquadSSE *)pf)->Data;
	
	f->i[0].m128_f32[channel] = 0;
	f->i[1].m128_f32[channel] = 0;
	f->o[0].m128_f32[channel] = 0;
	f->o[1].m128_f32[channel] = 0;

}


//*XDSP*S
static void BiquadSetCoeffs(CXBiquadFilter *pf, dword const channel, float const a0, float const a1, float const b0, float const b1, float const b2)
{
	CBiquadSSE *pbq = (CBiquadSSE *)pf;
	CBiquadSSE::CData *f = (CBiquadSSE::CData *)&((CBiquadSSE *)pf)->Data;

	M128_MOV(f->a[0], channel, a0);
	M128_MOV(f->a[1], channel, a1);
	M128_MOV(f->b[0], channel, b0);
	M128_MOV(f->b[1], channel, b1);
	M128_MOV(f->b[2], channel, b2);
	
}


static void BiquadRun1(CBiquadSSE::CData *f, float *p, dword const n)
{
	for (dword c = 0; c < n/2; c++, p += 2)
	{
		__m128 i = _mm_load_ss(p);
		__m128 x = _mm_mul_ss(f->b[0], i);
		x = _mm_add_ss(x, _mm_mul_ss(f->b[1], f->i[0]));
		x = _mm_add_ss(x, _mm_mul_ss(f->b[2], f->i[1]));

		x = _mm_add_ss(x, _mm_mul_ss(f->a[0], f->o[0]));
		x = _mm_add_ss(x, _mm_mul_ss(f->a[1], f->o[1]));
		
		_mm_store_ss(p, x);

		
		__m128 j = _mm_load_ss(p+1);
		__m128 y = _mm_mul_ss(f->b[0], j);
		y = _mm_add_ss(y, _mm_mul_ss(f->b[1], i));
		y = _mm_add_ss(y, _mm_mul_ss(f->b[2], f->i[0]));
	
		f->i[1] = i;
		f->i[0] = j;
		
		y = _mm_add_ss(y, _mm_mul_ss(f->a[0], x));
		y = _mm_add_ss(y, _mm_mul_ss(f->a[1], f->o[0]));
		
		f->o[1] = x;
		f->o[0] = y;
		
		_mm_store_ss(p+1, y);
		
	}

	if (n & 1)
	{
		__m128 i = _mm_load_ss(p);
		__m128 x = _mm_mul_ss(f->b[0], i);
		x = _mm_add_ss(x, _mm_mul_ss(f->b[1], f->i[0]));
		x = _mm_add_ss(x, _mm_mul_ss(f->b[2], f->i[1]));
		
		f->i[1] = f->i[0];
		f->i[0] = i;
		
		x = _mm_add_ss(x, _mm_mul_ss(f->a[0], f->o[0]));
		x = _mm_add_ss(x, _mm_mul_ss(f->a[1], f->o[1]));

		f->o[1] = f->o[0];
		f->o[0] = x;

		_mm_store_ss(p, x);
	}
}


static void BiquadRun4(CBiquadSSE::CData *f, float *p, dword const n)
{

	for (dword c = 0; c < n/2; c++, p += 8)
	{
		__m128 i = _mm_load_ps(p);
		__m128 x = _mm_mul_ps(f->b[0], i);
		x = _mm_add_ps(x, _mm_mul_ps(f->b[1], f->i[0]));
		x = _mm_add_ps(x, _mm_mul_ps(f->b[2], f->i[1]));
		
		x = _mm_add_ps(x, _mm_mul_ps(f->a[0], f->o[0]));
		x = _mm_add_ps(x, _mm_mul_ps(f->a[1], f->o[1]));
		
		_mm_store_ps(p, x);
		
		
		__m128 j = _mm_load_ps(p+4);
		__m128 y = _mm_mul_ps(f->b[0], j);
		y = _mm_add_ps(y, _mm_mul_ps(f->b[1], i));
		y = _mm_add_ps(y, _mm_mul_ps(f->b[2], f->i[0]));
		
		f->i[1] = i;
		f->i[0] = j;
		
		y = _mm_add_ps(y, _mm_mul_ps(f->a[0], x));
		y = _mm_add_ps(y, _mm_mul_ps(f->a[1], f->o[0]));
	
		f->o[1] = x;
		f->o[0] = y;
		
		_mm_store_ps(p+4, y);
		
	}
	
	if (n & 1)
	{
		__m128 i = _mm_load_ps(p);
		__m128 x = _mm_mul_ps(f->b[0], i);
		x = _mm_add_ps(x, _mm_mul_ps(f->b[1], f->i[0]));
		x = _mm_add_ps(x, _mm_mul_ps(f->b[2], f->i[1]));
		
		f->i[1] = f->i[0];
		f->i[0] = i;
		
		x = _mm_add_ps(x, _mm_mul_ps(f->a[0], f->o[0]));
		x = _mm_add_ps(x, _mm_mul_ps(f->a[1], f->o[1]));
		
		f->o[1] = f->o[0];
		f->o[0] = x;

		_mm_store_ps(p, x);
	}
}


static void BiquadRun2(CBiquadSSE::CData *f, float *p, dword const n)
{
	
	for (dword c = 0; c < n/2; c++, p += 4)
	{
		__m128 i = _mm_load_ps(p);
		__m128 x = _mm_mul_ps(f->b[0], i);
		x = _mm_add_ps(x, _mm_mul_ps(f->b[1], f->i[0]));
		x = _mm_add_ps(x, _mm_mul_ps(f->b[2], f->i[1]));
		
		x = _mm_add_ps(x, _mm_mul_ps(f->a[0], f->o[0]));
		x = _mm_add_ps(x, _mm_mul_ps(f->a[1], f->o[1]));
		
		
		__m128 j = _mm_shuffle_ps(i, i, _MM_SHUFFLE(3, 2, 3, 2));
		__m128 y = _mm_mul_ps(f->b[0], j);
		y = _mm_add_ps(y, _mm_mul_ps(f->b[1], i));
		y = _mm_add_ps(y, _mm_mul_ps(f->b[2], f->i[0]));
		
		f->i[1] = i;
		f->i[0] = j;
		
		y = _mm_add_ps(y, _mm_mul_ps(f->a[0], x));
		y = _mm_add_ps(y, _mm_mul_ps(f->a[1], f->o[0]));
		
		f->o[1] = x;
		f->o[0] = y;
		
		_mm_store_ps(p, _mm_movelh_ps(x, y));

		
	}
	
	if (n & 1)
	{
		__m128 i = _mm_set_ps(0, 0, p[1], p[0]);
		__m128 x = _mm_mul_ps(f->b[0], i);
		x = _mm_add_ps(x, _mm_mul_ps(f->b[1], f->i[0]));
		x = _mm_add_ps(x, _mm_mul_ps(f->b[2], f->i[1]));
		
		f->i[1] = f->i[0];
		f->i[0] = i;
		
		x = _mm_add_ps(x, _mm_mul_ps(f->a[0], f->o[0]));
		x = _mm_add_ps(x, _mm_mul_ps(f->a[1], f->o[1]));
		
		f->o[1] = f->o[0];
		f->o[0] = x;

		_mm_storel_pi((__m64 *)p, x);
		
	}
}


static void BiquadRun3(CBiquadSSE::CData *f, float *p, dword const n)
{
	
	for (dword c = 0; c < n/2; c++, p += 6)
	{
		__m128 i = _mm_loadu_ps(p);
		__m128 x = _mm_mul_ps(f->b[0], i);
		x = _mm_add_ps(x, _mm_mul_ps(f->b[1], f->i[0]));
		x = _mm_add_ps(x, _mm_mul_ps(f->b[2], f->i[1]));
		
		x = _mm_add_ps(x, _mm_mul_ps(f->a[0], f->o[0]));
		x = _mm_add_ps(x, _mm_mul_ps(f->a[1], f->o[1]));
		
		_mm_storel_pi((__m64 *)p, x);
		_mm_store_ss(p+2, _mm_shuffle_ps(x, x, _MM_SHUFFLE(3, 2, 1, 2)));
		
		
		__m128 j = _mm_loadu_ps(p+3);

		__m128 y = _mm_mul_ps(f->b[0], j);
		y = _mm_add_ps(y, _mm_mul_ps(f->b[1], i));
		y = _mm_add_ps(y, _mm_mul_ps(f->b[2], f->i[0]));
		
		f->i[1] = i;
		f->i[0] = j;
		
		y = _mm_add_ps(y, _mm_mul_ps(f->a[0], x));
		y = _mm_add_ps(y, _mm_mul_ps(f->a[1], f->o[0]));
		
		f->o[1] = x;
		f->o[0] = y;
		
		_mm_storel_pi((__m64 *)(p+3), y);
		_mm_store_ss(p+5, _mm_shuffle_ps(y, y, _MM_SHUFFLE(3, 2, 1, 2)));
		
	}
	
	if (n & 1)
	{
		__m128 i = _mm_set_ps(0, p[2], p[1], p[0]);
		__m128 x = _mm_mul_ps(f->b[0], i);
		x = _mm_add_ps(x, _mm_mul_ps(f->b[1], f->i[0]));
		x = _mm_add_ps(x, _mm_mul_ps(f->b[2], f->i[1]));
		
		f->i[1] = f->i[0];
		f->i[0] = i;
		
		x = _mm_add_ps(x, _mm_mul_ps(f->a[0], f->o[0]));
		x = _mm_add_ps(x, _mm_mul_ps(f->a[1], f->o[1]));
		
		f->o[1] = f->o[0];
		f->o[0] = x;
		
		_mm_storel_pi((__m64 *)p, x);
		_mm_store_ss(p+2, _mm_shuffle_ps(x, x, _MM_SHUFFLE(3, 2, 1, 2)));
		
	}
}



//*XDSP*S
static void BiquadRun2x2(CXBiquadFilter *pf, float *p, dword const n)
{
	CBiquadSSE::CData *f = &((CBiquadSSE *)pf)->Data;

	for (dword c = 0; c < n/2; c++, p += 4)
	{
		__m128 memin = _mm_load_ps(p);
		__m128 i = _mm_movelh_ps(memin, f->o[0]);

		__m128 x = _mm_mul_ps(f->b[0], _mm_add_ps(i, f->i[1]));
		x = _mm_add_ps(x, _mm_mul_ps(f->b[1], f->i[0]));
		
		x = _mm_add_ps(x, _mm_mul_ps(f->a[0], f->o[0]));
		x = _mm_add_ps(x, _mm_mul_ps(f->a[1], f->o[1]));
		
		__m128 j = _mm_shuffle_ps(memin, x, _MM_SHUFFLE(1, 0, 3, 2));
		__m128 y = _mm_mul_ps(f->b[0], _mm_add_ps(j, f->i[0]));
		y = _mm_add_ps(y, _mm_mul_ps(f->b[1], i));
		
		f->i[1] = i;
		f->i[0] = j;
		
		y = _mm_add_ps(y, _mm_mul_ps(f->a[0], x));
		y = _mm_add_ps(y, _mm_mul_ps(f->a[1], f->o[0]));
		
		f->o[1] = x;
		f->o[0] = y;
		
		_mm_store_ps(p, _mm_shuffle_ps(x, y, _MM_SHUFFLE(3, 2, 3, 2)));
		
		
	}
	
	if (n & 1)
	{
		__m128 i = _mm_set_ps(0, 0, p[1], p[0]);
		i = _mm_movelh_ps(i, f->o[0]);
		
		__m128 x = _mm_mul_ps(f->b[0], _mm_add_ps(i, f->i[1]));
		x = _mm_add_ps(x, _mm_mul_ps(f->b[1], f->i[0]));
		
		f->i[1] = f->i[0];
		f->i[0] = i;
		
		x = _mm_add_ps(x, _mm_mul_ps(f->a[0], f->o[0]));
		x = _mm_add_ps(x, _mm_mul_ps(f->a[1], f->o[1]));
		
		f->o[1] = f->o[0];
		f->o[0] = x;
		
		_mm_storel_pi((__m64 *)p, _mm_movehl_ps(x, x));
		
	}
}




//*XDSP*S
static void BiquadRun(CXBiquadFilter *pf, float *p, dword const n)
{
	CBiquadSSE *f = (CBiquadSSE *)pf;
	
	switch(f->numChannels)
	{
	case 1: BiquadRun1(&f->Data, p, n); break;
	case 2: BiquadRun2(&f->Data, p, n); break;
	case 3: BiquadRun3(&f->Data, p, n); break;
	case 4: BiquadRun4(&f->Data, p, n); break;
	default: break;
	}
}

/*
void SetBiquadSSE(CXDSPInterface *p)
{
	p->BiquadCreate = BiquadCreate;
	p->BiquadResetState = BiquadResetState;
	p->BiquadSetCoeffs = BiquadSetCoeffs;
	p->BiquadRun = BiquadRun;
	p->BiquadRun2x2 = BiquadRun2x2;
	
}
*/

#include "set/BiquadSSE.set"
           