
#include "stdafx.h"
#include "XDSPInterface.h"
#include "dspinternal.h"

void FPUClear_V(float *p, dword const n);
void FPUCopy_V_V(float *pout, float const *pin, dword const n);



class CCombFPU : public CXCombFilter
{
public:
	CCombFPU(dword const delay)
	{
		BufferPos = Buffer = NULL;
		BufferLength = 0;
		FilterPole = FilterGain = FilterLastOut = Feedback = 0;
		SetDelay(delay);
	}

	virtual ~CCombFPU()
	{
		delete[] Buffer;
	}

	virtual dword QueryMemoryUsage() const { return sizeof(this) + BufferLength * sizeof(float); }
	

	void SetDelay(dword d)
	{
		BufferLength = d;
		delete[] Buffer;
		BufferPos = Buffer = new float [d];
		FPUClear_V(Buffer, d);

	}

	void SetState(float const *state, float const onepolestate)
	{
		FilterLastOut = onepolestate;
		
		if (state == NULL)
			FPUClear_V(Buffer, BufferLength);
		else
			FPUCopy_V_V(Buffer, state, BufferLength);
	}

public:
	float FilterPole;
	float FilterGain;
	float FilterLastOut;
	float Feedback;
	dword BufferLength;
	float *BufferPos;
	float *Buffer;

};


//*XDSP*
static CXCombFilter *CombCreate(dword const delay)
{
	return new CCombFPU(delay);
}




//*XDSP*
static void CombSetFeedback(CXCombFilter *_pc, float const fb)
{
	CCombFPU *pc = (CCombFPU *)_pc;
	pc->Feedback = fb;		
}

//*XDSP*
static void CombSetFilterCoeffs(CXCombFilter *_pc, float const pole, float const gain)
{
	CCombFPU *pc = (CCombFPU *)_pc;
	pc->FilterPole = pole;
	pc->FilterGain = gain;
}

//*XDSP*
static void CombSetState(CXCombFilter *_pc, float const *state, float const onepolestate)
{
	CCombFPU *pc = (CCombFPU *)_pc;
	pc->SetState(state, onepolestate);	
}


//*XDSP*
static void CombRun(CXCombFilter *_pc, float *p, dword const n)
{
	CCombFPU *pc = (CCombFPU *)_pc;
	
	dword num = n;

	do
	{
		dword const x = __min(num, pc->BufferLength - (pc->BufferPos - pc->Buffer));
		num -= x;

		for (dword c = 0; c < x; c++, p++)
		{
			float out = *pc->BufferPos;

			float tobuf = *p + (out * pc->Feedback);
			undenormalise(tobuf);

			*pc->BufferPos++ = tobuf;
			*p = out;
		
		}
		
		if (pc->BufferPos == pc->Buffer + pc->BufferLength)
			pc->BufferPos = pc->Buffer;


	} while(num > 0);
	
}

//*XDSP*
static void CombOnepoleRun(CXCombFilter *_pc, float *p, dword const n)
{
	CCombFPU *pc = (CCombFPU *)_pc;
	
	dword num = n;
	
	do
	{
		dword const x = __min(num, pc->BufferLength - (pc->BufferPos - pc->Buffer));
		
		for (dword c = 0; c < x; c++, p++)
		{
			float out = *pc->BufferPos;
			undenormalise(out);
			pc->FilterLastOut = out * pc->FilterGain + pc->FilterLastOut * pc->FilterPole;
			undenormalise(pc->FilterLastOut);
			*pc->BufferPos = p[0] + (pc->FilterLastOut * pc->Feedback);
			p[0] = out;
			pc->BufferPos++;
			
			
		}
		
		if (pc->BufferPos == pc->Buffer + pc->BufferLength)
			pc->BufferPos = pc->Buffer;
		
		num -= x;
		
	} while(num > 0);
	
}

/*
void SetCombFPU(CXDSPInterface *p)
{
	p->CombCreate = CombCreate;
	p->CombSetState = CombSetState;
	p->CombSetFeedback = CombSetFeedback;
	p->CombSetFilterCoeffs = CombSetFilterCoeffs;
	p->CombRun = CombRun;
	p->CombOnepoleRun = CombOnepoleRun;
}
*/


#include "set/CombFPU.set"