
#include "stdafx.h"
#include "XDSPInterface.h"
#include <xmmintrin.h>
#include "dspinternal.h"

void SSEClear_V(float *p, dword const n);
void FPUCopy_V_V(float *pout, float const *pin, dword const n);

class CCombSSE : public CXCombFilter
{
public:
	CCombSSE(dword const delay)
	{
		Buffer = NULL;
		BufferLength = 0;
		BufferPos = NULL;

		FilterPole = FilterGain = FilterLastOut = Feedback = 0;
		SetDelay(delay);
	}
	
	virtual ~CCombSSE()
	{
		_aligned_free(Buffer);
	}
	
	virtual dword QueryMemoryUsage() const { return sizeof(this) + BufferLength * sizeof(float); }
	
	void SetDelay(dword d)
	{
		BufferLength = d;
		_aligned_free(Buffer);
		
		BufferPos = Buffer = (float *)_aligned_malloc(d * sizeof(float), 32);
		SSEClear_V(Buffer, d);
		
	}

	void SetState(float const *state, float const onepolestate)
	{
		FilterLastOut = onepolestate;

		if (state == NULL)
			SSEClear_V(Buffer, BufferLength);
		else
			FPUCopy_V_V(Buffer, state, BufferLength);
	}

	void *operator new(size_t size)
	{
		return _aligned_malloc(size, 32);
	}
	
	void operator delete(void *p)
	{
		_aligned_free(p);
	}

	

public:
	float FilterPole;
	float FilterGain;
	float FilterLastOut;
	float Feedback;

	dword BufferLength;
	float *BufferPos;
	float *Buffer;
	
};



//*XDSP*S
static CXCombFilter *CombCreate(dword const delay)
{
	return new CCombSSE(delay);
}

//*XDSP*S
static void CombSetFeedback(CXCombFilter *_pc, float const fb)
{
	CCombSSE *pc = (CCombSSE *)_pc;
	pc->Feedback = fb;		
}

//*XDSP*S
static void CombSetFilterCoeffs(CXCombFilter *_pc, float const pole, float const gain)
{
	CCombSSE *pc = (CCombSSE *)_pc;
	pc->FilterPole = pole;
	pc->FilterGain = gain;
}


//*XDSP*S
static void CombSetState(CXCombFilter *_pc, float const *state, float const onepolestate)
{
	CCombSSE *pc = (CCombSSE *)_pc;
	pc->SetState(state, onepolestate);	
}

//*XDSP*S
static void CombRun(CXCombFilter *_pc, float *p, dword const n)
{
	CCombSSE *pc = (CCombSSE *)_pc;
	

	dword num = n;
	
	__m128 const fb = _mm_load1_ps(&pc->Feedback);	
	
	do
	{
		dword x = __min(num, pc->BufferLength - (pc->BufferPos - pc->Buffer));
		num -= x;
	
		
		if ((dword)p & 15)
		{
			float *dest = (float *)(((dword)p + 15) & (~0x0f));
			while(p != dest && x > 0)
			{
				__m128 out = _mm_load_ss(pc->BufferPos);
				__m128 t = _mm_add_ss(_mm_load_ss(p), _mm_mul_ss(out, fb));
				_mm_store_ss(p, out);
				_mm_store_ss(pc->BufferPos++, t);
				
				p++;
				x--;
			}
			
		}
		
		for (dword c = 0; c < x/4; c++, p += 4)
		{

			__m128 const out = _mm_loadu_ps(pc->BufferPos);
				
			_mm_storeu_ps(pc->BufferPos, _mm_add_ps(_mm_load_ps(p), _mm_mul_ps(fb, out)));
			pc->BufferPos += 4;

			_mm_store_ps(p, out);
			

			
		}

		dword const last = x & 3;
		if (last > 0)
		{
			for (dword c = 0; c < last; c++, p++)
			{
				__m128 out = _mm_load_ss(pc->BufferPos);
				__m128 t = _mm_add_ss(_mm_load_ss(p), _mm_mul_ss(out, fb));
				_mm_store_ss(p, out);
				_mm_store_ss(pc->BufferPos++, t);
			}
			
		}
		

		if (pc->BufferPos == pc->Buffer + pc->BufferLength)
			pc->BufferPos = pc->Buffer;
		
		
	} while(num > 0);


}

//*XDSP*S
static void CombOnepoleRun(CXCombFilter *_pc, float *p, dword const n)
{
	CCombSSE *pc = (CCombSSE *)_pc;
	
	
	dword num = n;
	
	__m128 const fp = _mm_load_ss(&pc->FilterPole);
	__m128 flo = _mm_load_ss(&pc->FilterLastOut);
	
	__m128 const fg = _mm_load1_ps(&pc->FilterGain);
	__m128 const fb = _mm_load1_ps(&pc->Feedback);	
	
	__declspec(align(16)) float fbuf[4];
	
	do
	{
		dword x = __min(num, pc->BufferLength - (pc->BufferPos - pc->Buffer));
		num -= x;
		
		
		if ((dword)p & 15)
		{
			float *dest = (float *)(((dword)p + 15) & (~0x0f));
			while(p != dest && x > 0)
			{
				__m128 out = _mm_load_ss(pc->BufferPos);
				flo = _mm_add_ss(_mm_mul_ss(out, fg), _mm_mul_ss(flo, fp));
				__m128 t = _mm_add_ss(_mm_load_ss(p), _mm_mul_ss(flo, fb));
				_mm_store_ss(p, out);
				_mm_store_ss(pc->BufferPos++, t);
				
				p++;
				x--;
			}
			
		}
		
		for (dword c = 0; c < x/4; c++, p += 4)
		{
			
			__m128 const out = _mm_loadu_ps(pc->BufferPos);
			__m128 const t = _mm_mul_ps(out, fg);
			
			flo = _mm_add_ss(t, _mm_mul_ss(flo, fp));
			_mm_store_ss(fbuf, flo);
			
			flo = _mm_add_ss(_mm_shuffle_ps(t, t, _MM_SHUFFLE(3, 2, 1, 1)), _mm_mul_ss(flo, fp));
			_mm_store_ss(fbuf+1, flo);
			
			flo = _mm_add_ss(_mm_shuffle_ps(t, t, _MM_SHUFFLE(3, 2, 1, 2)), _mm_mul_ss(flo, fp));
			_mm_store_ss(fbuf+2, flo);
			
			flo = _mm_add_ss(_mm_shuffle_ps(t, t, _MM_SHUFFLE(3, 2, 1, 3)), _mm_mul_ss(flo, fp));
			_mm_store_ss(fbuf+3, flo);
			
			_mm_storeu_ps(pc->BufferPos, _mm_add_ps(_mm_load_ps(p), _mm_mul_ps(fb, _mm_load_ps(fbuf))));
			pc->BufferPos += 4;
			
			_mm_store_ps(p, out);
			
			
			
		}
		
		dword const last = x & 3;
		if (last > 0)
		{
			for (dword c = 0; c < last; c++, p++)
			{
				__m128 out = _mm_load_ss(pc->BufferPos);
				flo = _mm_add_ss(_mm_mul_ss(out, fg), _mm_mul_ss(flo, fp));
				__m128 t = _mm_add_ss(_mm_load_ss(p), _mm_mul_ss(flo, fb));
				_mm_store_ss(p, out);
				_mm_store_ss(pc->BufferPos++, t);
			}
			
		}
		
		
		if (pc->BufferPos == pc->Buffer + pc->BufferLength)
			pc->BufferPos = pc->Buffer;
		
		
	} while(num > 0);
	
	_mm_store_ss(&pc->FilterLastOut, flo);
	
}

/*
void SetCombSSE(CXDSPInterface *p)
{
	p->CombCreate = CombCreate;
	p->CombSetState = CombSetState;
	p->CombSetFeedback = CombSetFeedback;
	p->CombSetFilterCoeffs = CombSetFilterCoeffs;
	p->CombRun = CombRun;
	p->CombOnepoleRun = CombOnepoleRun;
}
*/

#include "set/CombSSE.set"