#include "stdafx.h"
#include "dspinternal.h"

extern CXDSPInterface xdsp;


//*XDSP*
static CXMWOscillator *MWOscillatorCreate()
{
	CXMWOscillatorImp *po = new CXMWOscillatorImp;
	po->pMW = NULL;
	po->Phase = 0;


	return po;
}

//*XDSP*
static void MWOscillatorSetMW(CXMWOscillator *_po, CXMipWave const *pmw)
{
	CXMWOscillatorImp *po = (CXMWOscillatorImp *)_po;
	po->pMW = (CXMipWaveImp const *)pmw;
	
}

static void RunOsc(float *pout, dword const n, short const *data, dword &phase, dword const rate, dword const shift)
{
	dword const mask = (1 << shift) - 1;
	
	for (dword i = 0; i < n; i++)
	{
		dword const ipos = phase >> shift;
		float const alpha = (phase & mask) * (1.0 / (1 << shift));
		
		pout[i] = 2 * (data[ipos] + alpha * (data[ipos+1] - data[ipos]));
		
		phase += rate;
		
	}
	
	
}

static void RunOsc3(float *pout, dword const n, short const *data, dword &phase, dword const rate, dword const shift)
{
	int *po = (int *)pout;
	dword p = phase;

	dword const sm16 = shift - 16;

	dword i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		dword alpha = p >> sm16;
		dword di = alpha >> 16;
		alpha &= 65535;
		p += rate;
		int s0 = data[di];
		int s1 = data[di+1];
		s1 -= s0;
		s1 *= alpha;
		s1 >>= 16;
		s1 += s0;
		po[i] = s1;

		alpha = p >> sm16;
		di = alpha >> 16;
		alpha &= 65535;
		p += rate;
		s0 = data[di];
		s1 = data[di+1];
		s1 -= s0;
		s1 *= alpha;
		s1 >>= 16;
		s1 += s0;
		po[i+1] = s1;

		alpha = p >> sm16;
		di = alpha >> 16;
		alpha &= 65535;
		p += rate;
		s0 = data[di];
		s1 = data[di+1];
		s1 -= s0;
		s1 *= alpha;
		s1 >>= 16;
		s1 += s0;
		po[i+2] = s1;

		alpha = p >> sm16;
		di = alpha >> 16;
		alpha &= 65535;
		p += rate;
		s0 = data[di];
		s1 = data[di+1];
		s1 -= s0;
		s1 *= alpha;
		s1 >>= 16;
		s1 += s0;
		po[i+3] = s1;
		
	}

	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
		{
			dword alpha = p >> sm16;
			dword di = alpha >> 16;
			alpha &= 65535;
			p += rate;
			int s0 = data[di];
			int s1 = data[di+1];
			s1 -= s0;
			s1 *= alpha;
			s1 >>= 16;
			s1 += s0;
			po[i] = s1;
			
		}
	}

	phase = p;

	xdsp.Copy_V_VIP(pout, n);
	
}

static void RunOsc3PM(float *pout, dword const n, short const *data, dword &phase, dword const rate, dword const shift, dword const *pmtbl)
{
	int *po = (int *)pout;
	dword p = phase;
	
	dword const sm16 = shift - 16;
	
	dword i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		dword alpha = (p + pmtbl[i]) >> sm16;
		dword di = alpha >> 16;
		alpha &= 65535;
		p += rate;
		int s0 = data[di];
		int s1 = data[di+1];
		s1 -= s0;
		s1 *= alpha;
		s1 >>= 16;
		s1 += s0;
		po[i] = s1;
		
		alpha = (p + pmtbl[i+1]) >> sm16;
		di = alpha >> 16;
		alpha &= 65535;
		p += rate;
		s0 = data[di];
		s1 = data[di+1];
		s1 -= s0;
		s1 *= alpha;
		s1 >>= 16;
		s1 += s0;
		po[i+1] = s1;
		
		alpha = (p + pmtbl[i+2]) >> sm16;
		di = alpha >> 16;
		alpha &= 65535;
		p += rate;
		s0 = data[di];
		s1 = data[di+1];
		s1 -= s0;
		s1 *= alpha;
		s1 >>= 16;
		s1 += s0;
		po[i+2] = s1;
		
		alpha = (p + pmtbl[i+3]) >> sm16;
		di = alpha >> 16;
		alpha &= 65535;
		p += rate;
		s0 = data[di];
		s1 = data[di+1];
		s1 -= s0;
		s1 *= alpha;
		s1 >>= 16;
		s1 += s0;
		po[i+3] = s1;
		
	}
	
	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
		{
			dword alpha = (p + pmtbl[i]) >> sm16;
			dword di = alpha >> 16;
			alpha &= 65535;
			p += rate;
			int s0 = data[di];
			int s1 = data[di+1];
			s1 -= s0;
			s1 *= alpha;
			s1 >>= 16;
			s1 += s0;
			po[i] = s1;
			
		}
	}
	
	phase = p;
	
	xdsp.Copy_V_VIP(pout, n);
	
}


void RunOsc3RampAdd(float *pout, dword const n, short const *data, dword &phase, dword const rate, dword const shift, dword &amp, dword const astep)
{
	int *po = (int *)pout;
	dword p = phase;
	dword a = amp;
	
	dword const sm16 = shift - 16;
	
	dword i = 0;
	
	for (dword c = 0; c < n; c++, i++)
	{
		dword alpha = p >> sm16;
		dword di = alpha >> 16;
		alpha &= 65535;
		p += rate;
		int s0 = data[di];
		int s1 = data[di+1];
		s1 -= s0;
		s1 *= alpha;
		s1 >>= 16;
		s1 += s0;

		s1 *= amp >> 16;
		s1 >>= 16;
		amp += astep;

		po[i] += s1;
		
	}

	amp = a;
	phase = p;

	xdsp.Copy_V_VIP(pout, n);
	
}

void RunOsc3PMRampAdd(float *pout, dword const n, short const *data, dword &phase, dword const rate, dword const shift, dword const *pmtbl, dword &amp, dword const astep)
{
	int *po = (int *)pout;
	dword p = phase;
	dword a = amp;
	
	dword const sm16 = shift - 16;
	
	dword i = 0;
	
	for (dword c = 0; c < n; c++, i++)
	{
		dword alpha = (p + pmtbl[i]) >> sm16;
		dword di = alpha >> 16;
		alpha &= 65535;
		p += rate;
		int s0 = data[di];
		int s1 = data[di+1];
		s1 -= s0;
		s1 *= alpha;
		s1 >>= 16;
		s1 += s0;
		
		s1 *= amp >> 16;
		s1 >>= 16;
		amp += astep;
		
		po[i] += s1;
		
	}
	
	amp = a;
	phase = p;
	
	xdsp.Copy_V_VIP(pout, n);
	
}


#pragma warning (disable : 4731) 


static void RunOsc2(float *pout, dword const n, short const *data, dword &phase, dword const rate, dword const shift)
{
	__asm
	{
		
		mov edi, [pout]
			mov esi, [data]
			mov ecx, [shift]
			
			mov eax, [phase]
			mov ebx, [eax]
			

			mov eax, [rate]
			push eax

			mov eax, [n]
			lea eax, [edi + eax * 4]	
			push eax
			
			sub ecx, 16

			push ebp

pou:		
			mov eax, ebx
			add edi, 4
			shr eax, cl
			mov edx, eax
			shr eax, 16
			movsx ebp, word ptr [esi + eax * 2]
			and edx, 65535
			movsx eax, word ptr [esi + eax * 2 + 2]
			add ebx, [esp+8]
			sub eax, ebp
			imul eax, edx
			sar eax, 16
			add eax, ebp
			mov [edi-4], eax
			cmp edi, [esp+4]
			jnz pou
			
			pop ebp

			add esp, 8
			
			mov eax, [phase]
			mov [eax], ebx

	}
	
	
	xdsp.Copy_V_VIP(pout, n);
}


// no-mmx version

/*
static void RunOsc2(float *pout, dword const n, short const *data, dword &phase, dword const rate, dword const shift)
{
	__asm
	{

			mov edi, [pout]
			mov esi, [data]
			mov ecx, [shift]

			mov eax, [phase]
			mov ebx, [eax]
			
			mov eax, [rate]
			push eax					// rate = [esp+12]
	
			mov eax, [n]
			lea eax, [edi + eax * 4]	
			push eax					// endofbuf = [esp+8]
			
			mov eax, 1
			shl eax, cl
			dec eax
			push eax					// mask = [esp+4]

			push ebp

pou:		
			mov eax, ebx
			mov edx, ebx
			shr eax, cl

			movsx ebp, word ptr [esi + eax * 2]
			movsx eax, word ptr [esi + eax * 2 + 2]

			sub eax, ebp
			
			add ebx, [esp+12]
			and edx, [esp+4]
			
			imul edx
			
			shrd eax, edx, cl

			add eax, ebp

			mov [edi], eax
			
			add edi, 4
			
			cmp edi, [esp+8]
			jne pou

			pop ebp

			add esp, 12

			mov eax, [phase]
			mov [eax], ebx
	
	}


	xdsp.IntToFloatInPlace(pout, n);
}

*/

//*XDSP*
static void MWOscillatorRun(CXMWOscillator *_po, float *pout, dword const n, float const rate)
{
	CXMWOscillatorImp *po = (CXMWOscillatorImp *)_po;
	CXMipWaveImp const *pmw = po->pMW;
	
	
	dword const drate = (int)(rate * 65536.0 * 65536.0);

	dword level;

	for (level = 0; level < pmw->Limits.size(); level++)
	{
		if (drate < pmw->Limits[level])
			break;
	}
		
	if (level == pmw->Limits.size() || rate > 0.5)
	{
		// TODO: update phase
		xdsp.Clear_V(pout, n);
		return;
	}

	short const *data = pmw->Data + level * (pmw->LevelSize+3);
	dword const shift = (32 - pmw->log2LevelSize);

	RunOsc3(pout, n, data, po->Phase, drate, shift);	
}




#include "set/Osc.set"