
#include "stdafx.h"
#include <malloc.h>
#include "XDSPInterface.h"
#include "dspinternal.h"
#include "xdsp.h"
#include <emmintrin.h>
#include "amaths.h"


//*XDSP*S
static void SinApprox_V_V(float *pout, float const *pin, dword const n)
{
#ifndef _DEBUG

	dword i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
		_mm_store_ps(pout+i, am_sin_ps(_mm_load_ps(pin+i)));

	_mm_empty();

	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			_mm_store_ss(pout+i, am_sin_ss(_mm_load_ss(pin+i)));
	}

#endif
}

//*XDSP*S
static void CosApprox_V_V(float *pout, float const *pin, dword const n)
{
#ifndef _DEBUG

	dword i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
		_mm_store_ps(pout+i, am_cos_ps(_mm_load_ps(pin+i)));

	_mm_empty();

	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			_mm_store_ss(pout+i, am_cos_ss(_mm_load_ss(pin+i)));
	}

#endif
}

//*XDSP*S
static void TanApprox_V_V(float *pout, float const *pin, dword const n)
{
#ifndef _DEBUG

	dword i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
		_mm_store_ps(pout+i, am_tan_ps(_mm_load_ps(pin+i)));

	_mm_empty();

	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			_mm_store_ss(pout+i, am_tan_ss(_mm_load_ss(pin+i)));
	}

#endif
}


//*XDSP*S
static void AtanApprox_V_V(float *pout, float const *pin, dword const n)
{
#ifndef _DEBUG

	dword i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
		_mm_store_ps(pout+i, am_atan_ps(_mm_load_ps(pin+i)));

	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			_mm_store_ss(pout+i, am_atan_ss(_mm_load_ss(pin+i)));
	}

#endif
}

//*XDSP*S
static void ExpApprox_V_V(float *pout, float const *pin, dword const n)
{
#ifndef _DEBUG

	dword i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
		_mm_store_ps(pout+i, am_exp_ps(_mm_load_ps(pin+i)));

	_mm_empty();

	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			_mm_store_ss(pout+i, am_exp_ss(_mm_load_ss(pin+i)));
	}

#endif
}

//*XDSP*S
static void Pow2Approx_V_V(float *pout, float const *pin, dword const n)
{
#ifndef _DEBUG

	dword i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
		_mm_store_ps(pout+i, am_exp2_ps(_mm_load_ps(pin+i)));

	_mm_empty();

	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			_mm_store_ss(pout+i, am_exp2_ss(_mm_load_ss(pin+i)));
	}

#endif
}

//*XDSP*S
static void LogApprox_V_V(float *pout, float const *pin, dword const n)
{
#ifndef _DEBUG

	dword i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
		_mm_store_ps(pout+i, am_log_ps(_mm_load_ps(pin+i)));

	_mm_empty();

	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			_mm_store_ss(pout+i, am_log_ss(_mm_load_ss(pin+i)));
	}

#endif
}

//*XDSP*S
static void Log2Approx_V_V(float *pout, float const *pin, dword const n)
{
#ifndef _DEBUG

	dword i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
		_mm_store_ps(pout+i, am_log2_ps(_mm_load_ps(pin+i)));

	_mm_empty();

	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
			_mm_store_ss(pout+i, am_log2_ss(_mm_load_ss(pin+i)));
	}

#endif
}


//*XDSP*S
static void TanhApprox_V_V(float *pout, float const *pin, dword const n)
{
#ifndef _DEBUG

	// tanh: (e^2x - 1) / (e^2x + 1)

	__m128 const one = _mm_set_ps1(1.0f);

	dword i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		__m128 x = _mm_load_ps(pin+i);
		x = am_exp_ps(_mm_add_ps(x, x));
		_mm_store_ps(pout+i, _mm_div_ps(_mm_sub_ps(x, one), _mm_add_ps(x, one)));
	}

	_mm_empty();

	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
		{
			__m128 x = _mm_load_ss(pin+i);
			x = am_exp_ss(_mm_add_ss(x, x));
			_mm_store_ss(pout+i, _mm_div_ss(_mm_sub_ss(x, one), _mm_add_ss(x, one)));
		}
	}

#endif
}

//*XDSP*S
static void SinhApprox_V_V(float *pout, float const *pin, dword const n)
{
#ifndef _DEBUG

	// sinh: 1/2 * (e^x - e^-x)

	__m128 const half = _mm_set_ps1(0.5f);
	__m128 const minusone = _mm_set_ps1(-1.0f);
	
	dword i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		__m128 x = _mm_load_ps(pin+i);
		x = _mm_sub_ps(am_exp_ps(x), am_exp_ps(_mm_mul_ps(x, minusone)));
		_mm_store_ps(pout+i, _mm_mul_ps(x, half));
	}

	_mm_empty();

	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
		{
			__m128 x = _mm_load_ss(pin+i);
			x = _mm_sub_ss(am_exp_ss(x), am_exp_ss(_mm_mul_ss(x, minusone)));
			_mm_store_ss(pout+i, _mm_mul_ss(x, half));
			
		}
	}

#endif
}


//*XDSP*S
static void CoshApprox_V_V(float *pout, float const *pin, dword const n)
{
#ifndef _DEBUG

	// sinh: 1/2 * (e^x - e^-x)

	__m128 const half = _mm_set_ps1(0.5f);
	__m128 const minusone = _mm_set_ps1(-1.0f);
	
	dword i = 0;
	
	for (dword c = 0; c < n/4; c++, i += 4)
	{
		__m128 x = _mm_load_ps(pin+i);
		x = _mm_add_ps(am_exp_ps(x), am_exp_ps(_mm_mul_ps(x, minusone)));
		_mm_store_ps(pout+i, _mm_mul_ps(x, half));
	}

	_mm_empty();

	dword r = n & 3;
	if (r > 0)
	{
		for (dword c = 0; c < r; c++, i++)
		{
			__m128 x = _mm_load_ss(pin+i);
			x = _mm_add_ss(am_exp_ss(x), am_exp_ss(_mm_mul_ss(x, minusone)));
			_mm_store_ss(pout+i, _mm_mul_ss(x, half));
			
		}
	}

#endif
}

#include "set/TrigSSE.set"