#include "stdafx.h"
#include "../dspinternal.h"
#include "XDSPInterface.h"
#include <math.h>

static void Reduce(dword &num, dword &dem)
{
	while(1)
	{
		dword maks = __min(num, dem);

		dword x;
		for (x = 2; x <= maks; x++)
		{
			if (!(num % x) && !(dem % x))
				break;

		}

		if (x > maks)
			break;

		num /= x;
		dem /= x;
	}
	

}



#define IzeroEPSILON 1E-21				 /* Max error acceptable in Izero */

static double Izero(double x)
{
	double sum, u, halfx, temp;
	int n;
	
	sum = u = n = 1;
	halfx = x/2.0;
	do {
		temp = halfx/(double)n;
		n += 1;
		temp *= temp;
		u *= temp;
		sum += u;
	} while (u >= IzeroEPSILON*sum);
	return(sum);
}

static void LpFilter(double c[], int N, double frq, double Beta, int Num)
{
   double IBeta, temp, inm1;
   int i;

   /* Calculate ideal lowpass filter impulse response coefficients: */
   c[0] = 2.0*frq;
   for (i=1; i<N; i++) {
	   temp = M_PI * (double)i/(double)Num;
	   c[i] = sin(2.0*temp*frq)/temp; /* Analog sinc function, cutoff = frq */
   }

   /*
	* Calculate and Apply Kaiser window to ideal lowpass filter.
	* Note: last window value is IBeta which is NOT zero.
	* You're supposed to really truncate the window here, not ramp
	* it to zero. This helps reduce the first sidelobe.
	*/
   IBeta = 1.0/Izero(Beta);
   inm1 = 1.0/((double)(N-1));
   for (i=1; i<N; i++) {
	   temp = (double)i * inm1;
	   c[i] *= Izero(Beta*sqrt(1.0-temp*temp)) * IBeta;
   }
}

#define SINC_SIZE		16384


static inline double SampleFilter(double const *coeff, double x)
{
	double const ax = fabs(SINC_SIZE * x);

	int const iax = (int)ax;
	assert(iax <= SINC_SIZE);
	
	double const alpha = ax - iax;

	if (iax == SINC_SIZE)
		return 0;

	if (iax == SINC_SIZE-1)
		return (1.0 - alpha) * coeff[iax];	// interpolate from last point to zero

	return (1.0 - alpha) * coeff[iax] + alpha * coeff[iax+1];
}

static void MakePolyphaseFilter(dword const num, dword const denom, dword const filtersize, double const *coeff, float *ppf, dword *inttbl)
{
	double const ratio = (double)num / denom;
	double const ooratio = (double)denom / num;

	for (dword p = 0; p < num; p++)
	{
		double ofs = p * ooratio;
		

		double alpha = ofs - floor(ofs);

		inttbl[p] = (dword)(floor((p+1) * ooratio) - floor(ofs));


		double start = -1 + (2.0 / filtersize) * (1 - alpha);
		
		dword n = 0;

		do
		{
			double out = SampleFilter(coeff, start + n * (2.0 / filtersize));

//			printf("%f ", out);

			ppf[p * filtersize + n] = (float)out;

			n++;
		} while(n < filtersize);

//		printf("\n\n");
	}

#ifdef _DEBUG

	dword sumdelta = 0;

	for (dword i = 0; i < num; i++)
		sumdelta += inttbl[i];

	assert(sumdelta == denom);

#endif

}

//*XDSP*
static CXResamplerFilter *CreateKaiserSincFilter(dword inrate, dword outrate, dword fsize, float cutoff, float beta)
{
	CXResamplerFilterImp *ps = new CXResamplerFilterImp;

	dword num = outrate, denom = inrate;
	Reduce(num, denom);

	float const ratio = (float)num / denom;
	if (ratio < 1)
		cutoff *= ratio;

	ps->numPhases = num;
	ps->numCoeffsPerPhase = fsize;
	ps->OutInRatio = ratio;
	ps->Coefficients = new float [num * fsize];
	ps->IntStepTable = new dword [num];

	double *sinc = new double [SINC_SIZE];

	LpFilter(sinc, SINC_SIZE, cutoff, beta, 4 * SINC_SIZE / fsize);

	
	MakePolyphaseFilter(num, denom, fsize, sinc, ps->Coefficients, ps->IntStepTable);

	delete[] sinc;

	return ps;
}




void SetFilterDesign(CXDSPInterface *p)
{
	p->CreateKaiserSincFilter = CreateKaiserSincFilter;
}


bool CXResamplerFilterImp::Serialize(CStreamOut &so)
{
	so.Write(CXObject::ResamplerFilter);
	so.Write((byte)1);		// version
	so.Write(numPhases);
	so.Write(numCoeffsPerPhase);
	so.Write(OutInRatio);
	so.WriteRaw(Coefficients, numPhases * numCoeffsPerPhase * sizeof(float));
	so.WriteRaw(IntStepTable, numPhases * sizeof(dword));

	return true;
}

CXResamplerFilterImp *CXResamplerFilterImp::Deserialize(CStreamIn &si)
{
	byte ver;
	si.Read(ver);
	if (ver != 1)
		return NULL;

	CXResamplerFilterImp *pi = new CXResamplerFilterImp;
	si.Read(pi->numPhases);
	si.Read(pi->numCoeffsPerPhase);
	si.Read(pi->OutInRatio);
	pi->Coefficients = new float [pi->numPhases * pi->numCoeffsPerPhase];
	pi->IntStepTable = new dword [pi->numPhases];
	si.ReadRaw(pi->Coefficients, pi->numPhases * pi->numCoeffsPerPhase * sizeof(float));
	si.ReadRaw(pi->IntStepTable, pi->numPhases * sizeof(dword));
	
	return pi;
}


dword CXResamplerFilterImp::QueryMemoryUsage() const
{
	return sizeof(this) + (numPhases * numCoeffsPerPhase) * sizeof(float);
}

#include "set/filterdesign.set"
