#include "stdafx.h"
#include "dspinternal.h"
#include <malloc.h>
#include <xmmintrin.h>

extern CXDSPInterface xdsp;

class CXOscillatorArrayImp : public CXOscillatorArray
{
public:
	virtual ~CXOscillatorArrayImp()
	{
		_aligned_free(Amplitudes);
	}

public:
	dword numOscs;
	dword numReqOscs;
	float *TempBuffer;

	float *CosPhase;
	float *SinPhase;
	float *CosFreq;
	float *SinFreq;
};

//*XDSP*S
static void OscArraySetPhases(CXOscillatorArray *_poa, float const *phases)
{
	CXOscillatorArrayImp *poa = (CXOscillatorArrayImp *)_poa;

	if (phases == NULL)
	{
		xdsp.Clear_V(poa->SinPhase, poa->numOscs);
		xdsp.Copy_V_S(poa->CosPhase, poa->numOscs, 1.0f);
	}
	else
	{
		xdsp.SinCos_V2N_V(poa->SinPhase, poa->CosPhase, phases, poa->numReqOscs);
	}
}

//*XDSP*S
static CXOscillatorArray *OscArrayCreate(dword const _numoscs, float *tempbuffer)
{
	if (_numoscs == 0 || tempbuffer == NULL)
		return NULL;

	CXOscillatorArrayImp *poa = new CXOscillatorArrayImp;

	poa->numReqOscs = _numoscs;
	poa->numOscs = (_numoscs+7) & ~7;
	poa->TempBuffer = tempbuffer;
	poa->Amplitudes = (float *)_aligned_malloc(6 * poa->numOscs * sizeof(float), 32);
	poa->AmplitudeDeltas = poa->Amplitudes + poa->numOscs;
	poa->CosPhase = poa->Amplitudes + 2 * poa->numOscs;
	poa->SinPhase= poa->Amplitudes + 3 * poa->numOscs;
	poa->CosFreq = poa->Amplitudes + 4 * poa->numOscs;
	poa->SinFreq = poa->Amplitudes + 5 * poa->numOscs;
	
	xdsp.Clear_V(poa->Amplitudes, 6 * poa->numOscs);
	
	OscArraySetPhases(poa, NULL);

	return poa;
}

//*XDSP*S
static void OscArraySetFrequencies(CXOscillatorArray *_poa, float const *freqs)
{
	CXOscillatorArrayImp *poa = (CXOscillatorArrayImp *)_poa;
	xdsp.SinCos_V2N_V(poa->SinFreq, poa->CosFreq, freqs, poa->numReqOscs);

	for (dword i = 0; i < poa->numReqOscs; i++)
	{
		if (freqs[i] >= JXPI)
		{
			poa->SinFreq[i] = 0;
			poa->CosFreq[i] = 0;
			poa->SinPhase[i] = 0;
			poa->CosPhase[i] = 1;
		}
	}

}



static void Run(float *pout, dword const n, CXOscillatorArrayImp *poa)
{
	__m128 const cos4 = _mm_load_ps(poa->CosFreq);
	__m128 const sin4 = _mm_load_ps(poa->SinFreq);
	__m128 const negsin4 = _mm_mul_ps(sin4, _mm_set_ps1(-1));
	__m128 a4 = _mm_load_ps(poa->Amplitudes);
	__m128 const da4 = _mm_load_ps(poa->AmplitudeDeltas);

	__m128 const cos42 = _mm_load_ps(poa->CosFreq+4);
	__m128 const sin42 = _mm_load_ps(poa->SinFreq+4);
	__m128 const negsin42 = _mm_mul_ps(sin42, _mm_set_ps1(-1));
	__m128 a42 = _mm_load_ps(poa->Amplitudes+4);
	__m128 const da42 = _mm_load_ps(poa->AmplitudeDeltas+4);
	
	__m128 x = _mm_load_ps(poa->CosPhase);
	__m128 y = _mm_load_ps(poa->SinPhase);
	
	__m128 x2 = _mm_load_ps(poa->CosPhase+4);
	__m128 y2 = _mm_load_ps(poa->SinPhase+4);
	
	
	for (dword i = 0; i < n; i++)
	{
		__m128 xn = _mm_mul_ps(y, a4);
		xn = _mm_add_ps(xn, _mm_mul_ps(y2, a42));
		
		a4 = _mm_add_ps(a4, da4);
		a42 = _mm_add_ps(a42, da42);
		
		_mm_store_ps(pout + i * 4, xn);
		
		xn = _mm_add_ps(_mm_mul_ps(cos4, x), _mm_mul_ps(negsin4, y));
		__m128 xn2 = _mm_add_ps(_mm_mul_ps(cos42, x2), _mm_mul_ps(negsin42, y2));

		y = _mm_add_ps(_mm_mul_ps(sin4, x), _mm_mul_ps(cos4, y));
		y2 = _mm_add_ps(_mm_mul_ps(sin42, x2), _mm_mul_ps(cos42, y2));

		x = xn;
		x2 = xn2;
		
		
	}
	
	
	_mm_store_ps(poa->SinPhase, y);
	_mm_store_ps(poa->CosPhase, x);
	_mm_store_ps(poa->Amplitudes, a4);

	_mm_store_ps(poa->SinPhase+4, y2);
	_mm_store_ps(poa->CosPhase+4, x2);
	_mm_store_ps(poa->Amplitudes+4, a42);
	
}

static void RunAdd(float *pout, dword const n, CXOscillatorArrayImp *poa, dword const oi)
{
	__m128 const cos4 = _mm_load_ps(poa->CosFreq + oi);
	__m128 const sin4 = _mm_load_ps(poa->SinFreq + oi);
	__m128 const negsin4 = _mm_mul_ps(sin4, _mm_set_ps1(-1));
	__m128 a4 = _mm_load_ps(poa->Amplitudes + oi);
	__m128 const da4 = _mm_load_ps(poa->AmplitudeDeltas + oi);

	__m128 const cos42 = _mm_load_ps(poa->CosFreq + oi + 4);
	__m128 const sin42 = _mm_load_ps(poa->SinFreq + oi + 4);
	__m128 const negsin42 = _mm_mul_ps(sin42, _mm_set_ps1(-1));
	__m128 a42 = _mm_load_ps(poa->Amplitudes + oi + 4);
	__m128 const da42 = _mm_load_ps(poa->AmplitudeDeltas + oi + 4);
	
	__m128 x = _mm_load_ps(poa->CosPhase + oi);
	__m128 y = _mm_load_ps(poa->SinPhase + oi);

	__m128 x2 = _mm_load_ps(poa->CosPhase + oi + 4);
	__m128 y2 = _mm_load_ps(poa->SinPhase + oi + 4);
	
	for (dword i = 0; i < n; i++)
	{
		__m128 xn = _mm_mul_ps(y, a4);
		xn = _mm_add_ps(xn, _mm_mul_ps(y2, a42));

		a4 = _mm_add_ps(a4, da4);
		a42 = _mm_add_ps(a42, da42);
		
		_mm_store_ps(pout + i * 4, _mm_add_ps(_mm_load_ps(pout + i * 4), xn));
		
		xn = _mm_add_ps(_mm_mul_ps(cos4, x), _mm_mul_ps(negsin4, y));
		__m128 xn2= _mm_add_ps(_mm_mul_ps(cos42, x2), _mm_mul_ps(negsin42, y2));
		
		y = _mm_add_ps(_mm_mul_ps(sin4, x), _mm_mul_ps(cos4, y));
		y2 = _mm_add_ps(_mm_mul_ps(sin42, x2), _mm_mul_ps(cos42, y2));

		x = xn;
		x2 = xn2;
	}
	
	
	_mm_store_ps(poa->SinPhase + oi, y);
	_mm_store_ps(poa->CosPhase + oi, x);
	_mm_store_ps(poa->Amplitudes + oi, a4);

	_mm_store_ps(poa->SinPhase + oi + 4, y2);
	_mm_store_ps(poa->CosPhase + oi + 4, x2);
	_mm_store_ps(poa->Amplitudes + oi + 4, a42);
	
} 
        

//*XDSP*S
static void OscArrayRun(CXOscillatorArray *_poa, float *pout, dword const n)
{
	CXOscillatorArrayImp *poa = (CXOscillatorArrayImp *)_poa;

	Run(poa->TempBuffer, n, poa);

	for (dword i = 8; i < poa->numOscs; i += 8)
		RunAdd(poa->TempBuffer, n, poa, i);
	
	xdsp.Copy_V_V4(pout, poa->TempBuffer, n);

}


#include "set/oscarray.set"