
#include "stdafx.h"
#include "xdspinterface.h"
#include "dspinternal.h"

template <dword BITS>
class CP2LUT
{
private:
	enum { SIZE = 1 << BITS };
	enum { ALPHA_BITS = (23 - BITS) };
	enum { ALPHA_MASK = ((1 << ALPHA_BITS)-1) };
	
public:
	CP2LUT()
	{
		assert(BITS >= 8);	// otherwise 32bit imul will fail 

		for (dword i = 0; i <= SIZE; i++)
			Data[i] = (int)((pow(2, i / (double)SIZE) - 1) * (1 << 23));

	}

	__forceinline dword LerpLookup(dword const mantissa) const
	{
		dword const ilut = mantissa >> ALPHA_BITS;
//		return Data[ilut] + fpmath<ALPHA_BITS>::MulSar((mantissa & ALPHA_MASK), Data[ilut+1] - Data[ilut]);
		return Data[ilut] + ((mantissa & ALPHA_MASK) * (Data[ilut+1] - Data[ilut]) >> ALPHA_BITS);
		
	}
	
	void Dump(std::ofstream &os) const
	{
		for (dword i = 0; i <= SIZE; i++)
			os << Data[i] << "\n";			
	}

private:
	dword Data[SIZE+1];
	
};

static CP2LUT<8> const Pow2LUT;

float XPow2(float const x)
{
	int intpart = FloatToIntFloor(x);
	
	float const frac = (x - (intpart-1));	// frac [1..2[
	dword const dwfrac = *(dword *)&frac;

	// fixes the case of frac being >= 2.0f because of float rounding issues
	intpart = (intpart << 23) + (dwfrac & 0x7f800000);

	dword const mantissa = dwfrac & ((1 << 23)-1);
	dword const newmantissa = Pow2LUT.LerpLookup(mantissa);
	dword const res = newmantissa | intpart;

	return *(float *)&res;
}

//*XDSP*
static void Pow2Approx2_V_V(float *pout, float const *pin, dword const n)
{
	if (n > 256)
		return;

	int intbuf[256];
	
	_control87(_RC_DOWN, _MCW_RC);

	__asm
	{
		mov esi, pin
		lea edi, intbuf
		mov ecx, n
loopi:
		fld [esi]
		fistp [edi]

		add esi, 4
		add edi, 4
		dec ecx
		jnz loopi

	}

	_control87(_RC_NEAR, _MCW_RC);
	

	for (dword i = 0; i < n; i++)
	{
		float x = pin[i];
		int intpart = intbuf[i];
		
		float const frac = (x - (intpart-1));	// frac [1..2[
		dword const dwfrac = *(dword *)&frac;
		
		// fixes the case of frac being >= 2.0f because of float rounding issues
		intpart = (intpart << 23) + (dwfrac & 0x7f800000);
		
		dword const mantissa = dwfrac & ((1 << 23)-1);
		dword const newmantissa = Pow2LUT.LerpLookup(mantissa);
		dword const res = newmantissa | intpart;
		
		((dword *)pout)[i] = res;
	}

}

/*
void SetPow2(CXDSPInterface *p)
{
	p->Pow2Approx_V = Pow2Approx_V;
}
*/

#include "set/pow2.set"