
#include "stdafx.h"
#include <malloc.h>
#include "XDSPInterface.h"
#include "dspinternal.h"
#include "xdsp.h"

class CXPreparedResamplerFilterFPU : public CXPreparedResamplerFilter
{
public:
	CXPreparedResamplerFilterFPU()
	{
		Coefficients = NULL;
		IntStepTable = NULL;
	}
	
	virtual ~CXPreparedResamplerFilterFPU()
	{
		delete[] Coefficients;
		delete[] IntStepTable;
	}
	
	virtual dword QueryMemoryUsage() const
	{
		return sizeof(this) + (numPhases + numCoeffsPerPhase + numPhases) * 4;
	}
	
public:
	dword numPhases;
	dword numCoeffsPerPhase;
	float *Coefficients;
	dword *IntStepTable;		
	dword numChannels;	
	float OutInRatio;

};

class CXResamplerFPU : public CXResampler
{
public:
	CXResamplerFPU()
	{
		Phase = 0;
		Buffer = NULL;
		OwnsBuffer = false;
	}

	virtual ~CXResamplerFPU()
	{
		if (OwnsBuffer)
			xdsp.DeleteBuffer(Buffer);
	}
		
	virtual dword QueryMemoryUsage() const
	{
		dword x = sizeof(this);
		if (OwnsBuffer)
			x += BufferSize * sizeof(float);
			
		return x;
	}
		
public:
	CXPreparedResamplerFilterFPU *pD;
	CXResamplerCallback pCB;
	void *Context;

	dword Phase;
	dword BufferSize;
	float *Buffer;

	bool OwnsBuffer;

};


//*XDSP*
static CXPreparedResamplerFilter *PrepareResamplerFilter(CXResamplerFilter *prf, dword numch)
{
	CXResamplerFilterImp *i = (CXResamplerFilterImp *)prf;

	CXPreparedResamplerFilterFPU *p = new CXPreparedResamplerFilterFPU;
	p->numChannels = numch;
	p->numPhases = i->numPhases;
	p->numCoeffsPerPhase = i->numCoeffsPerPhase;
	p->OutInRatio = i->OutInRatio;
	p->Coefficients = new float [p->numPhases * p->numCoeffsPerPhase];
	p->IntStepTable = new dword [p->numPhases];
	memcpy(p->Coefficients, i->Coefficients, p->numPhases * p->numCoeffsPerPhase * sizeof(float));
	memcpy(p->IntStepTable, i->IntStepTable, p->numPhases * sizeof(dword));
	

	return p;
}

//*XDSP*
static CXResampler *ResamplerCreate(CXPreparedResamplerFilter *pf, CXResamplerCallback pcb, void *cbcontext, dword const maxbuffersize, float *tempbuffer)
{
	CXResamplerFPU *pr = new CXResamplerFPU;
	pr->pD = (CXPreparedResamplerFilterFPU *)pf;
	pr->pCB = pcb;
	pr->Context = cbcontext;

	dword reqbufsize = (int)(maxbuffersize / pr->pD->OutInRatio) + 1;


	pr->BufferSize = pr->pD->numChannels * (pr->pD->numCoeffsPerPhase + reqbufsize);
	
	if (tempbuffer == NULL)
	{
		pr->OwnsBuffer = true;
		pr->Buffer = xdsp.NewBuffer(pr->BufferSize);
	}
	else
	{
		pr->OwnsBuffer = false;
		pr->Buffer = tempbuffer;
	}

	xdsp.Clear_V(pr->Buffer, pr->BufferSize);

	return pr;
}

extern float g_SilenceThreshold;


//*XDSP*
static void ResamplerRun(CXResampler *_pr, float *pout, dword const numsamples)
{
	CXResamplerFPU *pr = (CXResamplerFPU *)_pr;
	CXPreparedResamplerFilterFPU *pd = pr->pD;

	dword p = pr->Phase;
	dword needsamples = 0;
	
	for (dword i = 0; i < numsamples; i++)
	{
		needsamples += pd->IntStepTable[p];
		p++;
		if (p == pd->numPhases)
			p = 0;
	}
	
	float * const bufstart = pr->Buffer + pd->numChannels * pd->numCoeffsPerPhase;

	pr->pCB(bufstart, needsamples, pr->Context);

	if (!xdsp.AbsMaxGT_V_S(pr->Buffer, pr->BufferSize, g_SilenceThreshold))
	{
		xdsp.Clear_V(pout, numsamples * pd->numChannels);
		return;
	}
		

	dword inpos = 0;
	
	for (i = 0; i < numsamples; i++)
	{
		float const *pfc = pd->Coefficients + pr->Phase * pd->numCoeffsPerPhase;
		float const *pin = bufstart + pd->numChannels * (inpos - (pd->numCoeffsPerPhase-1));
		
		if (pd->numChannels == 1)
		{
			double acc = 0;
			
			for (dword c = 0; c < pd->numCoeffsPerPhase; c++, pfc++, pin++)
				acc += *pfc * *pin; 
			
			pout[i] = (float)acc;
			
		}
		else if (pd->numChannels == 2)
		{
			double lacc = 0, racc = 0;

			for (dword c = 0; c < pd->numCoeffsPerPhase; c++, pfc++, pin += 2)
			{
				lacc += *pfc * pin[0]; 
				racc += *pfc * pin[1];
			}
			
			pout[2*i+0] = (float)lacc;
			pout[2*i+1] = (float)racc;
		}
		
		
		inpos += pd->IntStepTable[pr->Phase];
		
		pr->Phase++;
		if (pr->Phase == pd->numPhases)
			pr->Phase = 0;
	}
	
	memcpy(bufstart - pd->numChannels * (pd->numCoeffsPerPhase-1), bufstart + pd->numChannels * (inpos - (pd->numCoeffsPerPhase-1)), sizeof(float) * pd->numChannels * (pd->numCoeffsPerPhase-1));



}


#include "set/resampler.set"