#pragma once

#include "jxstream.h"
#include "jxmath.h"
#include "xdspinterface.h"

class CXResamplerFilterImp : public CXResamplerFilter
{
public:

	CXResamplerFilterImp()
	{
		Coefficients = NULL;
		IntStepTable = NULL;
	}

	virtual ~CXResamplerFilterImp()
	{
		delete[] Coefficients;
		delete[] IntStepTable;
	}

	virtual dword QueryMemoryUsage() const;
	
	virtual bool Serialize(CStreamOut &so);

	static CXResamplerFilterImp *Deserialize(CStreamIn &si);

public:
	dword numPhases;
	dword numCoeffsPerPhase;
	float *Coefficients;
	dword *IntStepTable;
	float OutInRatio;
};


class CXMipWaveImp : public CXMipWave
{
public:
	virtual ~CXMipWaveImp() {}

	virtual dword QueryMemoryUsage() const;
	

public:
	std::vector<dword> Limits;
	dword LevelSize;
	dword log2LevelSize;
	short *Data;
};

class CXMWOscillatorImp : public CXMWOscillator
{
public:
	virtual ~CXMWOscillatorImp() {}
	
	virtual dword QueryMemoryUsage() const { return sizeof(this); }
	
public:
	CXMipWaveImp const *pMW;
	dword Phase;
	

};

#define undenormalise(sample) if(((*(unsigned int*)&sample)&0x7f800000)==0) sample=0.0f

#define _SHORT_MIN	-32768
#define _SHORT_MAX	32767

#define _INT_MIN		-2147483648.0
#define _INT_MAX		2147483647.0



#define _SHORT_MINI	-32768
#define _SHORT_MAXI	32767

#define _INT_MINI		0x80000000
#define _INT_MAXI		0x7fffffff


#define M128_MOV(DEST, I, SOURCE) *(int *)&(DEST.m128_f32[I]) = *(int *)&SOURCE

#define FI_USE_MMX		(1<<0)
#define FI_USE_SSE		(1<<1)
#define FI_USE_SSE2		(1<<2)
#define FI_PREF_P3 		(1<<3)
#define FI_PREF_P4 		(1<<4)
#define FI_PREF_ATHLON	(1<<5)

// return one when mask is 1, return zero when mask is 0
#define MASK_COMBINE(one, zero, mask) _mm_or_ps(_mm_and_ps((mask), (one)), _mm_andnot_ps((mask), (zero)))


struct CXFIInfo
{
	char const *Name;
	void *FuncPtr;
	int Flags;
};


typedef void (*SET_FUNC_PTR)(CXDSPInterface *p, int const f);