#include "stdafx.h"
#include "dspinternal.h"
#include "dsp/set/setlist.h"
#include "cpuid.h"
#include "xdsp.h"
#include "../BuildCount.include"
 
#pragma warning (disable : 4073)
#pragma init_seg(lib)


static char const VersionString[] = "XDSP build " BUILD_COUNT " " __DATE__ " " __TIME__ " Copyright 1997-2003 Oskari Tammelin";
 
char const *GetXDSPVersionString() { return VersionString; }
 
class CProcInfo
{
public:
	CProcInfo()
	{
		int i = _cpuid(&pinfo);
	}
	
	bool SSE() const { return (pinfo.feature & _CPU_FEATURE_SSE) && (pinfo.os_support & _CPU_FEATURE_SSE); }
	bool SSE2() const { return (pinfo.feature & _CPU_FEATURE_SSE2) && (pinfo.os_support & _CPU_FEATURE_SSE2); }
	bool MMX() const { return (pinfo.feature & _CPU_FEATURE_MMX) && (pinfo.os_support & _CPU_FEATURE_MMX); }
	bool _3DNOW() const { return (pinfo.feature & _CPU_FEATURE_3DNOW) && (pinfo.os_support & _CPU_FEATURE_3DNOW); }
	
private:
	_p_info pinfo;
	
};

static CProcInfo ProcInfo;


bool const MMXSupported = ProcInfo.MMX();
bool const SSESupported = ProcInfo.SSE();
bool const SSE2Supported = ProcInfo.SSE2();
bool const _3DNOWSupported = ProcInfo._3DNOW();

 

static void DoSet(CXDSPInterface *p, int const f)
{
	SET_FUNC_PTR *ps = SetFuncList;
	while(*ps)
	{
		(*ps)(p, f);
		ps++;
	}
}

void SetAll(CXDSPInterface *p, bool use_sse2, bool use_sse, bool use_mmx)
{
	use_sse2 &= SSE2Supported;
	use_sse &= SSESupported;
	use_mmx &= MMXSupported;
	

	DoSet(p, 0);	// fpu first

	if (use_mmx)
		DoSet(p, FI_USE_MMX);

	if (use_sse)
		DoSet(p, FI_USE_SSE);

	if (use_sse && use_mmx)
		DoSet(p, FI_USE_SSE | FI_USE_MMX);
	
	if (use_sse2)
		DoSet(p, FI_USE_SSE2);
	
	if (use_sse2 && use_mmx)
		DoSet(p, FI_USE_SSE2 | FI_USE_MMX);
}




